!******************************************************************************|
! channel.f90, the channel-flow solvers for diablo.                  VERSION 1.0
! This solver was written by John R. Taylor.
!******************************************************************************|
!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine init_chan
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  ! This subroutine initializes various variables for the serial channel
  ! flow subroutine. This is not called if MPI is turned on
  ! Initialize any constants here

  include 'header'
  integer i, j, k, n

  pi = 4.d0 * atan(1.d0)

  ! Defined starting and ending indices in the wall-bounded direction
  ! These will be used for loops in the GYF variables
  ! The start and end of these loops depends on the bounary conditions
  ! For example, Direchlet BCs do not need to update the wall values

  if (u_BC_Ymin == 0) then
    jstart = 2
  else if (u_BC_Ymin == 1) then
    jstart = 1
  else
    jstart = 2
  end if
  ! Now, set the indexing for the scalar equations
  do n = 1, N_th
    if (th_BC_Ymin(n) == 0) then
      jstart_th(n) = 2
    else if (th_BC_Ymin(n) == 1) then
      jstart_th(n) = 1
    else
      jstart_th(n) = 2
    end if
  end do
  if (u_BC_Ymax == 0) then
    jend = Nyp - 1
  else if (u_BC_Ymax == 1) then
    jend = Nyp
  else
    jend = Nyp - 1
  end if

  ! Set the upper and lower limits of timestepping of the scalar equations
  do n = 1, N_th
    if (th_BC_Ymax(n) == 0) then
      jend_th(n) = Nyp - 1
    else if (th_BC_Ymax(n) == 1) then
      jend_th(n) = Nyp
    else
      jend_th(n) = Nyp - 1
    end if
  end do

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine rk_chan_1
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  ! Main time-stepping algorithm for the channel-flow case.
  ! This algorithm uses Crank-Nicolson for viscous/diffusive terms and
  ! and 3rd order Runge-Kutta for the rest of the terms
  ! INPUTS  (in Fourier space):  CUi, CP, and (if k>1) CFi at (k-1)  (for i=1,2,3)
  ! OUTPUTS (in Fourier space):  CUi, CP, and (if k<3) CFi at (k)
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  include 'header'

  integer i, j, k, n, istart
  real(rkind) temp1, temp2, temp3, temp4, temp5, ubulk

  ! Define the constants that are used in the time-stepping
  ! For reference, see Numerical Renaissance
  temp1 = 0.5d0 * nu * h_bar(rk_step)
  temp2 = 0.5d0 * h_bar(rk_step)
  temp3 = zeta_bar(rk_step) * h_bar(rk_step)
  temp4 = h_bar(rk_step)
  temp5 = beta_bar(rk_step) * h_bar(rk_step)

  ! First, we will compute the explicit RHS terms and store in Ri
  ! Note, Momentum equation and hence the RHS is evaluated at the
  ! corresponding velocity points.

  ! Store the old velocity in the RHS vector
  do j = jstart, jend
    do k = 0, twoNkz
      do i = 0, Nxp - 1
        cr1(i, k, j) = cu1(i, k, j)
        cr3(i, k, j) = cu3(i, k, j)
      end do
    end do
  end do
  do j = 2, Nyp
    do k = 0, twoNkz
      do i = 0, Nxp - 1
        cr2(i, k, j) = cu2(i, k, j)
      end do
    end do
  end do

  ! Add the R-K term from the rk-1 step
  if (rk_step > 1) then
    do j = jstart, jend
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          cr1(i, k, j) = cr1(i, k, j) + temp3 * cf1(i, k, j)
          cr3(i, k, j) = cr3(i, k, j) + temp3 * cf3(i, k, j)
        end do
      end do
    end do
    do j = 2, Nyp
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          cr2(i, k, j) = cr2(i, k, j) + temp3 * cf2(i, k, j)
        end do
      end do
    end do
  end if

  ! Take the y-derivative of the pressure at GY points in Fourier space
  do j = 2, Nyp
    do k = 0, twoNkz
      do i = 0, Nxp - 1
        cs1(i, k, j) = (cp(i, k, j) - cp(i, k, j - 1)) / dy(j)
      end do
    end do
  end do

  ! Add the pressure gradient to the RHS as explicit Euler
  do j = jstart, jend
    do k = 0, twoNkz
      do i = 0, Nxp - 1
        cr1(i, k, j) = cr1(i, k, j) - temp4 * (cikx(i) * cp(i, k, j))
        cr3(i, k, j) = cr3(i, k, j) - temp4 * (cikz(k) * cp(i, k, j))
      end do
    end do
  end do

  do j = 2, Nyp
    do k = 0, twoNkz
      do i = 0, Nxp - 1
        cr2(i, k, j) = cr2(i, k, j) - temp4 * cs1(i, k, j)
      end do
    end do
  end do

  ! Here, add the constant, forcing pressure gradient
  ! There are two built-in ways of doing this
  ! f_type=1 -> Constant pressure gradient in the x-direction
  ! f_type=2 -> Oscillatory pressure gradient in the x-direction
  ! else -> No forcing added
  if (f_type == 1) then
    ! Add forcing for a constant pressure gradient
    do j = jstart, jend
      if (rankZ == 0) cr1(0, 0, j) = cr1(0, 0, j) - temp4 * px0
    end do
  else if (f_type == 2) then
    ! Oscillatory pressure gradient
    do j = jstart, jend
      if (rankZ == 0) cr1(0, 0, j) = cr1(0, 0, j) - &
                                       temp4 * (px0 + amp_omega0 * cos(omega0 * time))
    end do
    ! End if forcing type
  end if

  ! Now compute the term R-K term Ai
  ! Compile terms of Ai in CFi which will be saved for next time step
  ! First, store the horizontal viscous terms in CFi
  ! Note that Beta is the order of the Laplacian operator
  ! e.g. Beta=1 for second order viscosity/diffusivity
  ! or   Beta=2, etc. for hyper viscosity/diffusivity
  do j = jstart, jend
    do k = 0, twoNkz
      do i = 0, Nxp - 1
        cf1(i, k, j) = -nu * (kx2(i) + kz2(k))**beta * cu1(i, k, j)
        cf3(i, k, j) = -nu * (kx2(i) + kz2(k))**beta * cu3(i, k, j)
      end do
    end do
  end do
  do j = 2, Nyp
    do k = 0, twoNkz
      do i = 0, Nxp - 1
        cf2(i, k, j) = -nu * (kx2(i) + kz2(k))**beta * cu2(i, k, j)
      end do
    end do
  end do

  ! Add the terms owing to the system rotation (Coriolis terms)
  ! Assume that the flow is on an f-plane
  do k = 0, twoNkz
    do i = 0, Nxp - 1
      do j = jstart, jend
        cf1(i, k, j) = cf1(i, k, j) + (Ro_inv / delta) * cu3(i, k, j)
        cf3(i, k, j) = cf3(i, k, j) - (Ro_inv / delta) * cu1(i, k, j)
      end do
    end do
  end do

  ! Do for each scalar
  do n = 1, N_th

    ! If a scalar contributes to the denisty, RI is not equal to zero and
    ! add the buoyancy term as explicit R-K.  Don't add the 0,0 mode in the
    ! y-direction, which corresponds to the plane-average.
    ! The plane averaged density balances the hydrostati! pressure
    do j = 2, Nyp
      do k = 1, twoNkz
        do i = 0, Nxp - 1
          ! Use second order interpolation
          cf2(i, k, j) = cf2(i, k, j) + grav_y * &
                         (cth(i, k, j, n) * dyf(j - 1) + &
                          cth(i, k, j - 1, n) * dyf(j)) / (2.d0 * dy(j))
        end do
      end do
      k = 0
      if (rankZ == 0) then
        istart = 1
      else
        istart = 0
      end if
      do i = istart, Nxp - 1
        cf2(i, k, j) = cf2(i, k, j) + grav_y * &
                       (cth(i, k, j, n) * dyf(j - 1) + &
                        cth(i, k, j - 1, n) * dyf(j)) / (2.d0 * dy(j))
      end do
    end do

    do j = jstart, jend
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          cf1(i, k, j) = cf1(i, k, j) + grav_x * cth(i, k, j, n)
        end do
      end do
    end do

    do j = jstart, jend
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          cf3(i, k, j) = cf3(i, k, j) + grav_z * cth(i, k, j, n)
        end do
      end do
    end do

    ! Now, compute the RHS vector for the scalar equations
    ! Since TH is defined at horizontal velocity points, the
    ! scalar update equation will be very similar to the horizontal
    ! velocity update.

    ! We will store the RHS scalar terms in CRTH, RTH
    ! The k-1 term for the R-K stepping is saved in FTH, CFTH

    ! First, build the RHS vector, use CRTH
    do j = jstart_th(n), jend_th(n)
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          crth(i, k, j, n) = cth(i, k, j, n)
        enddo
      end do
    end do
    ! Add term from k-2 step to free up CFTH variable
    if (rk_step > 1) then
      do j = jstart_th(n), jend_th(n)
        do k = 0, twoNkz
          do i = 0, Nxp - 1
            crth(i, k, j, n) = crth(i, k, j, n) + temp3 * cfth(i, k, j, n)
          end do
        end do
      end do
    end if

    ! Now compute the explicit R-K term Ai
    ! Compile terms of Ai in CFi which will be saved for next time step
    do j = jstart_th(n), jend_th(n)
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          cfth(i, k, j, n) = -(nu / Pr(n)) * &
                              (kx2(i) + kz2(k))**beta * cth(i, k, j, n)
        end do
      end do
    end do

    ! Add advection acting on the background scalar gradient if present
    do j = jstart_th(n), jend_th(n)
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          cfth(i, k, j, n) = cfth(i, k, j, n) - cu1(i, k, j) * dTHdX(n) &
                             - cu3(i, k, j) * dTHdZ(n)
        end do
      end do
    end do

    ! End do number of passive scalars (N_th)
  end do

  ! Optionally, add user forcing to the right hand side
  ! Here, we have U1, U2, U3, and TH in Fourier space
  call user_rhs_chan_fourier

  ! If we are considering an LES, then add the subgrid scale stress:
  ! Here, velocity and CFi should be in Fourier space
  ! The subgrid scale stress is added to CFi:   CFi=CFi - d/dx_i tau_ij

  if (LES .and. ((.not. create_new_flow) .or. (time_step > 100))) then
    ! If we have created new flow with random perturbations, wait for a
    ! spinup before applying the subgrid model for stability purposes
    ! In the process, Ui is converted to physical space
    call les_chan

    ! APPLY constant SGS Prandlt number
    do n = 1, N_th
      do j = 1, Nyp + 1
        do k = 0, Nzp - 1
          do i = 0, Nxm1
            kappa_t(i, k, j, n) = 1.d0 * nu_t(i, k, j)
          end do
        end do
      end do
    end do

    ! Add the horizontal diffusive terms using explicit timestepping
    ! This is already done for the viscous terms inside les_chan.f

    do n = 1, N_th

    do j = 1, Nyp + 1
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          cs1(i, k, j) = cikx(i) * cth(i, k, j, n)
        end do
      end do
    end do
    call fft_xz_to_physical(cs1, s1, 0, Nyp + 1)
    do j = 1, Nyp + 1
      do k = 0, Nzp - 1
        do i = 0, Nxm1
          s1(i, k, j) = kappa_t(i, k, j, n) * s1(i, k, j)
        end do
      end do
    end do
    call fft_xz_to_fourier(s1, cs1, 0, Nyp + 1)
    do j = jstart_th(n), jend_th(n)
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          cfth(i, k, j, n) = cfth(i, k, j, n) + cikx(i)*cs1(i, k, j)
        end do
      end do
    end do

    do j = 1, Nyp + 1
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          cs1(i, k, j) = cikz(k) * cth(i, k, j, n)
        end do
      end do
    end do
    call fft_xz_to_physical(cs1, s1, 0, Nyp + 1)
    do j = 1, Nyp + 1
      do k = 0, Nzp - 1
        do i = 0, Nxm1
          s1(i, k, j) = kappa_t(i, k, j, n) * s1(i, k, j)
        end do
      end do
    end do
    call fft_xz_to_fourier(s1, cs1, 0, Nyp + 1)
    do j = jstart_th(n), jend_th(n)
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          cfth(i, k, j, n) = cfth(i, k, j, n) + cikz(k) * cs1(i, k, j)
        end do
      end do
    end do
    end do ! end do n

    ! Now, convert TH to physical space for calculation of nonlinear terms
    do n = 1, N_th
      cs1(:, :, :) = cth(:, :, :, n)
      call fft_xz_to_physical(cs1, s1, 0, Nyp + 1)
      th(:, :, :, n) = s1(:, :, :)
    end do


  else if (LES .and. flag_save_LES) then
    ! Save out blanks for the LES Stats to keep up with flow stats
    flag_save_LES = .false.
    call save_stats_LES_OOL(.true.)
  else
    ! If the subgrid model hasn't been called, then it is necessary to
    ! convert to physical space.
    call fft_xz_to_physical(cu1, u1, 0, Nyp + 1)
    call fft_xz_to_physical(cu2, u2, 0, Nyp + 1)
    call fft_xz_to_physical(cu3, u3, 0, Nyp + 1)
    ! Transform THETA to physical space for computation of nonlinear terms
    ! Here pass the first location in memory of the array for scalar n

    do n = 1, N_th
      cs1(:, :, :) = cth(:, :, :, n)
      call fft_xz_to_physical(cs1, s1, 0, Nyp + 1)
      th(:, :, :, n) = s1(:, :, :)
    end do

  end if

  ! Compute the nonlinear products in physical space, then transform
  ! back to Fourier space to compute the derivative.
  ! Here, we compute the horizontal derivatives of the nonlinear terms
  ! which will be treated with RKW3.
  ! Do terms one at a time to save on memory
  ! U1*U3
  do j = jstart, jend
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        s1(i, k, j) = u3(i, k, j) * u1(i, k, j)
      end do
    end do
  end do

  call fft_xz_to_fourier(s1, cs1, 0, Nyp + 1)

  do j = jstart, jend
    do k = 0, twoNkz
      do i = 0, Nxp - 1
        cf1(i, k, j) = cf1(i, k, j) - cikz(k) * cs1(i, k, j)
        cf3(i, k, j) = cf3(i, k, j) - cikx(i) * cs1(i, k, j)
      end do
    end do
  end do

  ! U1*U1
  do j = jstart, jend
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        s1(i, k, j) = u1(i, k, j) * u1(i, k, j)
      end do
    end do
  end do

  call fft_xz_to_fourier(s1, cs1, 0, Nyp + 1)

  do j = jstart, jend
    do k = 0, twoNkz
      do i = 0, Nxp - 1
        cf1(i, k, j) = cf1(i, k, j) - cikx(i) * cs1(i, k, j)
      end do
    end do
  end do

  ! U3*U3
  do j = jstart, jend
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        s1(i, k, j) = u3(i, k, j) * u3(i, k, j)
      end do
    end do
  end do

  call fft_xz_to_fourier(s1, cs1, 0, Nyp + 1)

  do j = jstart, jend
    do k = 0, twoNkz
      do i = 0, Nxp - 1
        cf3(i, k, j) = cf3(i, k, j) - cikz(k) * cs1(i, k, j)
      end do
    end do
  end do

  ! U1*U2
  do j = 2, Nyp
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        s1(i, k, j) = ((dyf(j) * u1(i, k, j) &
                        + dyf(j - 1) * u1(i, k, j - 1)) / (2.*dy(j))) &
                      * u2(i, k, j)
      end do
    end do
  end do

  call fft_xz_to_fourier(s1, cs1, 0, Nyp + 1)

  do j = 2, Nyp
    do k = 0, twoNkz
      do i = 0, Nxp - 1
        cf2(i, k, j) = cf2(i, k, j) - cikx(i) * cs1(i, k, j)
      end do
    end do
  end do

  ! U3*U2
  do j = 2, Nyp
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        s1(i, k, j) = ((dyf(j) * u3(i, k, j) &
                        + dyf(j - 1) * u3(i, k, j - 1)) / (2.*dy(j))) &
                      * u2(i, k, j)
      end do
    end do
  end do

  call fft_xz_to_fourier(s1, cs1, 0, Nyp + 1)

  do j = 2, Nyp
    do k = 0, twoNkz
      do i = 0, Nxp - 1
        cf2(i, k, j) = cf2(i, k, j) - cikz(k) * cs1(i, k, j)
      end do
    end do
  end do

  ! Add the vertical derivative term
  do j = jstart, jend
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        s1(i, k, j) = &
          (u1(i, k, j + 1) * u2(i, k, j + 1) + u1(i, k, j) * u2(i, k, j + 1) &
           - u1(i, k, j) * u2(i, k, j) - u1(i, k, j - 1) * u2(i, k, j)) / (2.d0 * dyf(j))
      end do
    end do
  end do

  call fft_xz_to_fourier(s1, cs1, 0, Nyp + 1)

  do j = jstart, jend
    do k = 0, twoNkz
      do i = 0, Nxp - 1
        cf1(i, k, j) = cf1(i, k, j) - cs1(i, k, j)
      end do
    end do
  end do

  ! Add the vertical derivative term explicitly
  do j = jstart, jend
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        s1(i, k, j) = &
          (u3(i, k, j + 1) * u2(i, k, j + 1) + u3(i, k, j) * u2(i, k, j + 1) &
           - u3(i, k, j) * u2(i, k, j) - u3(i, k, j - 1) * u2(i, k, j)) / (2.d0 * dyf(j))
      end do
    end do
  end do

  call fft_xz_to_fourier(s1, cs1, 0, Nyp + 1)

  do j = jstart, jend
    do k = 0, twoNkz
      do i = 0, Nxp - 1
        cf3(i, k, j) = cf3(i, k, j) - cs1(i, k, j)
      end do
    end do
  end do

  ! Add the vertical derivative term explicitly
  do j = 2, Nyp
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        s1(i, k, j) = &
          (0.25d0 * (u2(i, k, j) + u2(i, k, j + 1))**2.d0 &
           - 0.25d0 * (u2(i, k, j) + u2(i, k, j - 1))**2.d0) / dy(j)
      end do
    end do
  end do

  call fft_xz_to_fourier(s1, cs1, 0, Nyp + 1)

  do j = 2, Nyp
    do k = 0, twoNkz
      do i = 0, Nxp - 1
        cf2(i, k, j) = cf2(i, k, j) - cs1(i, k, j)
      end do
    end do
  end do

  ! -- At this point, we are done computing the nonlinear terms --

  ! Optionally, add user forcing to the right hand side
  ! Here, we have U1, U2, U3, and TH in physical space
  call user_rhs_chan_physical

  ! Finally, Add CFi to CRi
  do j = jstart, jend
    do k = 0, twoNkz
      do i = 0, Nxp - 1
        cr1(i, k, j) = cr1(i, k, j) + temp5 * cf1(i, k, j)
        cr3(i, k, j) = cr3(i, k, j) + temp5 * cf3(i, k, j)
      end do
    end do
  end do
  do j = 2, Nyp
    do k = 0, twoNkz
      do i = 0, Nxp - 1
        cr2(i, k, j) = cr2(i, k, j) + temp5 * cf2(i, k, j)
      end do
    end do
  end do

  ! Convert RHS terms to physical space
  call fft_xz_to_physical(cr1, r1, 0, Nyp + 1)
  call fft_xz_to_physical(cr2, r2, 2, Nyp)
  call fft_xz_to_physical(cr3, r3, 0, Nyp + 1)

  ! Compute the vertical viscous term in physical space and add to RHS
  ! This is the explicit part of the Crank-Nicolson term
  ! NU_V_SCALE is a coefficient defined as the ratio of the vertical
  ! to horizontal viscosity and can be used to add anisotropic viscosity
  do j = jstart, jend
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        r1(i, k, j) = r1(i, k, j) + temp1 * nu_v_scale * &
                      (((u1(i, k, j + 1) - u1(i, k, j)) / dy(j + 1) &
                        - (u1(i, k, j) - u1(i, k, j - 1)) / dy(j)) / dyf(j))
        r3(i, k, j) = r3(i, k, j) + temp1 * nu_v_scale * &
                      (((u3(i, k, j + 1) - u3(i, k, j)) / dy(j + 1) &
                        - (u3(i, k, j) - u3(i, k, j - 1)) / dy(j)) / dyf(j))
      end do
    end do
  end do
  do j = 2, Nyp
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        r2(i, k, j) = r2(i, k, j) + temp1 * nu_v_scale * &
                      (((u2(i, k, j + 1) - u2(i, k, j)) / dyf(j) &
                        - (u2(i, k, j) - u2(i, k, j - 1)) / dyf(j - 1)) / dy(j))
      end do
    end do
  end do

  ! If we are using a subgrid model, add the eddy viscosity term
  ! This is an added viscosity that will be treated just like the
  ! molecular viscosity with Crank-Nicolson for the vertical derivatives
  if (LES) then
    ! Note, nu_t is defined at GY points
    do j = jstart, jend
      do k = 0, Nzp - 1
        do i = 0, Nxm1
          r1(i, k, j) = r1(i, k, j) + temp2 * &
                        ((nu_t(i, k, j + 1) * (u1(i, k, j + 1) - u1(i, k, j)) / dy(j + 1) &
                          - nu_t(i, k, j) * (u1(i, k, j) - u1(i, k, j - 1)) / dy(j)) &
                         / dyf(j))
          r3(i, k, j) = r3(i, k, j) + temp2 * &
                        ((nu_t(i, k, j + 1) * (u3(i, k, j + 1) - u3(i, k, j)) / dy(j + 1) &
                          - nu_t(i, k, j) * (u3(i, k, j) - u3(i, k, j - 1)) / dy(j)) &
                         / dyf(j))
        end do
      end do
    end do
    ! Here, interpolate nu_t to GYF points
    do j = 2, Nyp
      do k = 0, Nzp - 1
        do i = 0, Nxm1
          r2(i, k, j) = r2(i, k, j) + temp2 * &
                        ((0.5d0 * (nu_t(i, k, j) + nu_t(i, k, j + 1)) * (u2(i, k, j + 1) - u2(i, k, j)) &
                          / dyf(j) &
                          - 0.5d0 * (nu_t(i, k, j) + nu_t(i, k, j - 1)) * (u2(i, k, j) - u2(i, k, j - 1)) &
                          / dyf(j - 1)) / dy(j))
        end do
      end do
    end do
  end if

  ! -- Here, we are done with computation of Velocity RHS, explicit terms --

  ! Now, build the explicit RHS terms for the passive scalar(s)

  do n = 1, N_th
    ! Do for each scalar:

    ! Compute the nonlinear terms that are present in the explicit term A
    ! U1*TH
    do j = jstart_th(n), jend_th(n)
      do k = 0, Nzp - 1
        do i = 0, Nxm1
          s1(i, k, j) = th(i, k, j, n) * u1(i, k, j)
        end do
      end do
    end do
    call fft_xz_to_fourier(s1, cs1, 0, Nyp + 1)
    do j = jstart_th(n), jend_th(n)
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          cfth(i, k, j, n) = cfth(i, k, j, n) - cikx(i) * cs1(i, k, j)
        end do
      end do
    end do
    ! U3*TH
    do j = jstart_th(n), jend_th(n)
      do k = 0, Nzp - 1
        do i = 0, Nxm1
          s1(i, k, j) = th(i, k, j, n) * u3(i, k, j)
        end do
      end do
    end do
    call fft_xz_to_fourier(s1, cs1, 0, Nyp + 1)
    do j = jstart_th(n), jend_th(n)
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          cfth(i, k, j, n) = cfth(i, k, j, n) - cikz(k) * cs1(i, k, j)
        end do
      end do
    end do

    ! We are done with the horizontal derivatives of the nonlinear terms
    ! Add the vertical derivative term explicitly
    do j = jstart_th(n), jend_th(n)
      do k = 0, Nzp - 1
        do i = 0, Nxm1
          s1(i, k, j) = &
            (th(i, k, j + 1, n) * u2(i, k, j + 1) + th(i, k, j, n) * u2(i, k, j + 1) &
             - th(i, k, j, n) * u2(i, k, j) - th(i, k, j - 1, n) * u2(i, k, j)) / (2.d0 * dyf(j))
        end do
      end do
    end do
    call fft_xz_to_fourier(s1, cs1, 0, Nyp + 1)
    do j = jstart_th(n), jend_th(n)
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          cfth(i, k, j, n) = cfth(i, k, j, n) - cs1(i, k, j)
        end do
      end do
    end do

    ! Add CFTH to the RHS vector CRTH
    do j = jstart_th(n), jend_th(n)
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          crth(i, k, j, n) = crth(i, k, j, n) + temp5 * cfth(i, k, j, n)
        end do
      end do
    end do
    ! Done with computation of RHS, explicit terms for the THETA equation
    ! Transform back to physical space

    cs1(:, :, :) = crth(:, :, :, n)
    call fft_xz_to_physical(cs1, s1, 0, Nyp + 1)
    rth(:, :, :, n) = s1(:, :, :)

    ! Compute the Explicit part of the Crank-Nicolson terms for the TH equation
    ! First, the vertical derivative viscous term
    do j = jstart_th(n), jend_th(n)
      do k = 0, Nzp - 1
        do i = 0, Nxm1
          rth(i, k, j, n) = rth(i, k, j, n) + (temp1 / Pr(n)) * nu_v_scale * ( &
                            ((th(i, k, j + 1, n) - th(i, k, j, n)) / dy(j + 1) &
                             - (th(i, k, j, n) - th(i, k, j - 1, n)) / dy(j)) / dyf(j))
        end do
      end do
    end do
    ! If we are using a subgrid model (LES) then add the eddy diffusivity here
    ! Note, KAPPA_T is defined at GY points
    if (LES) then
      do j = jstart_th(n), jend_th(n)
        do k = 0, Nzp - 1
          do i = 0, Nxm1
            rth(i, k, j, n) = rth(i, k, j, n) + temp2 * ( &
                              (kappa_t(i, k, j + 1, n) * (th(i, k, j + 1, n) - th(i, k, j, n)) / dy(j + 1) &
                               - kappa_t(i, k, j, n) * (th(i, k, j, n) - th(i, k, j - 1, n)) / dy(j)) / dyf(j))
          end do
        end do
      end do
    end if

    ! -- Now, timestep the passive scalar equation --
    ! We solve the the passive scalar before the velocity so that
    ! it is advected with the velocity from the previous R-K step
    ! which we have already made divergence free

    ! Solve the implicit equation for THETA
    ! Note that the system size is Nyp+1, but only 1..Nyp are used

    ! Initialize the matrix used to store implicit coefficients
    do j = 0, Nyp + 1
      do i = 0, Nxm1
        matl(i, j) = 0.
        matd(i, j) = 1.
        matu(i, j) = 0.
        vec(i, j) = 0.
      end do
    end do

    ! Build implicit matrix
    ! Use quasi-second order interpolation for TH on GY points
    do k = 0, Nzp - 1
      do j = jstart_th(n), jend_th(n)
        do i = 0, Nxm1
          matl(i, j) = -(temp1 / Pr(n) * nu_v_scale) / (dy(j) * dyf(j))
          matd(i, j) = 1.+(temp1 / Pr(n) * nu_v_scale) / (dy(j + 1) * dyf(j)) &
                       + (temp1 / Pr(n) * nu_v_scale) / (dy(j) * dyf(j))
          matu(i, j) = -(temp1 / Pr(n) * nu_v_scale) / (dy(j + 1) * dyf(j))
          ! Define RHS vector
          vec(i, j) = rth(i, k, j, n)
        end do
      end do
      ! IF using a subgrid model (LES) then add the eddy diffusivity part implicitly
      if (LES) then
        do j = jstart_th(n), jend_th(n)
          do i = 0, Nxm1
            matl(i, j) = matl(i, j) - temp2 * kappa_t(i, k, j, n) &
                         / (dy(j) * dyf(j))
            matd(i, j) = matd(i, j) + temp2 * kappa_t(i, k, j + 1, n) &
                         / (dy(j + 1) * dyf(j)) &
                         + temp2 * kappa_t(i, k, j, n) &
                         / (dy(j) * dyf(j))
            matu(i, j) = matu(i, j) - temp2 * kappa_t(i, k, j + 1, n) &
                         / (dy(j + 1) * dyf(j))
          end do
        end do
      end if

      ! If we are using MPI, then solve the implicit system in separate forward
      ! and backward sweeps for efficiency
      if (use_mpi) then

        call apply_BC_th_mpi(matl, matd, matu, vec, n)
        ! If we are using MPI, split the implicit solve into foward and
        ! backward sweeps for efficiency
        call thomas_forward_real_mpi(matl, matd, matu, vec, Nyp, Nx)
        call thomas_backward_real_mpi(matl, matd, matu, vec, Nyp, Nx)
      else
        ! Else we are running in serial mode
        call apply_BC_th_lower(matl, matd, matu, vec, n)
        call apply_BC_th_upper(matl, matd, matu, vec, n)
        call thomas_real(matl, matd, matu, vec, Nyp + 1, Nxm1)
        ! Apply the boundary conditions to our linear system
      end if

      do j = jstart_th(n) - 1, jend_th(n) + 1
        do i = 0, Nxm1
          th(i, k, j, n) = vec(i, j)
        end do
      end do

      ! END do k
    end do

    ! End do number of passive scalars
  end do

  ! Initialize the matrix to zeros to be used for implicit solves
  ! Note that the system size is Nyp+1, but only 1..Nyp are used

  ! Initialize the matrix used to store implicit coefficients
  do j = 0, Nyp + 1
    do i = 0, Nxm1
      matl(i, j) = 0.
      matd(i, j) = 1.
      matu(i, j) = 0.
      vec(i, j) = 0.
    end do
  end do

  ! Build implicit matrix for U2
  do k = 0, Nzp - 1
    do j = 2, Nyp
      do i = 0, Nxm1
        matl(i, j) = -temp1 * nu_v_scale / (dyf(j - 1) * dy(j))
        matd(i, j) = 1.+temp1 * nu_v_scale / (dyf(j) * dy(j)) &
                     + temp1 * nu_v_scale / (dyf(j - 1) * dy(j))
        matu(i, j) = -temp1 * nu_v_scale / (dyf(j) * dy(j))
        vec(i, j) = r2(i, k, j)
      end do
    end do
    if (LES) then
      ! IF using a subgrid model (LES) then add the eddy viscosity part implicitly
      do j = 2, Nyp
        do i = 0, Nxm1
          matl(i, j) = matl(i, j) &
                       - temp2 * 0.5d0 * (nu_t(i, k, j) + nu_t(i, k, j - 1)) / (dyf(j - 1) * dy(j))
          matd(i, j) = matd(i, j) &
                       + temp2 * 0.5d0 * (nu_t(i, k, j) + nu_t(i, k, j + 1)) / (dyf(j) * dy(j)) &
                       + temp2 * 0.5d0 * (nu_t(i, k, j) + nu_t(i, k, j - 1)) / (dyf(j - 1) * dy(j))
          matu(i, j) = matu(i, j) &
                       - temp2 * 0.5d0 * (nu_t(i, k, j) + nu_t(i, k, j + 1)) / (dyf(j) * dy(j))
        end do
      end do
    end if

    if (use_mpi) then

      ! First, apply the boundary conditions
      call apply_BC_u2_mpi(matl, matd, matu, vec)
      ! If we are using MPI, split the implicit solve into forward and
      ! backward sweeps for efficiency
      call thomas_forward_real_mpi(matl, matd, matu, vec, Nyp, Nx)
      call thomas_backward_real_mpi(matl, matd, matu, vec, Nyp, Nx)
    else
      ! Else, we are running in serial mode
      ! Set the boundary conditions for U2
      call apply_BC_2_lower(matl, matd, matu, vec)
      call apply_BC_2_upper(matl, matd, matu, vec)

      ! Now, solve the tridiagonal system for U2(i,:,k)
      call thomas_real(matl, matd, matu, vec, Nyp + 1, Nxm1)
    end if

    do j = 1, Nyp + 1
      do i = 0, Nxm1
        u2(i, k, j) = vec(i, j)
      end do
    end do
    ! End do k
  end do

  ! Solve for U1
  ! Note, here the matrix will be indexed from 1...Nyp+1 corresponding to U1(0:Nyp)

  ! Initialize the matrix used to store implicit coefficients
  do j = 0, Nyp + 1
    do i = 0, Nxm1
      matl(i, j) = 0.
      matd(i, j) = 1.
      matu(i, j) = 0.
      vec(i, j) = 0.
    end do
  end do

  ! Build the implicit system of equations for U1
  do k = 0, Nzp - 1
    do j = jstart, jend
      do i = 0, Nxm1
        matl(i, j) = -temp1 * nu_v_scale / (dy(j) * dyf(j))
        matd(i, j) = 1.-temp1 * nu_v_scale * (-1./(dy(j + 1) * dyf(j)) &
                                              - 1./(dy(j) * dyf(j)))
        matu(i, j) = -temp1 * nu_v_scale / (dy(j + 1) * dyf(j))
        vec(i, j) = r1(i, k, j)
      end do
    end do
    ! IF using a subgrid model (LES) then add the eddy viscosity part implicitly
    if (LES) then
      do j = jstart, jend
        do i = 0, Nxm1
          matl(i, j) = matl(i, j) - temp2 * nu_t(i, k, j) &
                       / (dy(j) * dyf(j))
          matd(i, j) = matd(i, j) + temp2 * nu_t(i, k, j + 1) &
                       / (dy(j + 1) * dyf(j)) &
                       + temp2 * nu_t(i, k, j) &
                       / (dy(j) * dyf(j))
          matu(i, j) = matu(i, j) - temp2 * nu_t(i, k, j + 1) &
                       / (dy(j + 1) * dyf(j))
        end do
      end do
    end if

    if (use_mpi) then
      ! First, apply the boundary conditions
      call apply_BC_u1_mpi(matl, matd, matu, vec)
      ! If we are using MPI, split the implicit solve into forward and
      ! backward sweeps for efficiency
      call thomas_forward_real_mpi(matl, matd, matu, vec, Nyp, Nx)
      call thomas_backward_real_mpi(matl, matd, matu, vec, Nyp, Nx)
    else
      ! Else, we are running in serial mode
      ! Set the boundary conditions for U1
      call apply_BC_1_lower(matl, matd, matu, vec)
      call apply_BC_1_upper(matl, matd, matu, vec)
      ! Now, solve the tridiagonal system for U1(:,k,:)
      call thomas_real(matl, matd, matu, vec, Nyp + 1, Nxm1)
    end if

    do j = jstart - 1, jend + 1
      do i = 0, Nxm1
        u1(i, k, j) = vec(i, j)
      end do
    end do

    ! End do k
  end do

  ! Initialize the matrix used to store implicit coefficients
  do j = 0, Nyp + 1
    do i = 0, Nxm1
      matl(i, j) = 0.
      matd(i, j) = 1.
      matu(i, j) = 0.
      vec(i, j) = 0.
    end do
  end do

  ! Solve for U3
  ! Note, here the matrix will be indexed from 1...Nyp+1 corresponding to U1(0:Nyp)
  ! Build the implicit system of equations for U3
  do k = 0, Nzp - 1
    do j = jstart, jend
      do i = 0, Nxm1
        matl(i, j) = -temp1 * nu_v_scale / (dy(j) * dyf(j))
        matd(i, j) = 1.-temp1 * nu_v_scale * (-1./(dy(j + 1) * dyf(j)) &
                                              - 1./(dy(j) * dyf(j)))
        matu(i, j) = -temp1 * nu_v_scale / (dy(j + 1) * dyf(j))
        vec(i, j) = r3(i, k, j)
      end do
    end do
    ! IF using a subgrid model (LES) then add the eddy viscosity part implicitly
    if (LES) then
      do j = jstart, jend
        do i = 0, Nxm1
          matl(i, j) = matl(i, j) - temp2 * nu_t(i, k, j) &
                       / (dy(j) * dyf(j))
          matd(i, j) = matd(i, j) + temp2 * nu_t(i, k, j + 1) &
                       / (dy(j + 1) * dyf(j)) &
                       + temp2 * nu_t(i, k, j) &
                       / (dy(j) * dyf(j))
          matu(i, j) = matu(i, j) - temp2 * nu_t(i, k, j + 1) &
                       / (dy(j + 1) * dyf(j))
        end do
      end do
    end if

    if (use_mpi) then
      ! First, apply the boundary conditions
      call apply_BC_u3_mpi(matl, matd, matu, vec)
      ! If we are using MPI, split the implicit solve into forward and
      ! backward sweeps for efficiency
      call thomas_forward_real_mpi(matl, matd, matu, vec, Nyp, Nx)
      call thomas_backward_real_mpi(matl, matd, matu, vec, Nyp, Nx)
    else
      ! Else, we are running in serial mode
      ! Set the boundary conditions for U3
      call apply_BC_3_lower(matl, matd, matu, vec)
      call apply_BC_3_upper(matl, matd, matu, vec)
      ! Now, solve the tridiagonal system for U3(i,:,k)
      call thomas_real(matl, matd, matu, vec, Nyp + 1, Nxm1)
    end if

    do j = jstart - 1, jend + 1
      do i = 0, Nxm1
        u3(i, k, j) = vec(i, j)
      end do
    end do
    ! End do k
  end do

  ! -- Done getting U1hat, U2hat, U3hat at new RK Step --

  ! If we are on the final RK step, optionally update the timestep
  ! based on the CFL criteria. This won't affect the current timestep
  ! since the TEMP1, etc. variables have already been set using
  ! the current timestep
  if (variable_dt .and. (rk_step == 3) &
      .and. (mod(time_step, update_dt) == 0)) then
    call courant
  end if

  ! Transform TH and U to Fourier Space
  call fft_xz_to_fourier(u1, cu1, 0, Nyp + 1)
  call fft_xz_to_fourier(u2, cu2, 0, Nyp + 1)
  call fft_xz_to_fourier(u3, cu3, 0, Nyp + 1)
  do n = 1, N_th
    s1(:, :, :) = th(:, :, :, n)
    call fft_xz_to_fourier(s1, cs1, 0, Nyp + 1)
    cth(:, :, :, n) = cs1(:, :, :)
  end do

  ! Begin second step of the Fractional Step algorithm, making u divergence free
  ! The following subroutine projects Uhat onto divergence free space

  call rem_div_chan

  ! Now, phi is stored in CR1, use this to update the pressure field
  ! Note, here we divide by H_BAR since it was absorbed into PHI in REM_DIV
  do j = jstart, jend
    do k = 0, twoNkz
      do i = 0, Nxp - 1
        cp(i, k, j) = cp(i, k, j) + cr1(i, k, j) / temp4
      end do
    end do
  end do

  ! Fix disparities at the boundary due to the thoms algorithm in parallel
  if (use_mpi) then
    call ghost_chan_mpi
  end if

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine rk_chan_2
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  ! Alternative time-stepping algorithm for the channel-flow case.
  ! This algorithm uses Crank-Nicolson for all viscous terms and
  ! third order Runge-Kutta for all nonlinear terms
  ! INPUTS  (in Fourier space):  CUi, P, and (if k>1) CFi at (k-1)  (for i=1,2,3)
  ! OUTPUTS (in Fourier space):  CUi, P, and (if k<3) CFi at (k)
  ! Each RK step, there are 11 FFT calls. 11 storage variables are used.
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|

  include 'header'

  integer i, j, k, n
  real(rkind) temp1, temp2, temp3, temp4, temp5, ubulk

  ! STOP -----------------------------------
  if (rank == 0) &
    write (*, '("RK_CHAN_2 not supported yet")')
  call mpi_finalize(ierror)
  stop
  ! ----------------------------------------

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine rem_div_chan
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|

  ! Compute varphi, store in variable CR1.
  ! Solves for phi in computational space
  ! H_BAR has been absorbed into PHI, so we are solving for H_BAR*PHI

  include 'header'
  integer i, j, k

  ! First, Initialize the matrix components
  do j = 0, Nyp + 1
    do i = 0, Nxp - 1
      matl_c(i, j) = 0.
      matd_c(i, j) = 1.
      matu_c(i, j) = 0.
      vec_c(i, j) = (0., 0.)
    end do
  end do

  ! The 2d FFT of Ui should have been taken and stored in CUi
  ! Solving for phi amounts to solving a tridiagonal system
  ! First, construct the system to be solved
  do k = 0, twoNkz
    do j = 1, Nyp
      do i = 0, Nxp - 1
        matl_c(i, j) = 1./(dy(j) * dyf(j))
        matd_c(i, j) = -kx2(i) - kz2(k) &
                       - 1./(dy(j + 1) * dyf(j)) - 1./(dy(j) * dyf(j))
        matu_c(i, j) = 1./(dy(j + 1) * dyf(j))
      end do
    end do

    ! Now, create the RHS vector
    do j = 1, Nyp
      do i = 0, Nxp - 1
        vec_c(i, j) = (cikx(i) * cu1(i, k, j) &
                       + (cu2(i, k, j + 1) - cu2(i, k, j)) / dyf(j) &
                       + cikz(k) * cu3(i, k, j))
      end do
    end do

    if (use_mpi) then
      ! If we are using the MPI package...
      call apply_BC_rem_div_mpi(matl_c, matd_c, matu_c, vec_c, k)
      ! First, do all forward sweeps
      call thomas_forward_complex_mpi(matl_c, matd_c, matu_c, vec_c &
                                      , Nyp, Nxp)
      ! Now, do the backward sweeps
      call thomas_backward_complex_mpi(matl_c, matd_c, matu_c, vec_c &
                                       , Nyp, Nxp)
    else
      ! Else we are running in serial mode
      do i = 0, Nkx
        if ((k == 0) .and. (i == 0)) then
          ! Use homogeneous dirichlet BCS for kx=kz=0 component at bottom wall
          ! Otherwise the matrix will be singular
          matl_c(i, 1) = 0.
          matd_c(i, 1) = 1.
          matu_c(i, 1) = 0.
          vec_c(i, 1) = (0., 0.)

          matl_c(i, Nyp) = 1.
          matd_c(i, Nyp) = -1.
          matu_c(i, Nyp) = 0.
          vec_c(i, Nyp) = (0., 0.)
        else
          ! Use Dirichlet boundary conditions, dp/dz=0 at walls
          matl_c(i, 1) = 0.
          matd_c(i, 1) = 1.
          matu_c(i, 1) = -1.
          vec_c(i, 1) = (0., 0.)

          matl_c(i, Nyp) = 1.
          matd_c(i, Nyp) = -1.
          matu_c(i, Nyp) = 0.
          vec_c(i, Nyp) = (0., 0.)
        end if
      end do
      ! Now solve the tridiagonal system for phi, store in CR1
      call thomas_complex(matl_c, matd_c, matu_c, vec_c, Nyp, Nkx)
    end if

    do j = 1, Nyp
      do i = 0, Nxp - 1
        cr1(i, k, j) = vec_c(i, j)
      end do
    end do

  end do

  ! Now, Solve for CUi, the divergenceless velocity field
  do j = 1, Nyp
    do k = 0, twoNkz
      do i = 0, Nxp - 1
        cu1(i, k, j) = cu1(i, k, j) - cikx(i) * cr1(i, k, j)
        cu3(i, k, j) = cu3(i, k, j) - cikz(k) * cr1(i, k, j)
      end do
    end do
  end do
  do j = 2, Nyp
    do k = 0, twoNkz
      do i = 0, Nxp - 1
        cu2(i, k, j) = cu2(i, k, j) - (cr1(i, k, j) &
                                       - cr1(i, k, j - 1)) / dy(j)
      end do
    end do
  end do

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine poisson_p_chan
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  ! We have CUi, need to compute CP.  Solve tridiagonal system exactly

  include 'header'

  integer i, j, k, n

  if (flavor == 'Basic') then
    if (rank == 0) &
      write (*, '("Computing cp from cui divergence")')

  end if

  ! First, construct the RHS vector, (dui/dxj)(duj/dxi)
  do j = 2, Nyp
    do k = 0, twoNkz
      do i = 0, Nxp - 1 ! Nkx
        cf1(i, k, j) = cikx(i) * cu1(i, k, j)
        cf2(i, k, j) = (cu2(i, k, j + 1) - cu2(i, k, j)) / dyf(j)
        cf3(i, k, j) = cikz(k) * cu3(i, k, j)
      end do
    end do
  end do

  call fft_xz_to_physical(cf1, f1, 0, Nyp + 1)
  call fft_xz_to_physical(cf2, f2, 0, Nyp + 1)
  call fft_xz_to_physical(cf3, f3, 0, Nyp + 1)

  do j = 2, Nyp
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        f1(i, k, j) = f1(i, k, j)**2.
        f2(i, k, j) = f2(i, k, j)**2.
        f3(i, k, j) = f3(i, k, j)**2.
      end do
    end do
  end do

  call fft_xz_to_fourier(f1, cf1, 0, Nyp + 1)
  call fft_xz_to_fourier(f2, cf2, 0, Nyp + 1)
  call fft_xz_to_fourier(f3, cf3, 0, Nyp + 1)

  ! Now we have the diagonal terms, add to the rhs term
  do j = 2, Nyp
    do k = 0, twoNkz
      do i = 0, Nxp - 1 ! Nkx
        cs1(i, k, j) = cf1(i, k, j) + cf2(i, k, j) + cf3(i, k, j)
      end do
    end do
  end do

  ! Now get the first of the off-diagonal terms
  do j = 2, Nyp
    do k = 0, twoNkz
      do i = 0, Nxp - 1 ! Nkx
        cf1(i, k, j) = (cu1(i, k, j + 1) - cu1(i, k, j - 1)) / (2.*dyf(j))
        cf2(i, k, j) = cikx(i) * 0.5 * (cu2(i, k, j) + cu2(i, k, j + 1))
      end do
    end do
  end do

  call fft_xz_to_physical(cf1, f1, 0, Nyp + 1)
  call fft_xz_to_physical(cf2, f2, 0, Nyp + 1)

  ! Compute product
  do j = 2, Nyp
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        f1(i, k, j) = 2.*f1(i, k, j) * f2(i, k, j)
      end do
    end do
  end do

  call fft_xz_to_fourier(f1, cf1, 0, Nyp + 1)

  ! Add to RHS term
  do j = 2, Nyp
    do k = 0, twoNkz
      do i = 0, Nxp - 1 ! Nkx
        cs1(i, k, j) = cs1(i, k, j) + cf1(i, k, j)
      end do
    end do
  end do

  ! Now get the second of the off-diagonal terms
  do j = 2, Nyp
    do k = 0, twoNkz
      do i = 0, Nxp - 1 ! Nkx
        cf1(i, k, j) = (cu3(i, k, j + 1) - cu3(i, k, j - 1)) / (2.*dyf(j))
        cf2(i, k, j) = cikz(k) * 0.5 * (cu2(i, k, j) + cu2(i, k, j + 1))
      end do
    end do
  end do

  ! Convert to Physical space
  call fft_xz_to_physical(cf1, f1, 0, Nyp + 1)
  call fft_xz_to_physical(cf2, f2, 0, Nyp + 1)

  ! Compute product
  do j = 2, Nyp
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        f1(i, k, j) = 2.*f1(i, k, j) * f2(i, k, j)
      end do
    end do
  end do

  call fft_xz_to_fourier(f1, cf1, 0, Nyp + 1)

  ! Add to RHS term
  do j = 2, Nyp
    do k = 0, twoNkz
      do i = 0, Nxp - 1 ! Nkx
        cs1(i, k, j) = cs1(i, k, j) + cf1(i, k, j)
      end do
    end do
  end do

  ! Now get the third of the off-diagonal terms
  do j = 2, Nyp
    do k = 0, twoNkz
      do i = 0, Nxp - 1 ! Nkx
        cf1(i, k, j) = cikz(k) * cu1(i, k, j)
        cf2(i, k, j) = cikx(i) * cu3(i, k, j)
      end do
    end do
  end do

  call fft_xz_to_physical(cf1, f1, 0, Nyp + 1)
  call fft_xz_to_physical(cf2, f2, 0, Nyp + 1)

  ! Compute product
  do j = 2, Nyp
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        f1(i, k, j) = 2.*f1(i, k, j) * f2(i, k, j)
      end do
    end do
  end do

  call fft_xz_to_fourier(f1, cf1, 0, Nyp + 1)

  ! Add to RHS term
  do j = 2, Nyp
    do k = 0, twoNkz
      do i = 0, Nxp - 1 ! Nkx
        cs1(i, k, j) = cs1(i, k, j) + cf1(i, k, j)
      end do
    end do
  end do

  ! Finally, if the buoyancy force is active, then we need to add
  ! the contribution of the density to the pressure.  Note that the
  ! plane averaged density and the corresponding hydrostatic part of the
  ! pressure have been cancelled, so skip the 0,0 mode
  do n = 1, N_th
    do j = 2, Nyp
      do k = 0, twoNkz
        do i = 0, Nxp - 1 ! Nkx
          if ((rankZ /= 0) .or. (i /= 0) .or. (k /= 0)) then
            cs1(i, k, j) = cs1(i, k, j) + &
                           (cth(i, k, j + 1, n) - cth(i, k, j - 1, n)) / (gyf(j + 1) - gyf(j - 1))
          end if
        end do
      end do
    end do
  end do

  ! Now, the RHS term should be stored in cs1

  ! Construct the tridiagonal system in Fourier space to solve for CP
  ! First, zero the vectors
  do j = 0, Nyp + 1
    do i = 0, Nxp - 1 ! Nkx
      matl_c(i, j) = 0.d0
      matd_c(i, j) = 1.d0
      matu_c(i, j) = 0.d0
      vec_c(i, j) = (0., 0.)
    end do
  end do

  do k = 0, twoNkz
    do j = 2, Nyp
      do i = 0, Nxp - 1 ! Nkx
        matl_c(i, j) = 1./(dy(j) * dyf(j))
        matd_c(i, j) = -kx2(i) - kz2(k) - 1./(dy(j + 1) * dyf(j)) &
                       - 1./(dy(j) * dyf(j))
        matu_c(i, j) = 1./(dy(j + 1) * dyf(j))
        vec_c(i, j) = -1.*cs1(i, k, j)
      end do
    end do

    if (use_mpi) then
      call apply_BC_poisson_mpi(matl_c, matd_c, matu_c, vec_c, k)
      ! First, do the forward sweeps
      call thomas_forward_complex_mpi(matl_c, matd_c, matu_c, vec_c &
                                      , Nyp, Nxp)
      ! Now, do the backwared sweeps to put the solution in VEC_C
      call thomas_backward_complex_mpi(matl_c, matd_c, matu_c, vec_c &
                                       , Nyp, Nxp)
    else
      ! Else we are running in serial mode
      ! Apply BCs
      do i = 0, Nkx
        ! Use dirichlet boundary condition at the lower wall to
        ! prevent the tridiagonal matrix from becomming singular for i,k=0
        if ((i == 0) .and. (k == 0)) then
          matd_c(i, 1) = 1.
          matu_c(i, 1) = 0.
          vec_c(i, 1) = (0., 0.)
          matd_c(i, Nyp) = -1.
          matl_c(i, Nyp) = 1.
          vec_c(i, Nyp) = (0., 0.)
        else
          ! Here, apply Neumann boundary conditions (dp/dz=0) at the walls
          matd_c(i, 1) = 1.
          matu_c(i, 1) = -1.
          vec_c(i, 1) = (0., 0.)
          matd_c(i, Nyp) = -1.
          matl_c(i, Nyp) = 1.
          vec_c(i, Nyp) = (0., 0.)
        end if
      end do
      ! Now, solve for CP
      call thomas_complex(matl_c, matd_c, matu_c, vec_c, Nyp, Nkx)
    end if

    do j = 1, Nyp
      do i = 0, Nxp - 1 ! Nkx
        cp(i, k, j) = vec_c(i, j)
      end do
    end do
  end do

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine input_chan
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  include 'header'
  real version, current_version
  integer i, j, k, n
  real(rkind) ro

  ! Read in input parameters specific for channel flow case
  open (11, file='input_chan.dat', form='formatted', status='old')
  ! Read input file.

  current_version = 3.2
  read (11, *)
  read (11, *)
  read (11, *)
  read (11, *)
  read (11, *) version
  if (version /= current_version) &
    stop 'Wrong input data format input_chan'
  read (11, *)
  read (11, *) time_ad_meth
  read (11, *)
  read (11, *) les_model_type
  read (11, *)
  read (11, *) IC_type, kick, physical_noise
  read (11, *)
  read (11, *) ro
  Ro_inv = 1.d0 / ro
  read (11, *)
  read (11, *) delta
  read (11, *)
  read (11, *) grav_x, grav_z, grav_y
  read (11, *)
  read (11, *) f_type, ubulk0, px0, omega0, amp_omega0
  read (11, *)
  read (11, *)
  read (11, *) u_BC_Ymin, u_BC_Ymin_c1
  read (11, *)
  read (11, *) w_BC_Ymin, w_BC_Ymin_c1
  read (11, *)
  read (11, *) v_BC_Ymin, v_BC_Ymin_c1
  read (11, *)
  read (11, *) u_BC_Ymax, u_BC_Ymax_c1
  read (11, *)
  read (11, *) w_BC_Ymax, w_BC_Ymax_c1
  read (11, *)
  read (11, *) v_BC_Ymax, v_BC_Ymax_c1
  read (11, *)
  ! Read in boundary conditions and background gradients for the N_th scalars
  do n = 1, N_th
    read (11, *)
    read (11, *) th_BC_Ymin(n), th_BC_Ymin_c1(n)
    read (11, *)
    read (11, *) th_BC_Ymax(n), th_BC_Ymax_c1(n)
  end do

  if (rank == 0) write (*, '("Ro Inverse = " ES26.18)') Ro_inv

  ! Compensate no-slip BC in the GS flow direction due to dTHdx
  !   AND also define dTHdx & dTHdz
  if (IC_Type == 4) then ! Infinite Sloping Front
    if (w_BC_Ymin == 1) then
      w_BC_Ymin_c1 = w_BC_Ymin_c1 - 1.d0
    end if
    if (w_BC_Ymax == 1) then
      w_BC_Ymax_c1 = w_BC_Ymax_c1 - 1.d0
    end if
    dTHdX(1) = Ro_inv / delta
    dTHdZ(1) = 0.d0

  else if (IC_Type == 5) then ! Finite Sloping Front
    if (w_BC_Ymin == 1) then
      w_BC_Ymin_c1 = w_BC_Ymin_c1 - 2.d0 * delta / Lx
    end if
    if (w_BC_Ymax == 1) then
      w_BC_Ymax_c1 = w_BC_Ymax_c1 - 2.d0 * delta / Lx
    end if
    dTHdX(1) = 2.d0 / Lx * Ro_inv
    dTHdZ(1) = 0.d0

  end if

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine init_chan_movie
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  include 'header'
  character(len=55) fname
  integer i, j, k

  ! Get the indices
  Nxmovie = int(XcMovie * Nx / Lx)
  NyMovie = nint(YcMovie * (((Nyp - 1) * Nprocs) / Ly))
  NzMovie = int(ZcMovie * Nz / Lz)

  rankzMovie = int(NzMovie / Nzp)
  NzMovie = NzMovie - rankzMovie * Nzp

  rankyMovie = -1
  if (gyf(jstart) <= YcMovie .and. gyf(jend + 1) > YcMovie) then
    rankyMovie = rankY
    i = 1
    do while (.not. &
              (gyf(i) <= YcMovie .and. gyf(i + 1) > YcMovie))
      i = i + 1
    end do
    NyMovie = i;
  end if

  if (rankY == rankyMovie .and. rankZ == rankzMovie) then
    write (*, '("Movie Parameters")')
    write (*, '("Rank = " I10)') rank
    write (*, '("  Xc = " ES26.18)') gx(Nxmovie)
    write (*, '("  Yc = " ES26.18)') gz(rankzMovie * Nzp + NzMovie)
    write (*, '("  Zc = " ES26.18)') gyf(NyMovie)
  end if

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine create_grid_chan
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  include 'header'
  character(len=55) fname
  integer i, j, k

  if (rank == 0) &
    write (*, '("Fourier in X")')
  do i = 0, Nx
    gx(i) = (i * Lx) / Nx
    dx(i) = Lx / Nx
    if (verbosity > 3 .and. rank == 0) &
      write (*, *) 'GX(', i, ') = ', gx(i)
  end do
  if (rank == 0) &
    write (*, '("Fourier in Y")')
  do k = 0, Nz
    gz(k) = (k * Lz) / Nz
    dz(k) = Lz / Nz
    if (rank == 0 .and. verbosity > 3) &
      write (*, *) 'GY(', k, ') = ', gz(k)
  end do
  if (rank == 0) &
    write (*, '("Finite-difference in Z")')

  if (rank == 0) &
    write (*, '("Use MPI: " L1)') use_mpi

  if (rank == 0) &
    write (*, '("Use LES: " L1)') LES

  fname = 'grid.h5'
  if (use_mpi) then
    call mpi_barrier(mpi_comm_world, ierror)
  end if
  call ReadGridHDF5(fname, 2)


  ! Define grid spacing
  do j = 1, Nyp + 1
    dy(j) = (gyf(j) - gyf(j - 1))
  end do
  do j = 1, Nyp
    dyf(j) = (gy(j + 1) - gy(j))
  end do
  ! dyf(Nyp + 1) = dyf(Nyp)

  ! Communicate  dyf(0)  and  dyf(Nyp + 1)
  if (rankY == 0) then
    dyf(0) = dyf(1)
  else
    call mpi_send(dyf(2), 1, mpi_double_precision, rankY - 1, &
                  100 + rankY, mpi_comm_y, ierror)
    call mpi_recv(dyf(0), 1, mpi_double_precision, rankY - 1, &
                  110 + rankY - 1, mpi_comm_y, status, ierror)
  end if
  if (rankY == NprocY - 1) then
    dyf(Nyp + 1) = dyf(Nyp)
  else
    call mpi_send(dyf(Nyp - 1), 1, mpi_double_precision, rankY + 1, &
                  110 + rankY, mpi_comm_y, ierror)
    call mpi_recv(dyf(Nyp + 1), 1, mpi_double_precision, rankY + 1, &
                  100 + rankY + 1, mpi_comm_y, status, ierror)
  end if

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|------
subroutine apply_BC_1_lower(matl, matd, matu, vec)
  !----*|--.---------.---------.---------.---------.---------.---------.-|-----
  include 'header'
  integer i

  ! Bottom Wall:
  if (u_BC_Ymin == 0) then
    ! Dirichlet
    do i = 0, Nxm1
      matl(i, 0) = 0.
      matd(i, 0) = 1.
      matu(i, 0) = 0.
      vec(i, 0) = 0.

      matl(i, 1) = 0.
      matd(i, 1) = 1.
      matu(i, 1) = 0.
      vec(i, 1) = u_BC_Ymin_c1
    end do
  else
    ! Neumann
    do i = 0, Nxm1
      matl(i, 0) = 0.
      matd(i, 0) = 1.
      matu(i, 0) = 0.
      vec(i, 0) = 0.
    end do
    do i = 0, Nxm1
      matl(i, 1) = 0.
      matd(i, 1) = -1.
      matu(i, 1) = 1.
      vec(i, 1) = dy(2) * u_BC_Ymin_c1
    end do

  end if

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|------
subroutine apply_BC_1_lower_c(matl_c, matd_c, matu_c, vec_c)
  !----*|--.---------.---------.---------.---------.---------.---------.-|-----
  include 'header'
  integer i

  ! Bottom Wall:
  if (u_BC_Ymin == 0) then
    ! Dirichlet
    do i = 0, Nxp - 1 ! Nkx
      matl_c(i, 0) = 0.
      matd_c(i, 0) = 1.
      matu_c(i, 0) = 0.
      vec_c(i, 0) = 0.

      matl_c(i, 1) = 0.
      matd_c(i, 1) = 1.
      matu_c(i, 1) = 0.
      vec_c(i, 1) = u_BC_Ymin_c1
    end do
  else
    ! Neumann
    do i = 0, Nxp - 1 ! Nkx
      matl_c(i, 0) = 0.
      matd_c(i, 0) = 1.
      matu_c(i, 0) = 0.
      vec_c(i, 0) = 0.
    end do
    do i = 0, Nxp - 1 ! Nkx
      matl_c(i, 1) = 0.
      matd_c(i, 1) = -1.
      matu_c(i, 1) = 1.
      vec_c(i, 1) = dy(2) * u_BC_Ymin_c1
    end do

  end if

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|----
subroutine apply_BC_1_upper(matl, matd, matu, vec)
  !----*|--.---------.---------.---------.---------.---------.---------.-|--
  include 'header'
  integer i

  ! Top wall
  if (u_BC_Ymax == 0) then
    ! Dirichlet
    do i = 0, Nxm1
      matl(i, Nyp + 1) = 0.
      matd(i, Nyp + 1) = 1.
      matu(i, Nyp + 1) = 0.
      vec(i, Nyp + 1) = 0.

      matl(i, Nyp) = 0.
      matd(i, Nyp) = 1.
      matu(i, Nyp) = 0.
      vec(i, Nyp) = u_BC_Ymax_c1
    end do
  else
    ! Neumann
    do i = 0, Nxm1
      matl(i, Nyp) = -1.
      matd(i, Nyp) = 1.
      matu(i, Nyp) = 0.
      vec(i, Nyp) = dy(Nyp) * u_BC_Ymax_c1
    end do
    do i = 0, Nxm1
      matl(i, Nyp + 1) = 0.
      matd(i, Nyp + 1) = 1.
      matu(i, Nyp + 1) = 0.
      vec(i, Nyp + 1) = 0.
    end do

  end if

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|----
subroutine apply_BC_1_upper_c(matl_c, matd_c, matu_c, vec_c)
  !----*|--.---------.---------.---------.---------.---------.---------.-|--
  include 'header'
  integer i

  ! Top wall
  if (u_BC_Ymax == 0) then
    ! Dirichlet
    do i = 0, Nxp - 1 ! Nkx
      matl_c(i, Nyp + 1) = 0.
      matd_c(i, Nyp + 1) = 1.
      matu_c(i, Nyp + 1) = 0.
      vec_c(i, Nyp + 1) = 0.

      matl_c(i, Nyp) = 0.
      matd_c(i, Nyp) = 1.
      matu_c(i, Nyp) = 0.
      vec_c(i, Nyp) = u_BC_Ymax_c1
    end do
  else
    ! Neumann
    do i = 0, Nxp - 1 ! Nkx
      matl_c(i, Nyp + 1) = 0.
      matd_c(i, Nyp + 1) = 1.
      matu_c(i, Nyp + 1) = 0.
      vec_c(i, Nyp + 1) = 0.
    end do
    do i = 0, Nxp - 1 ! Nkx
      matl_c(i, Nyp) = -1.
      matd_c(i, Nyp) = 1.
      matu_c(i, Nyp) = 0.
      vec_c(i, Nyp) = dy(Nyp) * u_BC_Ymax_c1
    end do

  end if

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|---
subroutine apply_BC_2_lower(matl, matd, matu, vec)
  !----*|--.---------.---------.---------.---------.---------.---------.-|--
  include 'header'
  integer i

  ! Bottom Wall:
  if (v_BC_Ymin == 0) then
    ! Dirichlet
    do i = 0, Nxm1
      matl(i, 1) = 0.d0
      matd(i, 1) = 1.d0
      matu(i, 1) = 0.d0
      vec(i, 1) = v_BC_Ymin_c1

      matl(i, 2) = 0.d0
      matd(i, 2) = 1.d0
      matu(i, 2) = 0.d0
      vec(i, 2) = v_BC_Ymin_c1
    end do
  else if (v_BC_Ymin == 1) then
    ! Neumann
    do i = 0, Nxm1
      matd(i, 1) = -1.d0
      matu(i, 1) = 1.d0
      matl(i, 1) = 0.d0
      vec(i, 1) = dyf(1) * v_BC_Ymin_c1
    end do
  end if

  ! The following is only a placeholder, this row is used for U1 and U3
  do i = 0, Nxm1
    matl(i, 0) = 0.
    matd(i, 0) = 1.
    matu(i, 0) = 0.
    vec(i, 0) = 0.
  end do

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|---
subroutine apply_BC_2_lower_c(matl_c, matd_c, matu_c, vec_c)
  !----*|--.---------.---------.---------.---------.---------.---------.-|--
  include 'header'
  integer i

  ! Bottom Wall:
  if (v_BC_Ymin == 0) then
    ! Dirichlet
    do i = 0, Nxp - 1 ! Nkx
      matl_c(i, 1) = 0.d0
      matd_c(i, 1) = 1.d0
      matu_c(i, 1) = 0.d0
      vec_c(i, 1) = v_BC_Ymin_c1

      matl_c(i, 2) = 0.d0
      matd_c(i, 2) = 1.d0
      matu_c(i, 2) = 0.d0
      vec_c(i, 2) = v_BC_Ymin_c1
    end do
  else if (v_BC_Ymin == 1) then
    ! Neumann
    do i = 0, Nxp - 1 ! Nkx
      matd_c(i, 1) = -1.d0
      matu_c(i, 1) = 1.d0
      matl_c(i, 1) = 0.d0
      vec_c(i, 1) = dyf(1) * v_BC_Ymin_c1
    end do
  end if

  ! The following is only a placeholder, this row is used for U1 and U3
  do i = 0, Nxp - 1 ! Nkx
    matl_c(i, 0) = 0.
    matd_c(i, 0) = 1.
    matu_c(i, 0) = 0.
    vec_c(i, 0) = 0.
  end do

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|--
subroutine apply_BC_2_upper(matl, matd, matu, vec)
  !----*|--.---------.---------.---------.---------.---------.---------.-|--
  include 'header'
  integer i
  ! Top wall
  if (v_BC_Ymax == 0) then
    ! Dirichlet
    do i = 0, Nxm1
      matl(i, Nyp + 1) = 0.
      matd(i, Nyp + 1) = 1.
      matu(i, Nyp + 1) = 0.
      vec(i, Nyp + 1) = v_BC_Ymax_c1

      matl(i, Nyp) = 0.
      matd(i, Nyp) = 1.
      matu(i, Nyp) = 0.
      vec(i, Nyp) = v_BC_Ymax_c1
    end do
  else if (v_BC_Ymax == 1) then
    ! Neumann
    do i = 0, Nxm1
      matl(i, Nyp + 1) = -1.
      matd(i, Nyp + 1) = 1.
      matu(i, Nyp + 1) = 0.
      vec(i, Nyp + 1) = dyf(Nyp) * v_BC_Ymax_c1
    end do
  end if
  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|--
subroutine apply_BC_2_upper_c(matl_c, matd_c, matu_c, vec_c)
  !----*|--.---------.---------.---------.---------.---------.---------.-|--
  include 'header'
  integer i
  ! Top wall
  if (v_BC_Ymax == 0) then
    ! Dirichlet
    do i = 0, Nxp - 1 ! Nkx
      matl_c(i, Nyp + 1) = 0.
      matd_c(i, Nyp + 1) = 1.
      matu_c(i, Nyp + 1) = 0.
      vec_c(i, Nyp + 1) = v_BC_Ymax_c1

      matl_c(i, Nyp) = 0.
      matd_c(i, Nyp) = 1.
      matu_c(i, Nyp) = 0.
      vec_c(i, Nyp) = v_BC_Ymax_c1
    end do
  else if (v_BC_Ymax == 1) then
    ! Neumann
    do i = 0, Nxp - 1 ! Nkx
      matl_c(i, Nyp + 1) = -1.
      matd_c(i, Nyp + 1) = 1.
      matu_c(i, Nyp + 1) = 0.
      vec_c(i, Nyp + 1) = dyf(Nyp) * v_BC_Ymax_c1
    end do
  end if
  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|--
subroutine apply_BC_3_lower(matl, matd, matu, vec)
  !----*|--.---------.---------.---------.---------.---------.---------.-|--
  include 'header'
  integer i

  ! Bottom Wall:
  if (w_BC_Ymin == 0) then
    ! Dirichlet
    do i = 0, Nxm1
      matl(i, 0) = 0.
      matd(i, 0) = 1.
      matu(i, 0) = 0.
      vec(i, 0) = 0.

      matl(i, 1) = 0.
      matd(i, 1) = 1.
      matu(i, 1) = 0.
      vec(i, 1) = w_BC_Ymin_c1
    end do
  else
    ! Neumann
    do i = 0, Nxm1
      matl(i, 0) = 0.
      matd(i, 0) = 1.
      matu(i, 0) = 0.
      vec(i, 0) = 0.
    end do
    do i = 0, Nxm1
      matl(i, 1) = 0.
      matd(i, 1) = -1.
      matu(i, 1) = 1.
      vec(i, 1) = dy(2) * w_BC_Ymin_c1
    end do

  end if

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|--
subroutine apply_BC_3_lower_c(matl_c, matd_c, matu_c, vec_c)
  !----*|--.---------.---------.---------.---------.---------.---------.-|--
  include 'header'
  integer i

  ! Bottom Wall:
  if (w_BC_Ymin == 0) then
    ! Dirichlet
    do i = 0, Nxp - 1 ! Nkx
      matl_c(i, 0) = 0.
      matd_c(i, 0) = 1.
      matu_c(i, 0) = 0.
      vec_c(i, 0) = 0.

      matl_c(i, 1) = 0.
      matd_c(i, 1) = 1.
      matu_c(i, 1) = 0.
      vec_c(i, 1) = w_BC_Ymin_c1
    end do
  else
    ! Neumann
    do i = 0, Nxp - 1 ! Nkx
      matl_c(i, 0) = 0.
      matd_c(i, 0) = 1.
      matu_c(i, 0) = 0.
      vec_c(i, 0) = 0.
    end do
    do i = 0, Nxp - 1 ! Nkx
      matl_c(i, 1) = 0.
      matd_c(i, 1) = -1.
      matu_c(i, 1) = 1.
      vec_c(i, 1) = dy(2) * w_BC_Ymin_c1
    end do

  end if

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|--
subroutine apply_BC_3_upper(matl, matd, matu, vec)
  !----*|--.---------.---------.---------.---------.---------.---------.-|--
  include 'header'
  integer i

  ! Top wall
  if (w_BC_Ymax == 0) then
    ! Dirichlet
    do i = 0, Nxm1
      matl(i, Nyp + 1) = 0.
      matd(i, Nyp + 1) = 1.
      matu(i, Nyp + 1) = 0.
      vec(i, Nyp + 1) = 0.

      matl(i, Nyp) = 0.
      matd(i, Nyp) = 1.
      matu(i, Nyp) = 0.
      vec(i, Nyp) = w_BC_Ymax_c1
    end do
  else
    ! Neumann
    do i = 0, Nxm1
      matl(i, Nyp) = -1.
      matd(i, Nyp) = 1.
      matu(i, Nyp) = 0.
      vec(i, Nyp) = dy(Nyp) * w_BC_Ymax_c1
    end do
    do i = 0, Nxm1
      matl(i, Nyp + 1) = 0.
      matd(i, Nyp + 1) = 1.
      matu(i, Nyp + 1) = 0.
      vec(i, Nyp + 1) = 0.
    end do

  end if

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|--
subroutine apply_BC_3_upper_c(matl_c, matd_c, matu_c, vec_c)
  !----*|--.---------.---------.---------.---------.---------.---------.-|--
  include 'header'
  integer i

  ! Top wall
  if (w_BC_Ymax == 0) then
    ! Dirichlet
    do i = 0, Nxp - 1 ! Nkx
      matl_c(i, Nyp + 1) = 0.
      matd_c(i, Nyp + 1) = 1.
      matu_c(i, Nyp + 1) = 0.
      vec_c(i, Nyp + 1) = 0.

      matl_c(i, Nyp) = 0.
      matd_c(i, Nyp) = 1.
      matu_c(i, Nyp) = 0.
      vec_c(i, Nyp) = w_BC_Ymax_c1
    end do
  else
    ! Neumann
    do i = 0, Nxp - 1 ! Nkx
      matl_c(i, Nyp) = -1.
      matd_c(i, Nyp) = 1.
      matu_c(i, Nyp) = 0.
      vec_c(i, Nyp) = dy(Nyp) * w_BC_Ymax_c1
    end do
    do i = 0, Nxp - 1 ! Nkx
      matl_c(i, Nyp + 1) = 0.
      matd_c(i, Nyp + 1) = 1.
      matu_c(i, Nyp + 1) = 0.
      vec_c(i, Nyp + 1) = 0.
    end do

  end if

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|--
subroutine apply_BC_th_lower(matl, matd, matu, vec, n)
  !----*|--.---------.---------.---------.---------.---------.---------.-|--
  include 'header'
  integer i, n
  ! Bottom Wall:
  if (th_BC_Ymin(n) == 0) then
    ! Dirichlet
    do i = 0, Nxm1
      matl(i, 0) = 0.
      matd(i, 0) = 1.
      matu(i, 0) = 0.
      vec(i, 0) = 0.

      matl(i, 1) = 0.
      matd(i, 1) = 1.
      matu(i, 1) = 0.
      vec(i, 1) = th_BC_Ymin_c1(n)
    end do
  else
    ! Neumann
    ! NOTE: BC enforced at GY(2)
    do i = 0, Nxm1
      matl(i, 1) = 0.
      matd(i, 1) = -1.
      matu(i, 1) = 1.
      vec(i, 1) = dy(2) * th_BC_Ymin_c1(n)
    end do
    do i = 0, Nxm1
      matl(i, 0) = 0.
      matd(i, 0) = 1.
      matu(i, 0) = 0.
      vec(i, 0) = 0.
    end do
  end if
  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|--
subroutine apply_BC_th_lower_c(matl_c, matd_c, matu_c, vec_c, n)
  !----*|--.---------.---------.---------.---------.---------.---------.-|--
  include 'header'
  integer i, n
  ! Bottom Wall:
  if (th_BC_Ymin(n) == 0) then
    ! Dirichlet
    do i = 0, Nxp - 1 ! Nkx
      matl_c(i, 0) = 0.
      matd_c(i, 0) = 1.
      matu_c(i, 0) = 0.
      vec_c(i, 0) = 0.

      matl_c(i, 1) = 0.
      matd_c(i, 1) = 1.
      matu_c(i, 1) = 0.
      vec_c(i, 1) = th_BC_Ymin_c1(n)
    end do
  else
    ! Neumann
    ! NOTE: BC enforced at GY(2)
    do i = 0, Nxm1
      matl(i, 1) = 0.
      matd(i, 1) = -1.
      matu(i, 1) = 1.
      vec(i, 1) = dy(2) * th_BC_Ymin_c1(n)
    end do
    do i = 0, Nxm1
      matl(i, 0) = 0.
      matd(i, 0) = 1.
      matu(i, 0) = 0.
      vec(i, 0) = 0.
    end do

  end if
  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|--
subroutine apply_BC_th_upper(matl, matd, matu, vec, n)
  !----*|--.---------.---------.---------.---------.---------.---------.-|--
  include 'header'
  integer i, n
  ! Top wall
  if (th_BC_Ymax(n) == 0) then
    ! Dirichlet
    do i = 0, Nxm1
      matl(i, Nyp + 1) = 0.
      matd(i, Nyp + 1) = 1.
      matu(i, Nyp + 1) = 0.
      vec(i, Nyp + 1) = 0.

      matl(i, Nyp) = 0.
      matd(i, Nyp) = 1.
      matu(i, Nyp) = 0.
      vec(i, Nyp) = th_BC_Ymax_c1(n)
    end do
  else
    ! Neumann
    ! NOTE: BC enforced at GY(Nyp)
    do i = 0, Nxm1
      matl(i, Nyp) = -1.
      matd(i, Nyp) = 1.
      matu(i, Nyp) = 0.
      vec(i, Nyp) = dy(Nyp) * th_BC_Ymax_c1(n)
    end do
    do i = 0, Nxm1
      matl(i, Nyp + 1) = 0.
      matd(i, Nyp + 1) = 1.
      matu(i, Nyp + 1) = 0.
      vec(i, Nyp + 1) = 0.
    end do

  end if
  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|--
subroutine apply_BC_th_upper_c(matl_c, matd_c, matu_c, vec_c, n)
  !----*|--.---------.---------.---------.---------.---------.---------.-|--
  include 'header'
  integer i, n
  ! Top wall
  if (th_BC_Ymax(n) == 0) then
    ! Dirichlet
    do i = 0, Nxp - 1 ! Nkx
      matl_c(i, Nyp + 1) = 0.
      matd_c(i, Nyp + 1) = 1.
      matu_c(i, Nyp + 1) = 0.
      vec_c(i, Nyp + 1) = 0.

      matl_c(i, Nyp) = 0.
      matd_c(i, Nyp) = 1.
      matu_c(i, Nyp) = 0.
      vec_c(i, Nyp) = th_BC_Ymax_c1(n)
    end do
  else
    ! Neumann
    ! NOTE: BC enforced at GY(Nyp)

    do i = 0, Nxp - 1 ! Nkx
      matl_c(i, Nyp) = -1.
      matd_c(i, Nyp) = 1.
      matu_c(i, Nyp) = 0.
      vec_c(i, Nyp) = dy(Nyp) * th_BC_Ymax_c1(n)
    end do
    do i = 0, Nxp - 1 ! Nkx
      matl_c(i, Nyp + 1) = 0.
      matd_c(i, Nyp + 1) = 1.
      matu_c(i, Nyp + 1) = 0.
      vec_c(i, Nyp + 1) = 0.
    end do

  end if
  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|--
subroutine apply_BC_vel_lower
  !----*|--.---------.---------.---------.---------.---------.---------.-|--
  ! This subroutine is called after initializing the flow
  ! It sets the appropriate boundary conditions including ghost cell values
  !  on the velocity field in Fourier space
  include 'header'
  integer i, k

  ! Now, apply the boundary conditions depending on the type specified
  if (u_BC_Ymin == 0) then
    ! Dirichlet
    ! Start with zero
    do k = 0, twoNkz
      do i = 0, Nxp - 1
        cu1(i, k, 0) = 0.d0
        cu1(i, k, 1) = 0.d0
      end do
    end do
    ! Now, set only the mean
    if (rankZ == 0) then
      cu1(0, 0, 1) = u_BC_Ymin_c1
      cu1(0, 0, 0) = u_BC_Ymin_c1
    end if
  else if (u_BC_Ymin == 1) then
    ! Neumann
    do k = 0, twoNkz
      do i = 0, Nxp - 1
        cu1(i, k, 1) = cu1(i, k, 2)
        cu1(i, k, 0) = cu1(i, k, 1)
      end do
    end do
    ! Now, Apply BC to mean
    if (rankZ == 0) then
      cu1(0, 0, 1) = cu1(0, 0, 2) - dy(2) * u_BC_Ymin_c1
      cu1(0, 0, 0) = cu1(0, 0, 1) - dy(1) * u_BC_Ymin_c1
    end if
  else
    stop 'Error: u_BC_Zmin must be 0, or 1'
  end if

  if (w_BC_Ymin == 0) then
    ! Dirichlet
    ! Start with zero
    do k = 0, twoNkz
      do i = 0, Nxp - 1
        cu3(i, k, 0) = 0.d0
        cu3(i, k, 1) = 0.d0
      end do
    end do
    ! Now, set only the mean
    if (rankZ == 0) then
      cu3(0, 0, 1) = w_BC_Ymin_c1
      cu3(0, 0, 0) = w_BC_Ymin_c1
    end if
  else if (w_BC_Ymin == 1) then
    ! Neumann
    do k = 0, twoNkz
      do i = 0, Nxp - 1
        cu3(i, k, 1) = cu3(i, k, 2)
        cu3(i, k, 0) = cu3(i, k, 1)
      end do
    end do
    ! Now, Apply BC to mean
    if (rankZ == 0) then
      cu3(0, 0, 1) = cu3(0, 0, 2) - dy(2) * w_BC_Ymin_c1
      cu3(0, 0, 0) = cu3(0, 0, 1) - dy(1) * w_BC_Ymin_c1
    end if
  else
    stop 'Error: v_BC_Zmin must be 0, or 1'
  end if

  if (v_BC_Ymin == 0) then
    ! Dirichlet
    ! Set the vertical velocity at GYF(1) (halfway between GY(2) and GY(1))
    do k = 0, twoNkz
      do i = 0, Nxp - 1
        cu2(i, k, 1) = 2.d0 * v_BC_Ymin_c1 - cu2(i, k, 2)
        cu2(i, k, 0) = cu2(i, k, 1)
      end do
    end do
  else if (v_BC_Ymin == 1) then
    ! Neumann
    do k = 0, twoNkz
      do i = 0, Nxp - 1
        cu2(i, k, 1) = cu2(i, k, 2)
        cu2(i, k, 0) = cu2(i, k, 1)
      end do
    end do
    if (rankZ == 0) then
      cu2(0, 0, 1) = cu2(0, 0, 2) - dyf(1) * v_BC_Ymin_c1
      cu2(0, 0, 0) = cu2(0, 0, 1) - dyf(1) * v_BC_Ymin_c1
    end if

  else if (v_BC_Ymin == 2) then
    ! Upstream-travelling wave proposed by Speyer/Kim
    ! (initialize as zero)
    if (rankZ == 0) cu2(0, 0, 1) = -cu2(0, 0, 2)
  else
    stop 'Error: w_BC_Zmin must be 0, 1, or 2'
  end if

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|--
subroutine apply_BC_vel_upper
  !----*|--.---------.---------.---------.---------.---------.---------.-|--
  ! This subroutine is called after initializing the flow
  ! It sets the appropriate boundary conditions including ghost cell values
  !  on the velocity field in Fourier space
  include 'header'
  integer i, k

  ! Now, apply boundary conditions to the top of the domain
  if (u_BC_Ymax == 0) then
    ! Dirichlet
    ! Start with zero
    do k = 0, twoNkz
      do i = 0, Nxp - 1
        cu1(i, k, Nyp) = 0.d0
        cu1(i, k, Nyp + 1) = 0.d0
      end do
    end do
    ! Now, set only the mean
    if (rankZ == 0) then
      cu1(0, 0, Nyp) = u_BC_Ymax_c1
      cu1(0, 0, Nyp + 1) = u_BC_Ymax_c1
    end if
  else if (u_BC_Ymax == 1) then
    ! Neumann
    do k = 0, twoNkz
      do i = 0, Nxp - 1
        cu1(i, k, Nyp) = cu1(i, k, Nyp - 1)
        cu1(i, k, Nyp + 1) = cu1(i, k, Nyp)
      end do
    end do
    ! Now, Apply BC to mean
    if (rankZ == 0) then
      cu1(0, 0, Nyp) = cu1(0, 0, Nyp - 1) + dy(Nyp) * u_BC_Ymax_c1
      cu1(0, 0, Nyp + 1) = cu1(0, 0, Nyp) + dy(Nyp) * u_BC_Ymax_c1
    end if
  else
    stop 'Error: u_BC_Zmax must be 0, or 1'
  end if

  if (w_BC_Ymax == 0) then
    ! Dirichlet
    ! Start with zero
    do k = 0, twoNkz
      do i = 0, Nxp - 1
        cu3(i, k, Nyp) = 0.d0
        cu3(i, k, Nyp + 1) = 0.d0
      end do
    end do
    ! Now, set only the mean
    if (rankZ == 0) then
      cu3(0, 0, Nyp) = w_BC_Ymax_c1
      cu3(0, 0, Nyp + 1) = w_BC_Ymax_c1
    end if
    ! Ghost cell not used
    cu3(0, 0, Nyp + 1) = 0.d0
  else if (w_BC_Ymax == 1) then
    ! Neumann
    do k = 0, twoNkz
      do i = 0, Nxp - 1
        cu3(i, k, Nyp) = cu3(i, k, Nyp - 1)
        cu3(i, k, Nyp + 1) = cu3(i, k, Nyp)
      end do
    end do
    ! Now, Apply BC to mean
    if (rankZ == 0) then
      cu3(0, 0, Nyp) = cu3(0, 0, Nyp - 1) + dy(Nyp) * w_BC_Ymax_c1
      cu3(0, 0, Nyp + 1) = cu3(0, 0, Nyp) + dy(Nyp) * w_BC_Ymax_c1
    end if
  else
    stop 'Error: v_BC_Zmax must be 0, or 1'
  end if

  if (v_BC_Ymax == 0) then
    ! Dirichlet
    ! Set the vertical velocity at GYF(Nyp) (halfway between GY(Nyp) and GY(Nyp+1))
    do k = 0, twoNkz
      do i = 0, Nxp - 1
        cu2(0, 0, Nyp + 1) = 2.d0 * v_BC_Ymax_c1 - cu2(0, 0, Nyp)
      end do
    end do
  else if (v_BC_Ymax == 1) then
    ! Neumann
    do k = 0, twoNkz
      do i = 0, Nxp - 1
        cu2(i, k, Nyp) = cu2(i, k, Nyp - 1)
        cu2(i, k, Nyp + 1) = cu2(i, k, Nyp)
      end do
    end do
    ! Now, Apply BC to mean
    if (rankZ == 0) then
      cu2(0, 0, Nyp + 1) = cu2(0, 0, Nyp) + dy(Nyp) * v_BC_Ymax_c1
    end if
  else if (v_BC_Ymax == 2) then
    ! Upstream-travelling wave proposed by Speyer/Kim
    ! (initialize as zero gradient)
    if (rankZ == 0) cu2(0, 0, Nyp + 1) = -cu2(0, 0, Nyp)
  else
    stop 'Error: w_BC_Zmax must be 0, 1, or 2'
  end if

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|--
subroutine apply_BC_th_phys_lower
  !----*|--.---------.---------.---------.---------.---------.---------.-|--
  ! This subroutine is called after initializing the flow
  ! It sets the appropriate boundary conditions including ghost cell values
  !  on the velocity field in Physical space
  include 'header'
  integer i, k, n

  do n = 1, N_th

    ! Now, apply the boundary conditions depending on the type specified
    if (th_BC_Ymin(n) == 0) then
      ! Dirichlet
      ! Start with zero
      do k = 0, Nzp - 1
        do i = 0, Nxm1
          th(i, k, 0, n) = th_BC_Ymin_c1(n)
          th(i, k, 1, n) = th_BC_Ymin_c1(n)
        end do
      end do
    else if (th_BC_Ymin(1) == 1) then
      ! Neumann
      do k = 0, Nzp - 1
        do i = 0, Nxm1
          th(i, k, 1, n) = th(i, k, 2, n) - dy(2) * th_BC_Ymin_c1(n)
          th(i, k, 0, n) = th(i, k, 1, n) - dy(1) * th_BC_Ymin_c1(n)
        end do
      end do
    else
      stop 'Error: TH_BC_Zmin must be 0, or 1'
    end if

  end do

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|--
subroutine apply_BC_th_phys_upper
  !----*|--.---------.---------.---------.---------.---------.---------.-|--
  ! This subroutine is called after initializing the flow
  ! It sets the appropriate boundary conditions including ghost cell values
  !  on the velocity field in Fourier space
  include 'header'
  integer i, k, n

  do n = 1, N_th

    ! Now, apply boundary conditions to the top of the domain
    if (th_BC_Ymax(n) == 0) then
      ! Dirichlet
      ! Start with zero
      do k = 0, Nzp - 1
        do i = 0, Nxm1
          th(i, k, Nyp, n) = th_BC_Ymax_c1(n)
          th(i, k, Nyp + 1, n) = th_BC_Ymax_c1(n)
        end do
      end do
    else if (th_BC_Ymax(n) == 1) then
      ! Neumann
      do k = 0, Nzp - 1
        do i = 0, Nxm1
          th(i, k, Nyp, n) = th(i, k, Nyp - 1, n) + dy(Nyp) * th_BC_Ymax_c1(n)
          th(i, k, Nyp + 1, n) = th(i, k, Nyp, n) + dy(Nyp) * th_BC_Ymax_c1(n)
        end do
      end do
    else
      stop 'Error: TH_BC_Zmax must be 0, or 1'
    end if

  end do

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|--
subroutine apply_BC_vel_phys_lower
  !----*|--.---------.---------.---------.---------.---------.---------.-|--
  ! This subroutine is called after initializing the flow
  ! It sets the appropriate boundary conditions including ghost cell values
  !  on the velocity field in Physical space
  include 'header'
  integer i, k

  ! Now, apply the boundary conditions depending on the type specified
  if (u_BC_Ymin == 0) then
    ! Dirichlet
    ! Start with zero
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        u1(i, k, 0) = u_BC_Ymin_c1
        u1(i, k, 1) = u_BC_Ymin_c1
      end do
    end do
  else if (u_BC_Ymin == 1) then
    ! Neumann
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        u1(i, k, 1) = u1(i, k, 2) - dy(2) * u_BC_Ymin_c1
        u1(i, k, 0) = u1(i, k, 1) - dy(1) * u_BC_Ymin_c1
      end do
    end do
  else
    stop 'Error: u_BC_Zmin must be 0, or 1'
  end if

  if (w_BC_Ymin == 0) then
    ! Dirichlet
    ! Start with zero
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        u3(i, k, 0) = w_BC_Ymin_c1
        u3(i, k, 1) = w_BC_Ymin_c1
      end do
    end do
  else if (w_BC_Ymin == 1) then
    ! Neumann
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        u3(i, k, 1) = u3(i, k, 2) - dy(2) * w_BC_Ymin_c1
        u3(i, k, 0) = u3(i, k, 1) - dy(1) * w_BC_Ymin_c1
      end do
    end do
  else
    stop 'Error: v_BC_Zmin must be 0, or 1'
  end if

  if (v_BC_Ymin == 0) then
    ! Dirichlet
    ! Set the vertical velocity at GYF(1) (halfway between GY(2) and GY(1))
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        u2(i, k, 1) = 2.d0 * v_BC_Ymin_c1 - u2(i, k, 2)
        u2(i, k, 0) = u2(i, k, 1)
      end do
    end do
  else if (v_BC_Ymin == 1) then
    ! Neumann
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        u2(i, k, 1) = u2(i, k, 2) - dyf(1) * v_BC_Ymin_c1
        u2(i, k, 0) = u2(i, k, 1) - dyf(1) * v_BC_Ymin_c1
      end do
    end do
    !      ELSE IF (v_BC_Ymin==2) THEN
    ! Upstream-travelling wave proposed by Speyer/Kim
    ! (initialize as zero)
    !         IF (rankZ==0) CU2(0,0,1)=-CU2(0,0,2)
  else
    stop 'Error: w_BC_Zmin must be 0, 1'
  end if

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|--
subroutine apply_BC_vel_phys_upper
  !----*|--.---------.---------.---------.---------.---------.---------.-|--
  ! This subroutine is called after initializing the flow
  ! It sets the appropriate boundary conditions including ghost cell values
  !  on the velocity field in Fourier space
  include 'header'
  integer i, k

  ! Now, apply boundary conditions to the top of the domain
  if (u_BC_Ymax == 0) then
    ! Dirichlet
    ! Start with zero
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        u1(i, k, Nyp) = u_BC_Ymax_c1
        u1(i, k, Nyp + 1) = u_BC_Ymax_c1
      end do
    end do
  else if (u_BC_Ymax == 1) then
    ! Neumann
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        u1(i, k, Nyp) = u1(i, k, Nyp - 1) + dy(Nyp) * u_BC_Ymax_c1
        u1(i, k, Nyp + 1) = u1(i, k, Nyp) + dy(Nyp) * u_BC_Ymax_c1
      end do
    end do
  else
    stop 'Error: u_BC_Zmax must be 0, or 1'
  end if

  if (w_BC_Ymax == 0) then
    ! Dirichlet
    ! Start with zero
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        u3(i, k, Nyp) = w_BC_Ymax_c1
        u3(i, k, Nyp + 1) = w_BC_Ymax_c1
      end do
    end do
  else if (w_BC_Ymax == 1) then
    ! Neumann
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        u3(i, k, Nyp) = u3(i, k, Nyp - 1) + dy(Nyp) * w_BC_Ymax_c1
        u3(i, k, Nyp + 1) = u3(i, k, Nyp) + dy(Nyp) * w_BC_Ymax_c1
      end do
    end do
  else
    stop 'Error: v_BC_Zmax must be 0, or 1'
  end if

  if (v_BC_Ymax == 0) then
    ! Dirichlet
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        u2(i, k, Nyp + 1) = v_BC_Ymax_c1
        u2(i, k, Nyp) = v_BC_Ymax_c1
      end do
    end do
  else if (v_BC_Ymax == 1) then
    ! Neumann
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        u2(i, k, Nyp + 1) = u2(i, k, Nyp) + dy(Nyp) * v_BC_Ymax_c1
      end do
    end do
    !      ELSE IF (v_BC_Ymax==2) THEN
    ! Upstream-travelling wave proposed by Speyer/Kim
    ! (initialize as zero gradient)
    !         IF (rankZ==0) CU2(0,0,Nyp+1)=-CU2(0,0,Nyp)
  else
    stop 'Error: w_BC_Zmax must be 0, 1'
  end if

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine thomas_real(a, b, c, g, Nyp, Nx)
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  ! Uses the Thomas algorithm to solve Ax=b for tridiagonal A
  ! The RHS vector and solution are real
  ! Input lower, main, and upper diagonals, ld, md, ud, and rhs x
  ! Returns solution in x
  ! The indexing should be done by ROW, ie.
  ! [ b1  c1   0   0   0 ...
  ! [ a2  b2  c2   0   0 ...
  ! [  0  a3  b3   c3  0 ...

  integer i, j, Nx, Nyp
  real(kind(0.d0)) a(0:Nx, 0:Nyp), b(0:Nx, 0:Nyp), c(0:Nx, 0:Nyp), g(0:Nx, 0:Nyp)

  do j = 0, Nyp - 1
    do i = 0, Nx
      a(i, j + 1) = -a(i, j + 1) / b(i, j)
      b(i, j + 1) = b(i, j + 1) + a(i, j + 1) * c(i, j)
      g(i, j + 1) = g(i, j + 1) + a(i, j + 1) * g(i, j)
    end do
  end do
  do i = 0, Nx
    g(i, Nyp) = g(i, Nyp) / b(i, Nyp)
  end do
  do j = Nyp - 1, 0, -1
    do i = 0, Nx
      g(i, j) = (g(i, j) - c(i, j) * g(i, j + 1)) / b(i, j)
    end do
  end do

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine thomas_complex(a, b, c, g, Nyp, Nx)
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|

  ! Uses the Thomas algorithm to solve Ax=b for tridiagonal A
  ! The RHS vector and solution is complex
  ! Input lower, main, and upper diagonals, ld, md, ud, and rhs x
  ! Returns solution in x
  ! The indexing should be done by ROW, ie.
  ! [ b1  c1   0   0   0 ...
  ! [ a2  b2  c2   0   0 ...
  ! [  0  a3  b3   c3  0 ...

  integer i, j, k, Nyp, Nx
  real(kind(0.d0)) a(0:Nx, 0:Nyp), b(0:Nx, 0:Nyp), c(0:Nx, 0:Nyp)
  complex(kind(0.d0)) g(0:Nx, 0:Nyp)

  do j = 0, Nyp - 1
    do i = 0, Nx
      a(i, j + 1) = -a(i, j + 1) / b(i, j)
      b(i, j + 1) = b(i, j + 1) + a(i, j + 1) * c(i, j)
      g(i, j + 1) = g(i, j + 1) + a(i, j + 1) * g(i, j)
    end do
  end do
  do i = 0, Nx
    g(i, Nyp) = g(i, Nyp) / b(i, Nyp)
  end do
  do i = 0, Nx
    do j = Nyp - 1, 0, -1
      g(i, j) = (g(i, j) - c(i, j) * g(i, j + 1)) / b(i, j)
    end do
  end do

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|--
subroutine apply_BC_scalar_lower
  !----*|--.---------.---------.---------.---------.---------.---------.-|--
  ! This subroutine is called after initializing the flow
  ! It sets the appropriate boundary conditions including ghost cell values
  !  on the scalar fields in Fourier space
  include 'header'
  integer i, k, n

  do n = 1, N_th

    ! Now, apply the boundary conditions depending on the type specified
    if (th_BC_Ymin(n) == 0) then
      ! Dirichlet
      ! Start with zero
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          cth(i, k, 0, n) = 0.d0
          cth(i, k, 1, n) = 0.d0
        end do
      end do
      ! Now, set only the mean
      if (rankZ == 0) then
        cth(0, 0, 1, n) = th_BC_Ymin_c1(n)
        cth(0, 0, 0, n) = th_BC_Ymin_c1(n)
      end if
    else if (th_BC_Ymin(n) == 1) then
      ! Neumann
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          cth(i, k, 1, n) = cth(i, k, 2, n)
          cth(i, k, 0, n) = cth(i, k, 1, n)
        end do
      end do
      ! Now, Apply BC to mean
      if (rankZ == 0) then
        cth(0, 0, 1, n) = cth(0, 0, 2, n) - dy(2) * th_BC_Ymin_c1(n)
        cth(0, 0, 0, n) = cth(0, 0, 1, n) - dy(1) * th_BC_Ymin_c1(n)
      end if
    else
      stop 'Error: TH_BC_Zmin must be 0, or 1'
    end if

  end do

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|--
subroutine apply_BC_scalar_upper
  !----*|--.---------.---------.---------.---------.---------.---------.-|--
  ! This subroutine is called after initializing the flow
  ! It sets the appropriate boundary conditions including ghost cell values
  !  on the scalar fields in Fourier space
  include 'header'
  integer i, k, n

  do n = 1, N_th

    ! Now, apply boundary conditions to the top of the domain
    if (th_BC_Ymax(n) == 0) then
      ! Dirichlet
      ! Start with zero
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          cth(i, k, Nyp, n) = 0.d0
          cth(i, k, Nyp + 1, n) = 0.d0
        end do
      end do
      ! Now, set only the mean
      if (rankZ == 0) then
        cth(0, 0, Nyp, n) = th_BC_Ymax_c1(n)
        cth(0, 0, Nyp + 1, n) = th_BC_Ymax_c1(n)
      end if
    else if (th_BC_Ymax(n) == 1) then
      ! Neumann
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          cth(i, k, Nyp, n) = cth(i, k, Nyp - 1, n)
          cth(i, k, Nyp + 1, n) = cth(i, k, Nyp, n)
        end do
      end do
      ! Now, Apply B! to mean
      if (rankZ == 0) then
        cth(0, 0, Nyp, n) = cth(0, 0, Nyp - 1, n) + dy(Nyp) * th_BC_Ymax_c1(n)
        cth(0, 0, Nyp + 1, n) = cth(0, 0, Nyp, n) + dy(Nyp) * th_BC_Ymax_c1(n)
      end if
    else
      stop 'Error: TH_BC_Zmax must be 0, or 1'
    end if

  end do

  return
end
