!******************************************************************************|
! diablo.f90 -> DNS In A Box, Laptop Optimized                       VERSION 3.0
!
! This Fortran 90 code computes incompressible flow in a box.
!
! Primative variables (u,v,w,p) are used, and continuity is enforced with a
! fractional step algorithm.
!
! SPATIAL DERIVATIVES:
!   0, 1, 2, or 3 directions are taken to be periodic and handled spectrally
!   (these cases are referred to as the "periodic", "channel", "duct", and
!    "cavity" cases respectively).
!   The remaining directions are taken to be bounded by walls and handled with
!   momentum- and energy-conserving second-order central finite differences.
!
! TIME ADVANCEMENT
!   Two main approaches are implemented:
!     1. RKW3 on nonlinear terms and CN on viscous terms over each RK substep.
!     2. RKW3 on y-derivative terms and CN on other terms over each RK substep.
!
! The emphasis in this introductory code is on code simplicity:
!   -> All variables are in core.
!   -> The code is not explicitly designed for use with either MPI or SMP.
!   -> Overindexing is not used.
! A few simple high-performance programming constructs are used:
!   -> The inner 2 loops are broken out in such a way as to enable out-of-order
!      execution of these loops as much as possible, thereby leveraging
!      vector and superscalar CPU architectures.
!   -> The outer loops are fairly long (including as many operations as
!      possible inside on a single J plane of data) in order to make effective
!      use of cache.
! Multiple time advancement algorithms are implemented for the periodic,
! channel, duct, and cavity cases in order to compare their efficiency for
! various flows on different computational architectures.  In a few of the
! algorithms, the code is broken out fully into a PASS1/PASS2 architecture
! to maximize the efficient use of cache.
!
!******************************************************************************|
!
! This code is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by the
! Free Software Foundation; either version 2 of the License, or (at your
! option) any later version. This code is distributed in the hope that it
! will be useful, but WITHOUT any WARRANTY; without even the implied
! warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
! GNU General Public License for more details. You should have received a
! copy of the GNU General Public License along with this code; if not,
! write to the Free Software Foundation, Inc., 59 Temple Place - Suite
! 330, Boston, MA 02111-1307, USA.
!
!******************************************************************************|

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
program diablo
  include 'header'
  integer n
  logical flag

  call initialize

  ! Initialize start_wall_time for run timing
  call wall_time(start_wall_time)

  first_time = .true.

  do
    time_step = time_step + 1

    if (verbosity > 2 .and. rank == 0) &
      write (*, '("Now beginning time step ", I10)') time_step

    do rk_step = 1, 3
      if (num_per_dir == 3) then
        if (time_ad_meth == 1) call rk_per_1
        if (time_ad_meth == 2) call rk_per_2
      elseif (num_per_dir == 2) then
        if (time_ad_meth == 1) call rk_chan_1
        if (time_ad_meth == 2) call rk_chan_2
      end if
    end do
    time = time + delta_t
    first_time = .false.

    ! Optionally apply a filter to the scalar field
    do n = 1, N_th
      if (filter_th(n) &
          .and. (mod(time_step, filter_int(n)) == 0)) then
        call filter_chan(n)
      end if
    end do

    if (use_mpi) then
      call end_run_mpi(flag)
    else
      call end_run(flag)
    end if

    ! Save statistics to an output file
    if (time >= save_stats_time) then

      flag_save_LES = .true.
      if (time >= save_movie_time) then
        call save_stats(.true.,.false.)
        save_movie_time = save_stats_time + save_movie_dt ! Increment from stats_time
      else
        call save_stats(.false.,.false.)
      end if
      if (flag .and. LES) then ! Won't get into the next RK to save LES
        call save_stats_LES_OOL(.false.)
      end if
      save_stats_time = save_stats_time + save_stats_dt

      ! Timing Diagnostics
      call wall_time(end_wall_time)
      if (rank == 0) then
        write (*,'("Wall Seconds per Iteration: ", ES18.3)') &
                  (end_wall_time - previous_wall_time) / float(time_step - previous_time_step)
        write (*,'("Wall Seconds per Simulation Time: ", ES12.3)') &
                  (end_wall_time - previous_wall_time) / save_stats_dt

        write (*, *)

      end if
      call wall_time(previous_wall_time)
      previous_time_step = time_step

    end if

    ! Save entire flow to a file (Only really a restart file if end.h5)
    if (time >= save_flow_time) then
      save_flow_time = save_flow_time + save_flow_dt
      call save_flow(.false.)
    end if

    if (flag) then
      exit
    end if
  end do

  ! Calculate and display the runtime for the simulation
  call wall_time(end_wall_time)
  if (rank == 0) then
    write (*, '("Elapsed Wall Time (sec): ", ES11.1)') end_wall_time - start_wall_time
    write (*, '("Wall Seconds per Iteration: ", ES12.5)') (end_wall_time - start_wall_time) / time_step
  end if

  !time_step = time_step - 1
  call save_flow(.true.)
  !call save_stats(save_movie_dt/=0,.true.)
  if (use_mpi) then
    call mpi_finalize(ierror)
  end if

end
