!******************************************************************************|
! fft.f90, the FFT package for diablo.                               VERSION 0.9
!
! This file isolates all calls to the FFTW package (available at: www.fftw.org)
! These wrapper routines were written by T. Bewley (spring 2001).
!******************************************************************************|
!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
! The arrangement of the significant real numbers in the arrays (denoted by +)
! in physical space, in Fourier space, and in Fourier space after packing are
! shown below for the 2D (X-Z) plane.  The third direction (Y) is handled in
! an identical matter as the Z direction shown here.
!
!      oooooooooooooooooo         oooooooooooooooooo         oooooooooooooooooo
!      oooooooooooooooooo         oooooooooooooooooo         oooooooooooooooooo
! Nz-1 ++++++++++++++++oo     -1  ++++++++++++oooooo         oooooooooooooooooo
!      ++++++++++++++++oo     -2  ++++++++++++oooooo         oooooooooooooooooo
!      ++++++++++++++++oo     -3  ++++++++++++oooooo         oooooooooooooooooo
!      ++++++++++++++++oo         ++++++++++++oooooo         oooooooooooooooooo
!      ++++++++++++++++oo    -NkZ ++++++++++++oooooo         oooooooooooooooooo
!      ++++++++++++++++oo         oooooooooooooooooo     -1  ++++++++++++oooooo
!      ++++++++++++++++oo         oooooooooooooooooo     -2  ++++++++++++oooooo
!      ++++++++++++++++oo         oooooooooooooooooo     -3  ++++++++++++oooooo
!      ++++++++++++++++oo         oooooooooooooooooo         ++++++++++++oooooo
!      ++++++++++++++++oo         oooooooooooooooooo    -NkZ ++++++++++++oooooo
!      ++++++++++++++++oo     NkZ ++++++++++++oooooo     NkZ ++++++++++++oooooo
!      ++++++++++++++++oo         ++++++++++++oooooo         ++++++++++++oooooo
!   3  ++++++++++++++++oo      3  ++++++++++++oooooo      3  ++++++++++++oooooo
!   2  ++++++++++++++++oo      2  ++++++++++++oooooo      2  ++++++++++++oooooo
!   1  ++++++++++++++++oo      1  ++++++++++++oooooo      1  ++++++++++++oooooo
!   0  ++++++++++++++++oo      0  +o++++++++++oooooo      0  +o++++++++++oooooo
!      ^^^^           ^           ^ ^ ^     ^                ^ ^ ^     ^
!      0123           Nx-1        0 1 2     Nkx              0 1 2     Nkx
!
!       PHYSICAL SPACE              FOURIER SPACE         FOURIER SPACE (PACKED)
!
! After the Real->Fourier transform, the significant coefficients are put next
! to each other in the array, so a loop such as
!
!        DO K=0,twoNkz           [where twoNkz = 2*NkZ = 2*(Nz/3) ]
!          DO I=0,Nkx          [where  Nkx = Nx/3             ]
!            CP(I,K,J)= ...
!          END DO
!        END DO
!
! includes all the Fourier coefficients of interest.  The subsequent loops in
! Fourier space just work on these coefficients in the matrix.
!
! Before a Fourier->Real transform, the significant coefficients are unpacked
! and the higher wavenumbers are SET TO ZERO before the inverse transform.
! This has the effect of doing the required dealiasing.
!
!----*|--.---------.---------.---------.---------.---------.---------.-|-------|

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine init_fft
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  include 'header'
  integer i, j, k

  integer, parameter :: fftw_forward = -1
  integer, parameter :: fftw_backward = 1
  integer, parameter :: fftw_estimate = 0
  integer, parameter :: fftw_measure = 1
  integer, parameter :: fftw_out_of_place = 0
  integer, parameter :: fftw_in_place = 8
  integer, parameter :: fftw_use_wisdom = 16
  integer, parameter :: fftw_threadsafe = 128


  if (verbosity > 2 .and. rank == 0) &
    write (*, '("Initializing FFTW package....")')

  pi = 4.*atan(1.0)
  ci = cmplx(0.0, 1.0)
  eps = 0.000000001

  if (num_per_dir > 0) then
    call rfftwnd_f77_create_plan(fftw_x_to_f_plan, 1, Nx, &
                                 fftw_forward, fftw_measure)
    call rfftwnd_f77_create_plan(fftw_x_to_p_plan, 1, Nx, &
                                 fftw_backward, fftw_measure)
    rnx = 1.0 * Nx
    do i = 0, Nxp - 1
      kx(i) = (i + Nxp * rankZ) * (2.*pi) / Lx
      kx2(i) = kx(i) * kx(i)
      cikx(i) = ci * kx(i)
    end do

  end if

  if (num_per_dir > 1) then
    call fftwnd_f77_create_plan(fftw_z_to_f_plan, 1, Nz, &
                                fftw_forward, fftw_measure + fftw_in_place)
    call fftwnd_f77_create_plan(fftw_z_to_p_plan, 1, Nz, &
                                fftw_backward, fftw_measure + fftw_in_place)
    rnz = 1.0 * Nz
    do k = 0, NkZ
      kz(k) = k * (2.*pi) / Lz
    end do
    do k = 1, NkZ
      kz(twoNkz + 1 - k) = -k * (2.*pi) / Lz
    end do
    do k = 0, twoNkz
      kz2(k) = kz(k) * kz(k)
      cikz(k) = ci * kz(k)
    end do
  end if

  if (rank == 0) then
    write (*, '("Nkx   = " I10)') Nkx
    write (*, '("2Nkz  = " I10)') twoNkz
  end if

  if (num_per_dir > 2) then
    call fftwnd_f77_create_plan(fftw_y_to_f_plan, 1, Nyp, &
                                fftw_forward, fftw_measure + fftw_in_place)
    call fftwnd_f77_create_plan(fftw_y_to_p_plan, 1, Nyp, &
                                fftw_backward, fftw_measure + fftw_in_place)

    rNyp = 1.0 * Nyp
    ky(0) = 0.
    ky2(0) = 0.
    ciky(0) = (0.0, 0.0)
    do j = 1, Nky
      ky(j) = j * (2.*pi) / Ly
    end do
    do j = 1, Nky
      ky(twoNky + 1 - j) = -j * (2.*pi) / Ly
    end do
    do j = 1, twoNky
      ky2(j) = ky(j) * ky(j)
      ciky(j) = ci * ky(j)
    end do
  end if

  do i = 0, Nkx
    do k = 0, Nz
      czx_plane(k, i) = cmplx(0.0, 0.0)
    end do
  end do
  do k = 0, twoNkz
    do j = 0, Nyp
      cyz_plane(j, k) = cmplx(0.0, 0.0)
    end do
  end do

  if (verbosity > 2 .and. rank == 0) &
    write (*, '("FFTW package initialised.")')

  return
end

!******************************************************************************|
!-------------> The transform routines for the duct flow follow. <-------------|
!******************************************************************************|

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine fft_x_to_fourier(u, cu, jmin, jmax, kmin, kmax)
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  ! This routine transforms (in 1 direction) planes JMIN-JMAX to Fourier space.
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  include 'header'
  integer jmin, jmax, kmin, kmax, i, j, k
  real(rkind) u(0:Nx + 1, 0:Nz + 1, 0:Nyp + 1)
  complex(rkind) cu(0:Nx / 2, 0:Nz + 1, 0:Nyp + 1)

  ! Looping over the planes of interest, simply perform a real -> complex
  ! transform in place in the big storage array, scaling appropriately.

  do j = jmin, jmax
    call rfftwnd_f77_real_to_complex(fftw_x_to_f_plan, (kmax - kmin + 1), &
                                     u(0, kmin, j), 1, Nx + 2, cu(0, kmin, j), 1, Nx / 2 + 1)
    do k = kmin, kmax
      do i = 0, Nkx
        cu(i, k, j) = cu(i, k, j) / rnx
      end do
    end do
  end do

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine fft_x_to_physical(cu, u, jmin, jmax, kmin, kmax)
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  ! This routine transforms (in 1 direction) planes JMIN-JMAX to physical space.
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  include 'header'
  integer jmin, jmax, kmin, kmax, i, j, k
  real(rkind) u(0:Nx + 1, 0:Nz + 1, 0:Nyp + 1)
  complex(rkind) cu(0:Nx / 2, 0:Nz + 1, 0:Nyp + 1)

  ! Looping over the planes of interest, simply set the higher wavenumbers to
  ! zero and then perform a complex -> real transform in place in the big
  ! storage array.

  do j = jmin, jmax
    do k = kmin, kmax
      do i = Nkx + 1, Nx / 2
        cu(i, k, j) = 0.
      end do
    end do
    call rfftwnd_f77_complex_to_real(fftw_x_to_p_plan, (kmax - kmin + 1), &
                                     cu(0, kmin, j), 1, Nx / 2 + 1, u(0, kmin, j), 1, Nx + 2)
  end do

  return
end

!******************************************************************************|
!-----------> The transform routines for the channel flow follow. <------------|
!******************************************************************************|

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|

!
!     THE CHANNEL TRANSFORM FFT_XZ_TO_* HAVE BEEN MOVED TO THE MPI FILE
!     SINCE THEY REQUIRE 2D PARALLELIZATION
!

!******************************************************************************|
!--------> The transform routines for the fully-periodic box follow. <---------|
!******************************************************************************|

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine fft_xzy_to_fourier(u, cu)
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  ! This routine transforms (in 3 directions) the entire box to Fourier space.
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  include 'header'
  integer i, j, k
  real(rkind) u(0:Nx + 1, 0:Nz + 1, 0:Nyp + 1)
  complex(rkind) cu(0:Nx / 2, 0:Nz + 1, 0:Nyp + 1)

  ! First, transform in the X & Z directions using FFT_XZ_TO_FOURIER.  Then,
  ! looping over the planes of interest, put the data into the CYZ_PLANE
  ! temporary storage variable, perform a complex -> complex transform in the
  ! y direction, then put the data back into the big storage array, packing the
  ! data towards J=0 and scaling appropriately.

  call fft_xz_to_fourier(u, cu, 0, Nypm1)
  do i = 0, Nkx
    do k = 0, twoNkz
      do j = 0, Nypm1
        cyz_plane(j, k) = cu(i, k, j)
      end do
    end do
    call fftwnd_f77(fftw_y_to_f_plan, twoNkz + 1, &
                    cyz_plane(0, 0), 1, Nyp + 1, cyz_plane(0, 0), 1, Nyp + 1)
    do k = 0, twoNkz
      do j = 0, Nky
        cu(i, k, j) = cyz_plane(j, k) / rNyp
      end do
      do j = 1, Nky
        cu(i, k, Nky + j) = cyz_plane(Nypm1 - Nky + j, k) / rNyp
      end do
    end do
  end do

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine fft_xzy_to_physical(cu, u)
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  ! This routine transforms (in 3 directions) the entire box to physical space.
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  include 'header'
  integer i, j, k
  real(rkind) u(0:Nx + 1, 0:Nz + 1, 0:Nyp + 1)
  complex(rkind) cu(0:Nx / 2, 0:Nz + 1, 0:Nyp + 1)

  ! Looping over the planes of interest, unpack data into the CYZ_PLANE temporary
  ! storage variable (setting higher wavenumbers to zero), perform a complex ->
  ! complex transform in the y direction, then put the data back into the big
  ! storage array.  Finally, transform in the X & Z directions using
  ! FFT_XZ_TO_PHYSICAL.

  do i = 0, Nkx
    do k = 0, twoNkz
      do j = 0, Nky
        cyz_plane(j, k) = cu(i, k, j)
      end do
      do j = Nky + 1, Nypm1 - Nky
        cyz_plane(j, k) = cmplx(0.0, 0.0)
      end do
      do j = 1, Nky
        cyz_plane(Nypm1 - Nky + j, k) = cu(i, k, Nky + j)
      end do
    end do
    call fftwnd_f77(fftw_y_to_p_plan, twoNkz + 1, &
                    cyz_plane(0, 0), 1, Nyp + 1, cyz_plane(0, 0), 1, Nyp + 1)
    do k = 0, twoNkz
      do j = 0, Nypm1
        cu(i, k, j) = cyz_plane(j, k)
      end do
    end do
  end do
  call fft_xz_to_physical(cu, u, 0, Nypm1)

  return
end
