!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine les_chan
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  ! This subroutine models the terms owing to the subgrid scale stress
  ! if the computation is to be treated as an LES not a DNS
  ! This subroutine should be called when the velocity is in fourier space
  ! in the periodic directions, on output, the velocity will be
  ! in physical space.
  ! It is assumed that the test filter and the LES filter are performed
  ! by the same operation
  ! On output S1 should contain |S| which may be used again in les_chan_th
  ! if for the subgrid scalar dissipation

  include 'header'
  include 'header_les'

  integer i, j, k, l, m, ij

  real(rkind) S1_mean(0:Nyp + 1)
  real(rkind) NU_T_mean(0:Nyp + 1)
  real(rkind) eps_sgs1_mean(0:Nyp + 1)
  real(rkind) U3_bar(0:Nyp + 1)
  real(rkind) U1_bar(0:Nyp + 1)

  real(rkind), parameter :: c_smag = 0.17d0
  real(rkind) delta_y(0:Nyp + 1), delta_yf(0:Nyp + 1)
  real(rkind) denominator_sum

  ! Array for writing HDF5 files
  real(rkind) Diag(1:Nyp)
  character(len=20) gname

  character(len=35) fname

  ! Array to store the velocity index for each component of the strain rate tensor
  integer U_index1(6)
  integer U_index2(6)

  ! Here, alpha is the test/LES filter width ratio
  real(rkind), parameter :: alpha_sgs = 2.449d0
  ! beta is the LES/grid filter width ratio
  real(rkind), parameter :: beta_sgs = 1.5d0

  ! Set the indices that are used when adding the off-diagnoal SGS stress terms
  if (rankY == NprocY - 1) then
    ! We are at the upper wall
    j1 = jstart
    j2 = Nyp - 1
  else if (rankY == 0) then
    ! We are at the lower wall
    j1 = 2
    j2 = jend
  else
    ! We are on a middle process
    j1 = 2
    j2 = Nyp
  end if

  ! First, for all models, apply boundary conditions to the velocity field
  ! (fill ghost cells) to ensure accurate calculation of gradients
  ! Apply Boundary conditions to velocity field
  if (use_mpi) then
    call apply_BC_vel_mpi
    call ghost_chan_mpi
    if (flag_save_LES .and. rk_step == 1) then
      call ghost_chan_mpi_j0 ! Need u1/u3 at j = 0 for saving stats
    end if
  else
    call apply_BC_vel_lower
    call apply_BC_vel_upper
  end if

  ! If we are using Neuman boundary conditions, over-write the values of the
  ! velocity at the ghost cells so that the LES model doesn't use the large
  ! velocity gradient
  call apply_BC_les

  if (les_model_type == 1) then
    ! Constant Smagorinsky model

    ! First, compute the rate of strain tensor S_ij
    call compute_strain_chan

    ! To remove mean shear: first get the mean horizontal velocity as in save_stats
    ! Save the mean velocity (maybe JSTART to JEND)
    if (rankZ == 0) then
      do j = 0, Nyp + 1
        U1_bar(j) = dble(cu1(0, 0, j))
        U3_bar(j) = dble(cu3(0, 0, j))
      end do
    end if
    if (use_mpi) then
      call mpi_bcast(U1_bar, Nyp + 2, mpi_double_precision, 0, &
                     mpi_comm_z, ierror)
      call mpi_bcast(U3_bar, Nyp + 2, mpi_double_precision, 0, &
                     mpi_comm_z, ierror)
    end if

    ! Convert the velocity to physical space
    call fft_xz_to_physical(cu1, u1, 0, Nyp + 1)
    call fft_xz_to_physical(cu2, u2, 0, Nyp + 1)
    call fft_xz_to_physical(cu3, u3, 0, Nyp + 1)

    ! Compute |S| at GYF points, store in S1
    ! Interpolation to GYF points is easy since by definition
    ! GYF points are exactly midway between neighboring GY points
    ! Sij4 and Sij6 have the mean shear
    ! remove the zero value U1,3bar as in compute strain
    do j = 1, Nyp ! Need j = 1 for saving stats
      do k = 0, Nzp - 1
        do i = 0, Nxm1
          s1(i, k, j) = sqrt( &
                        2.d0 * Sij1(i, k, j)**2.d0 &
                        + 4.d0 * (0.5d0 * (Sij4(i, k, j + 1) + Sij4(i, k, j) &
                                           ! Optionally remove mean shear
                                           - 0.5d0 * (U1_bar(j) - U1_bar(j - 1)) / dy(j) &
                                           - 0.5d0 * (U1_bar(j + 1) - U1_bar(j)) / dy(j + 1) &
                                           ))**2.d0 &
                        + 4.d0 * Sij5(i, k, j)**2.d0 &
                        + 2.d0 * Sij2(i, k, j)**2.d0 &
                        + 4.d0 * (0.5d0 * (Sij6(i, k, j + 1) + Sij6(i, k, j) &
                                           ! Optionally remove mean shear
                                           - 0.5d0 * (U3_bar(j) - U3_bar(j - 1)) / dy(j) &
                                           - 0.5d0 * (U3_bar(j + 1) - U3_bar(j)) / dy(j + 1) &
                                           ))**2.d0 &
                        + 2.d0 * Sij3(i, k, j)**2.d0)
        end do
      end do
    end do

    ! Compute |S| at GY points and store in temp
    do j = 1, Nyp + 1  ! Need j = 1 for saving stats
      do k = 0, Nzp - 1
        do i = 0, Nxm1
          temp(i, k, j) = sqrt( &
                          2.d0 * ((Sij1(i, k, j) * dyf(j - 1) + Sij1(i, k, j - 1) * dyf(j)) &
                                  / (2.d0 * dy(j)))**2.d0 &
                          + 4.d0 * (Sij4(i, k, j) &
                                    ! optionally remove mean shear
                                    - 0.5d0 * (U1_bar(j) - U1_bar(j - 1)) / dy(j))**2.d0 &
                          + 4.d0 * ((Sij5(i, k, j) * dyf(j - 1) + Sij5(i, k, j - 1) * dyf(j)) &
                                    / (2.d0 * dy(j)))**2.d0 &
                          + 2.d0 * ((Sij2(i, k, j) * dyf(j - 1) + Sij2(i, k, j - 1) * dyf(j)) &
                                    / (2.d0 * dy(j)))**2.d0 &
                          + 4.d0 * (Sij6(i, k, j) &
                                    ! remove mean shear
                                    - 0.5d0 * (U3_bar(j) - U3_bar(j - 1)) / dy(j))**2.d0 &
                          + 2.d0 * ((Sij3(i, k, j) * dyf(j - 1) + Sij3(i, k, j - 1) * dyf(j)) &
                                    / (2.d0 * dy(j)))**2.d0 )
        end do
      end do
    end do

    ! Now, compute |S|*S_ij, storing in Sij
    ! First compute at GYF points
    do j = 1, Nyp  ! Need j = 1 for saving stats
      do k = 0, Nzp - 1
        do i = 0, Nxm1
          Sij1(i, k, j) = s1(i, k, j) * Sij1(i, k, j)
          Sij5(i, k, j) = s1(i, k, j) * Sij5(i, k, j)
          ! Sij2 is added through an implicit eddy viscosity
          Sij2(i, k, j) = 0.d0
          Sij3(i, k, j) = s1(i, k, j) * Sij3(i, k, j)
        end do
      end do
    end do

    ! Now, compute at |S|*S_ij at GY points
    do j = 1, Nyp + 1  ! Need j = 1 for saving stats
      do k = 0, Nzp - 1
        do i = 0, Nxm1
          ! The terms dU1/dy and dU3/dy in CSij4(:,:,:) and CSij6(:,:,:) respectively
          ! are subtracted out from Sij here since they are treated implicitly
          ! in eddy viscosity terms
          Sij4(i, k, j) = temp(i, k, j) &
                          * (Sij4(i, k, j) - 0.5 * (u1(i, k, j) - u1(i, k, j - 1)) / dy(j))
          Sij6(i, k, j) = temp(i, k, j) &
                          * (Sij6(i, k, j) - 0.5 * (u3(i, k, j) - u3(i, k, j - 1)) / dy(j))
        end do
      end do
    end do

    ! We now have |S|*S_ij stored in Sij in Physical space

    ! Compute the filter lengthscale
    ! Absorb -2.d0*C_SMAG**2.d0 here for efficiency
    do j = 1, Nyp  ! Need j = 1 for saving stats
      ! At GYF points:
      ! Constant Smagorinsky
      delta_yf(j) = -2.d0 * c_smag**2.d0 &
                    * (dx(1) * beta_sgs * dyf(j) * 1.d0 * dz(1) * beta_sgs)**(2.d0 / 3.d0)
      ! Wall Damping
      ! delta_yf(j) = -2.d0 * (0.1d0 * (1.d0 - exp((-gyf(j)/(nu*25.d0))**3.d0)))**2.d0 &
      !                 * (dx(1)*beta_sgs*dyf(j)*2.d0*dz(1)*beta_sgs)**(2.d0/3.d0)

    end do

    do j = 1, Nyp + 1  ! Need j = 1 for saving stats
      ! At GY points:
      ! Constant Smagorinsky
      delta_y(j) = -2.d0 * c_smag**2.d0 &
                   * (dx(1) * beta_sgs * dy(j) * 2.d0 * dz(1) * beta_sgs)**(2.d0 / 3.d0)
      ! Wall Damping
      ! delta_y(j) = -2.d0 * (0.1d0 * (1.d0 - exp((-gy(j)/(nu*25.d0))**3.d0)))**2.d0 &
      !                * (dx(1)*beta_sgs*dy(j)*2.d0*dz(1)*beta_sgs)**(2.d0/3.d0)
    end do

    ! Get the eddy viscosity at GY points
    ! NU_T = (C_S^2 * DELTA^2)*|S|
    do j = 2, Nyp
      do k = 0, Nzp - 1
        do i = 0, Nxm1
          nu_t(i, k, j) = -0.5d0 * delta_y(j) * temp(i, k, j)
        end do
      end do
    end do

    ! Now that we have calculated NU_T, set the value at the ghost cells
    ! by sharing with neighboring processes.  This subroutine also sets
    ! the value of NU_T to zero at both walls
    call ghost_les_mpi

    ! Convert the stress tensor to Fourier space

    call fft_xz_to_fourier(Sij1, CSij1, 0, Nyp + 1)
    ! Sij2 is added through an implicit eddy viscosity
    ! call fft_xz_to_fourier(Sij2, CSij2, 0 , Nyp + 1)
    call fft_xz_to_fourier(Sij3, CSij3, 0, Nyp + 1)
    call fft_xz_to_fourier(Sij4, CSij4, 0, Nyp + 1)
    call fft_xz_to_fourier(Sij5, CSij5, 0, Nyp + 1)
    call fft_xz_to_fourier(Sij6, CSij6, 0, Nyp + 1)

    ! Now, compute TAU, store in the corresponging Sij
    do j = 1, Nyp  ! Need j = 1 for saving stats
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          CSij1(i, k, j) = delta_yf(j) * CSij1(i, k, j)
          CSij5(i, k, j) = delta_yf(j) * CSij5(i, k, j)
          ! CSij2 is added through an implicit eddy viscosity
          ! CSij2(i, k, j) = delta_yf(j) * CSij2(i, k, j)
          CSij3(i, k, j) = delta_yf(j) * CSij3(i, k, j)
        end do
      end do
    end do
    do j = 1, Nyp + 1  ! Need j = 1 for saving stats
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          CSij4(i, k, j) = delta_y(j) * CSij4(i, k, j)
          CSij6(i, k, j) = delta_y(j) * CSij6(i, k, j)
        end do
      end do
    end do

    ! tau_ij is now contained in CSij in Fourier space

  else if ((les_model_type == 2) .or. (les_model_type == 3)) then
    ! Here, use a dynamic smagorinsky model with or without scale similar part

    stop 'Error: LES_MODEL_TYPE=2, 3 not supported yet in MPI'

  end if

  ! Now, add the subgrid scale forcing to CFi
  ! (This includes the subgrid scale stress as an explicit R-K term)

  do j = j1, j2
    do k = 0, twoNkz
      do i = 0, Nxp - 1
        cf1(i, k, j) = cf1(i, k, j) &
                       - cikx(i) * CSij1(i, k, j) &
                       - (CSij4(i, k, j + 1) - CSij4(i, k, j)) / dyf(j) &
                       - cikz(k) * CSij5(i, k, j)
        cf3(i, k, j) = cf3(i, k, j) &
                       - cikx(i) * CSij5(i, k, j) &
                       - (CSij6(i, k, j + 1) - CSij6(i, k, j)) / dyf(j) &
                       - cikz(k) * CSij3(i, k, j)
      end do
    end do
  end do
  do j = 2, Nyp
    do k = 0, twoNkz
      do i = 0, Nxp - 1
        cf2(i, k, j) = cf2(i, k, j) &
                      - cikx(i) * CSij4(i, k, j) &
        ! Sij2 is added through an implict eddy viscosity
        !             - (CSij2(i, k, j) - CSij2(i, k, j - 1)) / dy(j) &
                      - cikz(k) * CSij6(i, k, j)
      end do
    end do
  end do

  ! Add back the Cross-terms using an explicit treatment
  do k = 0, Nzp - 1
    do i = 0, Nxm1
      do j = 2, Nyp
        temp(i, k, j) = nu_t(i, k, j) * (u1(i, k, j) - u1(i, k, j - 1)) / dy(j)
      end do
    end do
  end do
  call fft_xz_to_fourier(temp, ctemp, 2, Nyp)
  do k = 0, twoNkz
    do i = 0, Nxp - 1
      do j = 2, Nyp
        cf2(i, k, j) = cf2(i, k, j) + cikx(i) * ctemp(i, k, j)
      end do
    end do
  end do
  do k = 0, Nzp - 1
    do i = 0, Nxm1
      do j = 2, Nyp
        temp(i, k, j) = nu_t(i, k, j) * (u3(i, k, j) - u3(i, k, j - 1)) / dy(j)
      end do
    end do
  end do
  call fft_xz_to_fourier(temp, ctemp, 2, Nyp)
  do k = 0, twoNkz
    do i = 0, Nxp - 1
      do j = 2, Nyp
        cf2(i, k, j) = cf2(i, k, j) + cikz(k) * ctemp(i, k, j)
      end do
    end do
  end do



  ! Periodically, output mean quantities
  if (flag_save_LES .and. rk_step == 1) then
    flag_save_LES = .false.
    ! NOTE eps_sgs1 now is consistently evaluated at GY points

    ! Get plane averages
    do j = 0, Nyp + 1
      S1_mean(j) = 0.d0
      NU_T_mean(j) = 0.d0
      do i = 0, Nxm1
        do k = 0, Nzp - 1
          S1_mean(j) = S1_mean(j) + s1(i, k, j)
          NU_T_mean(j) = NU_T_mean(j) + nu_t(i, k, j)
        end do
      end do
    end do

    eps_sgs1_mean = 0.d0 ! NOTE: eps_sgs1 is actually negative of the convention
    ! compute part of sgs term
    ! U1*F1+U2*F2+U3*F3
    ! compute part of CF1, then interpolate onto GY
    do j = 1, Nyp
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          ctemp(i, k, j) = &
            - cikx(i) * CSij1(i, k, j) &
            - (CSij4(i, k, j + 1) - CSij4(i, k, j)) / dyf(j) &
            - cikz(k) * CSij5(i, k, j)
        end do
      end do
    end do
    call fft_xz_to_physical(ctemp, temp, 0, Nyp + 1)
    do j = 2, Nyp
      do i = 0, Nxm1
        do k = 0, Nzp - 1
          eps_sgs1_mean(j) = eps_sgs1_mean(j) + &
                              0.5d0 * (u1(i, k, j) * temp(i, k, j) &
                                     + u1(i, k, j - 1) * temp(i, k, j - 1))
        end do
      end do
    end do
    ! compute part of CF3, then interpolate onto GY
    do j = 1, Nyp
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          ctemp(i, k, j) = &
            - cikx(i) * CSij5(i, k, j) &
            - (CSij6(i, k, j + 1) - CSij6(i, k, j)) / dyf(j) &
            - cikz(k) * CSij3(i, k, j)
        end do
      end do
    end do
    call fft_xz_to_physical(ctemp, temp, 0, Nyp + 1)
    do j = 2, Nyp
      do i = 0, Nxm1
        do k = 0, Nzp - 1
          eps_sgs1_mean(j) = eps_sgs1_mean(j) + &
                              0.5d0 * (u3(i, k, j) * temp(i, k, j) &
                                     + u3(i, k, j - 1) * temp(i, k, j - 1))
        end do
      end do
    end do
    ! compute part of CF2, on GY
    do j = 2, Nyp
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          ctemp(i, k, j) = &
            - cikx(i) * CSij4(i, k, j) &
            ! Sij2 is added through an implict eddy viscosity
            ! - (CSij2(i, k, j) - CSij2(i, k, j - 1)) / dy(j)
            - cikz(k) * CSij6(i, k, j)
        end do
      end do
    end do
    call fft_xz_to_physical(ctemp, temp, 0, Nyp + 1)
    do j = 2, Nyp
      do i = 0, Nxm1
        do k = 0, Nzp - 1
          eps_sgs1_mean(j) = eps_sgs1_mean(j) + temp(i, k, j) * u2(i, k, j)
        end do
      end do
    end do
    ! Add back the Cross-terms (previously subtracted from Sij, thus tau_ij)
    ! using an explicit treatment
    ! compute second part of CF2, on GY
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        do j = 2, Nyp
          temp(i, k, j) = nu_t(i, k, j) * (u1(i, k, j) - u1(i, k, j - 1)) / dy(j)
        end do
      end do
    end do
    call fft_xz_to_fourier(temp, ctemp, 2, Nyp)
    do k = 0, twoNkz
      do i = 0, Nxp - 1
        do j = 2, Nyp
          ctemp(i, k, j) = cikx(i) * ctemp(i, k, j)
        end do
      end do
    end do
    call fft_xz_to_physical(ctemp, temp, 0, Nyp + 1)
    do j = 2, Nyp
      do i = 0, Nxm1
        do k = 0, Nzp - 1
          eps_sgs1_mean(j) = eps_sgs1_mean(j) + temp(i, k, j) * u2(i, k, j)
        end do
      end do
    end do
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        do j = 2, Nyp
          temp(i, k, j) = nu_t(i, k, j) * (u3(i, k, j) - u3(i, k, j - 1)) / dy(j)
        end do
      end do
    end do
    call fft_xz_to_fourier(temp, ctemp, 2, Nyp)
    do k = 0, twoNkz
      do i = 0, Nxp - 1
        do j = 2, Nyp
          ctemp(i, k, j) = cikz(k) * ctemp(i, k, j)
        end do
      end do
    end do
    call fft_xz_to_physical(ctemp, temp, 0, Nyp + 1)
    do j = 2, Nyp
      do i = 0, Nxm1
        do k = 0, Nzp - 1
          eps_sgs1_mean(j) = eps_sgs1_mean(j) + temp(i, k, j) * u2(i, k, j)
        end do
      end do
    end do

    call mpi_allreduce(mpi_in_place, S1_mean, Nyp + 2, mpi_double_precision, &
                       mpi_sum, mpi_comm_z, ierror)
    call mpi_allreduce(mpi_in_place, NU_T_mean, Nyp + 2 &
                       , mpi_double_precision, &
                       mpi_sum, mpi_comm_z, ierror)
    call mpi_allreduce(mpi_in_place, eps_sgs1_mean, Nyp + 2 &
                       , mpi_double_precision, &
                       mpi_sum, mpi_comm_z, ierror)

    do j = 0, Nyp + 1
      S1_mean(j) = S1_mean(j) / float(Nx * Nz)
      NU_T_mean(j) = NU_T_mean(j) / float(Nx * Nz)
      eps_sgs1_mean(j) = eps_sgs1_mean(j) / float(Nx * Nz)
    end do

    fname = 'mean.h5'

    if (rankZ == 0) then

      gname = 'nu_sgs'
      Diag = NU_T_mean(1:Nyp)
      call WriteStatH5_Y(fname, gname, Diag)

      gname = 'eps_sgs1'
      Diag = eps_sgs1_mean(1:Nyp)
      call WriteStatH5_Y(fname, gname, Diag)

    end if

  end if

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine compute_strain_chan
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  ! This subroutine computes S_ij for the filtered velocity field
  ! The input velocity field should be in fourier space in the periodic
  ! directions.
  ! For use in the LES model in channel flow (2 periodic directions)
  include 'header'
  include 'header_les'

  integer i, j, k, ij

  do j = 0, Nyp + 1 ! Need j = 0 for saving stats
    do k = 0, twoNkz
      do i = 0, Nxp - 1
        CSij1(i, k, j) = cikx(i) * cu1(i, k, j)
        CSij3(i, k, j) = cikz(k) * cu3(i, k, j)
        CSij5(i, k, j) = 0.5d0 * (cikz(k) * cu1(i, k, j) &
                                  + cikx(i) * cu3(i, k, j))
        if (j /= Nyp + 1) then
          CSij2(i, k, j) = (cu2(i, k, j + 1) - cu2(i, k, j)) / dyf(j)
        end if
      end do
    end do
  end do
  do j = 1, Nyp + 1 ! Need j = 1 for saving stats
    do k = 0, twoNkz
      do i = 0, Nxp - 1
        CSij4(i, k, j) = 0.5d0 * ((cu1(i, k, j) - cu1(i, k, j - 1)) / dy(j) &
                                  + cikx(i) * cu2(i, k, j))
        CSij6(i, k, j) = 0.5d0 * (cikz(k) * cu2(i, k, j) &
                                  + (cu3(i, k, j) - cu3(i, k, j - 1)) / dy(j))
      end do
    end do
  end do

  ! Need to communicate down CSij2 to each j = Nyp + 1
  call ghost_CSij2_mpi


  call fft_xz_to_physical(CSij1, Sij1, 0, Nyp + 1)
  call fft_xz_to_physical(CSij2, Sij2, 0, Nyp + 1)
  call fft_xz_to_physical(CSij3, Sij3, 0, Nyp + 1)
  call fft_xz_to_physical(CSij4, Sij4, 0, Nyp + 1)
  call fft_xz_to_physical(CSij5, Sij5, 0, Nyp + 1)
  call fft_xz_to_physical(CSij6, Sij6, 0, Nyp + 1)

  ! We now have S_ij in Physical space

  return
end

! !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
! subroutine les_filter_chan(a, jstart, jend)
!   !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
!   ! This subroutine applies the LES filter to the input field
!   ! The indices to the start and end of the array in the y-direction
!   ! are also inputted to make the routine cablable of filtering fields
!   ! at either GYF or GY points.
!   ! The array that is passed should be in physical space
!   include 'header'
!
!   integer i, j, k, n
!
!   real(rkind) a(0:Nx + 1, 0:Nz + 1, 0:Nyp + 1)
!   real(rkind) b(0:Nx - 1, 0:Nz - 1, 0:Nyp + 1)
!
!   integer im2(0:Nx - 1), im1(0:Nx - 1), ip1(0:Nx + 1), ip2(0:Nx + 2)
!   integer km2(0:Nz - 1), km1(0:Nz - 1), kp1(0:Nz + 1), kp2(0:Nz + 2)
!
!   ! These are the weights for the filtering operation used
!   real(rkind) w0, w1, w2, Wm1, Wm2, Wm1_j, W0_j, W1_j
!
!   ! The following is for the 3-point trapezoidal rule, alpha*beta=sqrt(6)
!   !      Wm2=0.d0
!   !      Wm1=1.d0/4.d0
!   !      W0=1.d0/2.d0
!   !      W1=1.d0/4.d0
!   !      W2=0.d0
!   Wm1_j = 1.d0 / 4.d0
!   W0_j = 1.d0 / 2.d0
!   W1_j = 1.d0 / 4.d0
!   ! The following is for the 5-point trapezoidal rule, alpha*beta=9
!   Wm2 = 1.d0 / 8.d0
!   Wm1 = 1.d0 / 4.d0
!   w0 = 1.d0 / 4.d0
!   w1 = 1.d0 / 4.d0
!   w2 = 1.d0 / 8.d0
!
!   Nxm1 = Nx - 1
!   Nzm1 = Nz - 1
!
!   !      do j=0,Nyp+1
!   !        do k=0,Nzm1
!   !          do i=0,Nxm1
!   !            B(i,k,j)=A(i,k,j)
!   !          end do
!   !        end do
!   !      end do
!
!   ! Filter in the periodic directions using cshift
!   ! Note, cshift if not used since it appears to be slower
!   ! Apply filter in the x-direction
!   !      B=Wm2*CSHIFT(B,-2,1)+Wm1*CSHIFT(B,-1,1)+W0*B+W1*CSHIFT(B,1,1)
!   !     &       +W2*CSHIFT(B,2,1)
!
!   ! Filter using more efficient F77 syntax:
!   ! Set up array to loop around periodic directions
!   do i = 2, Nxm1
!     im2(i) = i - 2
!   end do
!   im2(1) = Nxm1
!   im2(0) = Nx - 2
!   do i = 1, Nxm1
!     im1(i) = i - 1
!   end do
!   im1(0) = Nxm1
!   do i = 0, Nx - 2
!     ip1(i) = i + 1
!   end do
!   ip1(Nxm1) = 0
!   do i = 0, Nx - 3
!     ip2(i) = i + 2
!   end do
!   ip2(Nx - 2) = 0
!   ip2(Nxm1) = 1
!
!   do j = jstart, jend
!     do k = 0, Nzm1
!       do i = 0, Nxm1
!         b(i, k, j) = Wm2 * a(im2(i), k, j) + Wm1 * a(im1(i), k, j) + w0 * a(i, k, j) &
!                      + w1 * a(ip1(i), k, j) + w2 * a(ip2(i), k, j)
!       end do
!     end do
!   end do
!
!   ! Apply filter in the z-diretion
!   !      B=Wm2*CSHIFT(B,-2,2)+Wm1*CSHIFT(B,-1,2)+W0*B+W1*CSHIFT(B,1,2)
!   !     &       +W2*CSHIFT(B,2,2)
!   ! Filter using more efficient F77 syntax:
!   ! Set up array to loop around periodic directions
!   do k = 2, Nzm1
!     km2(k) = k - 2
!   end do
!   km2(1) = Nzm1
!   km2(0) = Nz - 2
!   do k = 1, Nzm1
!     km1(k) = k - 1
!   end do
!   km1(0) = Nzm1
!   do k = 0, Nz - 2
!     kp1(k) = k + 1
!   end do
!   kp1(Nzm1) = 0
!   do k = 0, Nz - 3
!     kp2(k) = k + 2
!   end do
!   kp2(Nz - 2) = 0
!   kp2(Nzm1) = 1
!
!   do j = jstart, jend
!     do k = 0, Nzm1
!       do i = 0, Nxm1
!         a(i, k, j) = Wm2 * b(i, km2(k), j) + Wm1 * b(i, km1(k), j) + w0 * b(i, k, j) &
!                      + w1 * b(i, kp1(k), j) + w2 * b(i, kp2(k), j)
!       end do
!     end do
!   end do
!
!   ! Apply filter in the vertical direction at all physical cells
!   ! (filter is not applied to ghost cells, but the values of the ghost cells
!   ! is used for averaging)
!   !      B(:,:,jstart+1:jend-1)=W0*B(:,:,jstart:jend-2)
!   !     &                      +W1*B(:,:,jstart+1:jend-1)
!   !     &                      +W2*B(:,:,jstart+2:jend)
!   ! Use more efficient F77 syntax:
!   !       do j=jstart+1,jend-1
!   !         do k=0,Nzm1
!   !           do i=0,Nxm1
!   !             B(i,k,j)=Wm1_j*B(i,k,j-1)+W0_j*B(i,k,j)+W1_j*B(i,k,j+1)
!   !           end do
!   !         end do
!   !       end do
!
!   !      do j=jstart,jend
!   !        do k=0,Nzm1
!   !          do i=0,Nxm1
!   !            A(i,k,j)=B(i,k,j)
!   !          end do
!   !        end do
!   !      end do
!
!   return
! end

! !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
! subroutine les_filter_chan_fourier(a, jstart, jend)
!   !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
!   ! This subroutine applies the LES filter to the input field
!   ! The filter is a spectral cutoff filter
!   ! The indices to the start and end of the array in the y-direction
!   ! are also inputted to make the routine cablable of filtering fields
!   ! at either GYF or GY points.
!   ! The array that is passed should be in physical space
!   include 'header'
!
!   integer i, j, k
!
!   real(rkind) pi, Lx, Lz
!   integer Nkx, NkZ, twoNkz
!
!   real(rkind) kx(0:Nx / 3), kz(0:2 * (Nz / 3))
!
!   real(rkind) a(0:Nx + 1, 0:Nz + 1, 0:Nyp + 1)
!
!   real(rkind), parameter :: alpha = 2.d0 ! Ratio of filter scales
!
!   real(rkind) b(0:Nx + 1, 0:Nz + 1, 0:Nyp + 1)
!
!   complex(rkind) cb(0:Nx / 2, 0:Nz + 1, 0:Nyp + 1)
!
!   equivalence(b, cb)
!
!   pi = 4.*atan(1.0)
!
!   Lx = pi
!   Lz = 2.d0 * pi
!
!   ! Get the wavenumber vectors:
!   Nkx = Nx / 3
!   do i = 0, Nkx
!     kx(i) = i * (2.*pi) / Lx
!   end do
!
!   NkZ = Nz / 3
!   twoNkz = NkZ * 2
!   do k = 0, NkZ
!     kz(k) = k * (2.*pi) / Lz
!   end do
!   do k = 1, NkZ
!     kz(twoNkz + 1 - k) = -k * (2.*pi) / Lz
!   end do
!
!   do j = 0, Nyp + 1
!     do k = 0, Nzm1
!       do i = 0, Nxm1
!         b(i, k, j) = a(i, k, j)
!       end do
!     enddo
!   end do
!
!   ! Convert to fourier space
!   call fft_xz_to_fourier(b, cb, jstart, jend)
!
!   ! Perform the filtering
!   do j = jstart, jend
!     do k = 0, twoNkz
!       do i = 0, Nkx
!         if (sqrt(kx(i)**2.d0 + kz(k)**2.d0) &
!             > sqrt(kx(Nkx)**2.d0 + kz(NkZ)**2.d0) / alpha) then
!           cb(i, k, j) = 0.d0
!         end if
!       end do
!     end do
!   end do
!
!   ! Now, convert back to physical space
!   call fft_xz_to_physical(cb, b, jstart, jend)
!
!   do j = jstart, jend
!     do k = 0, Nzm1
!       do i = 0, Nxm1
!         a(i, k, j) = b(i, k, j)
!       end do
!     end do
!   end do
!
!   return
! end

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine apply_BC_les
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  include 'header'
  integer i, j, k

  ! If we are using Neuman boundary conditions, over-write the values of the
  ! velocity at the ghost cells so that the LES model doesn't use the large
  ! velocity gradient
  if (u_BC_Ymax == 1) then
    if ((rankY == NprocY - 1) .or. (.not. use_mpi)) then
      ! We are on process at the upper wall
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          cu1(i, k, Nyp) = cu1(i, k, Nyp - 1)
          cu1(i, k, Nyp + 1) = cu1(i, k, Nyp)
        end do
      end do
    end if
  end if

  if (w_BC_Ymax == 1) then
    if ((rankY == NprocY - 1) .or. (.not. use_mpi)) then
      ! We are on process at the upper wall
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          cu3(i, k, Nyp) = cu3(i, k, Nyp - 1)
          cu3(i, k, Nyp + 1) = cu3(i, k, Nyp)
        end do
      end do
    end if
  end if

  if (u_BC_Ymin == 1) then
    if ((rankY == 0) .or. (.not. use_mpi)) then
      ! We are on a process at the bottom wall
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          cu1(i, k, 1) = cu1(i, k, 2)
          cu1(i, k, 0) = cu1(i, k, 1)
        end do
      end do
    end if
  end if

  if (w_BC_Ymin == 1) then
    if ((rankY == 0) .or. (.not. use_mpi)) then
      ! We are on a process at the bottom wall
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          cu3(i, k, 1) = cu3(i, k, 2)
          cu3(i, k, 0) = cu3(i, k, 1)
        end do
      end do
    end if
  end if

  return
end
