!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine ghost_chan_mpi
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  ! This subroutine is part of the MPI package for the channel flow
  ! Diablo package.
  ! Here, we define a set of ghost cells on each process
  ! the ghost cells contain information from the neighboring nodes
  ! and allow us to compute finite differences over the local gridpoints.
  ! We need to update the contents of the ghost cells at the start of
  ! each Runge-Kutta substep

  include 'header'

  integer i, j, k, n

  ! Define the arrays that will be used for data packing.  This makes the
  ! communication between processes more efficient by only requiring one
  ! send and recieve.
  ! The communication will be done in Fourier space, so these arrays should
  ! be complex arrays to match the velocity
  ! The size of the buffer array is 0:Nkx,0:twoNkz,# of variables
  complex(rkind) ocpack(0:Nxp - 1, 0:twoNkz, 4 + N_th)
  complex(rkind) icpack(0:Nxp - 1, 0:twoNkz, 4 + N_th)

  ! If we are using more than one processor, then we need to pass data

  if (NprocY > 1) then

    ! First, Pass data up the chain to higher ranked processes

    if (rankY == 0) then
      ! If we are the lowest ranked process, then we don't need to recieve
      ! data at the lower ghost cells, these will be filled with boundary
      ! condition information
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          ocpack(i, k, 1) = cu1(i, k, Nyp)
          ocpack(i, k, 2) = cu2(i, k, Nyp)
          ocpack(i, k, 3) = cu3(i, k, Nyp)
          ocpack(i, k, 4) = cp(i, k, Nyp)
          do n = 1, N_th
            ocpack(i, k, 4 + n) = cth(i, k, Nyp, n)
          end do
        end do
      end do
      ! Now, we have packed the data into a compact array, pass the data up
      call mpi_send(ocpack, (4 + N_th) * (Nxp) * (twoNkz + 1) &
                    , mpi_double_complex &
                    , rankY + 1, 1, mpi_comm_y, ierror)

      ! End if RANK=0
    else if (rankY < NprocY - 1) then
      ! Here, we are one of the middle processes and we need to pass data
      ! up and recieve data from below
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          ocpack(i, k, 1) = cu1(i, k, Nyp)
          ocpack(i, k, 2) = cu2(i, k, Nyp)
          ocpack(i, k, 3) = cu3(i, k, Nyp)
          ocpack(i, k, 4) = cp(i, k, Nyp)
          do n = 1, N_th
            ocpack(i, k, 4 + n) = cth(i, k, Nyp, n)
          end do
        end do
      end do
      ! Use MPI_SENDRECV since we need to recieve and send data
      call mpi_send(ocpack, (4 + N_th) * (Nxp) * (twoNkz + 1) &
                    , mpi_double_complex &
                    , rankY + 1, 1, mpi_comm_y, ierror)

      call mpi_recv(icpack, (4 + N_th) * (Nxp) * (twoNkz + 1) &
                    , mpi_double_complex &
                    , rankY - 1, 1, mpi_comm_y, status, ierror)
      ! Now, unpack the data that we have recieved
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          cu1(i, k, 1) = icpack(i, k, 1)
          cu2(i, k, 1) = icpack(i, k, 2)
          cu3(i, k, 1) = icpack(i, k, 3)
          cp(i, k, 1) = icpack(i, k, 4)
          do n = 1, N_th
            cth(i, k, 1, n) = icpack(i, k, 4 + n)
          end do
        end do
      end do

    else
      ! Otherwise, we must be the uppermost process with RANK=Nprocs-1
      ! Here, we need to recieve data from below, but don't need to send data up
      call mpi_recv(icpack, (4 + N_th) * (Nxp) * (twoNkz + 1) &
                    , mpi_double_complex &
                    , rankY - 1, 1, mpi_comm_y, status, ierror)
      ! Unpack the data that we have recieved
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          cu1(i, k, 1) = icpack(i, k, 1)
          cu2(i, k, 1) = icpack(i, k, 2)
          cu3(i, k, 1) = icpack(i, k, 3)
          cp(i, k, 1) = icpack(i, k, 4)
          do n = 1, N_th
            cth(i, k, 1, n) = icpack(i, k, 4 + n)
          end do
        end do
      end do
    end if

    ! AT this point we have passed data up the chain
    if (rankY == NprocY - 1) then
      ! If we are the higest ranked process, then we don't need to recieve
      ! data at the upper ghost cells, these will be filled with boundary
      ! condition information
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          ocpack(i, k, 1) = cu1(i, k, 2)
          ocpack(i, k, 2) = cu2(i, k, 2)
          ocpack(i, k, 3) = cu3(i, k, 2)
          ocpack(i, k, 4) = cp(i, k, 2)
          do n = 1, N_th
            ocpack(i, k, 4 + n) = cth(i, k, 2, n)
          end do
        end do
      end do
      ! Now, we have packed the data into a compact array, pass the data up
      call mpi_send(ocpack, (4 + N_th) * (Nxp) * (twoNkz + 1) &
                    , mpi_double_complex &
                    , rankY - 1, 3, mpi_comm_y, ierror)
    else if (rankY > 0) then
      ! Here, we are one of the middle processes and we need to pass data
      ! down and recieve data from above us
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          ocpack(i, k, 1) = cu1(i, k, 2)
          ocpack(i, k, 2) = cu2(i, k, 2)
          ocpack(i, k, 3) = cu3(i, k, 2)
          ocpack(i, k, 4) = cp(i, k, 2)
          do n = 1, N_th
            ocpack(i, k, 4 + n) = cth(i, k, 2, n)
          end do
        end do
      end do

      call mpi_send(ocpack, (4 + N_th) * (Nxp) * (twoNkz + 1) &
                    , mpi_double_complex &
                    , rankY - 1, 3, mpi_comm_y, ierror)

      call mpi_recv(icpack, (4 + N_th) * (Nxp) * (twoNkz + 1) &
                    , mpi_double_complex &
                    , rankY + 1, 3, mpi_comm_y, status, ierror)
      ! Now, unpack the data that we have recieved
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          cu1(i, k, Nyp + 1) = icpack(i, k, 1)
          cu2(i, k, Nyp + 1) = icpack(i, k, 2)
          cu3(i, k, Nyp + 1) = icpack(i, k, 3)
          cp(i, k, Nyp + 1) = icpack(i, k, 4)
          do n = 1, N_th
            cth(i, k, Nyp + 1, n) = icpack(i, k, 4 + n)
          end do
        end do
      end do
    else
      ! Here, we must be the lowest process (RANK=0) and we need to recieve
      ! data from above
      call mpi_recv(icpack, (4 + N_th) * (Nxp) * (twoNkz + 1) &
                    , mpi_double_complex &
                    , rankY + 1, 3, mpi_comm_y, status, ierror)
      ! Unpack the data that we have recieved
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          cu1(i, k, Nyp + 1) = icpack(i, k, 1)
          cu2(i, k, Nyp + 1) = icpack(i, k, 2)
          cu3(i, k, Nyp + 1) = icpack(i, k, 3)
          cp(i, k, Nyp + 1) = icpack(i, k, 4)
          do n = 1, N_th
            cth(i, k, Nyp + 1, n) = icpack(i, k, 4 + n)
          end do
        end do
      end do
    end if

  end if

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine ghost_les_mpi
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  ! This subroutine is part of the MPI package for the LES subroutine
  ! Here, after calculating the SGS viscosity, NU_T on each core,
  ! We need to share the ghost cells between neighboring processors

  include 'header'

  integer i, j, k, n

  ! Define the arrays that will be used for data packing.  This makes the
  ! communication between processes more efficient by only requiring one
  ! send and recieve.
  ! The communication will be done in Fourier space, so these arrays should
  ! be complex arrays to match the velocity
  ! The size of the buffer array is 0:Nxm1,0:NZP-1
  real(rkind) ocpack(0:Nxm1, 0:Nzp - 1)
  real(rkind) icpack(0:Nxm1, 0:Nzp - 1)

  ! If we are using more than one processor, then we need to pass data

  if (NprocY > 1) then

    ! First, Pass data up the chain to higher ranked processes

    if (rankY == 0) then
      ! If we are the lowest ranked process, then we don't need to recieve
      ! data at the lower ghost cells. Instead, set NU_T=0 at the lower wall
      do k = 0, Nzp - 1
        do i = 0, Nxm1
          nu_t(i, k, 1) = 0.d0
          nu_t(i, k, 2) = 0.d0
        end do
      end do

      ! Pass data up to the next process from GY(Nyp)
      do k = 0, Nzp - 1
        do i = 0, Nxm1
          ocpack(i, k) = nu_t(i, k, Nyp)
        end do
      end do
      ! Now, we have packed the data into a compact array, pass the data up
      call mpi_send(ocpack, (Nxm1 + 1) * (Nzp) &
                    , mpi_double_precision &
                    , rankY + 1, 1, mpi_comm_y, ierror)

      ! End if RANK=0
    else if (rankY < NprocY - 1) then
      ! Here, we are one of the middle processes and we need to pass data
      ! up and recieve data from below
      do k = 0, Nzp - 1
        do i = 0, Nxm1
          ocpack(i, k) = nu_t(i, k, Nyp)
        end do
      end do
      ! Use MPI_SENDRECV since we need to recieve and send data
      call mpi_send(ocpack, (Nxm1 + 1) * (Nzp) &
                    , mpi_double_precision &
                    , rankY + 1, 1, mpi_comm_y, ierror)

      call mpi_recv(icpack, (Nxm1 + 1) * (Nzp) &
                    , mpi_double_precision &
                    , rankY - 1, 1, mpi_comm_y, status, ierror)
      ! Now, unpack the data that we have recieved
      do k = 0, Nzp - 1
        do i = 0, Nxm1
          nu_t(i, k, 1) = icpack(i, k)
        end do
      end do

    else
      ! Otherwise, we must be the uppermost process with RANK=Nprocs-1
      ! Here, we need to recieve data from below, but don't need to send data up
      call mpi_recv(icpack, (Nxm1 + 1) * (Nzp) &
                    , mpi_double_precision &
                    , rankY - 1, 1, mpi_comm_y, status, ierror)
      ! Unpack the data that we have recieved
      do k = 0, Nzp - 1
        do i = 0, Nxm1
          nu_t(i, k, 1) = icpack(i, k)
        end do
      end do
    end if

    ! Now, we have hit the top process.  Set the BCs and pass data down

    if (rankY == NprocY - 1) then
      ! If we are the higest ranked process, then we don't need to recieve
      ! data at the upper ghost cells.
      ! Set NU_T=0 at the upper wall
      do k = 0, Nzp - 1
        do i = 0, Nxm1
          nu_t(i, k, Nyp) = 0.d0
          nu_t(i, k, Nyp + 1) = 0.d0
        end do
      end do

      ! Now, send data down the chain
      do k = 0, Nzp - 1
        do i = 0, Nxm1
          ocpack(i, k) = nu_t(i, k, 2)
        end do
      end do
      ! Now, we have packed the data into a compact array, pass the data up
      call mpi_send(ocpack, (Nxm1 + 1) * (Nzp) &
                    , mpi_double_precision &
                    , rankY - 1, 3, mpi_comm_y, ierror)
    else if (rankY > 0) then
      ! Here, we are one of the middle processes and we need to pass data
      ! down and recieve data from above us
      do k = 0, Nzp - 1
        do i = 0, Nxm1
          ocpack(i, k) = nu_t(i, k, 2)
        end do
      end do

      call mpi_send(ocpack, (Nxm1 + 1) * (Nzp) &
                    , mpi_double_precision &
                    , rankY - 1, 3, mpi_comm_y, ierror)

      call mpi_recv(icpack, (Nxm1 + 1) * (Nzp) &
                    , mpi_double_precision &
                    , rankY + 1, 3, mpi_comm_y, status, ierror)
      ! Now, unpack the data that we have recieved
      do k = 0, Nzp - 1
        do i = 0, Nxm1
          nu_t(i, k, Nyp + 1) = icpack(i, k)
        end do
      end do
    else
      ! Here, we must be the lowest process (RANK=0) and we need to recieve
      ! data from above
      call mpi_recv(icpack, (Nxm1 + 1) * (Nzp) &
                    , mpi_double_precision &
                    , rankY + 1, 3, mpi_comm_y, status, ierror)
      ! Unpack the data that we have recieved
      do k = 0, Nzp - 1
        do i = 0, Nxm1
          nu_t(i, k, Nyp + 1) = icpack(i, k)
        end do
      end do
    end if

  else
    ! Here, NprocY=1, so we just need to set the boundary values
    ! Set NU_T=0 at the lower wall
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        nu_t(i, k, 1) = 0.d0
        nu_t(i, k, 2) = 0.d0
      end do
    end do
    ! Set NU_T=0 at the upper wall
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        nu_t(i, k, Nyp) = 0.d0
        nu_t(i, k, Nyp + 1) = 0.d0
      end do
    end do

  end if

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine ghost_grid_mpi
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  ! This subroutine is part of the MPI package for the channel flow
  ! Diablo package.
  ! Here, we define a set of ghost cells on each process
  ! the ghost cells contain information from the neighboring nodes
  ! and allow us to compute finite differences over the local gridpoints.
  ! We need to update the contents of the ghost cells at the start of
  ! each Runge-Kutta substep

  include 'header'

  integer i, j, k, n

  real(rkind) ocpack(3), icpack(3)

  ! First, Pass data up the chain to higher ranked processes

  if (rankY == 0) then

    ! Set the lower ghost cells
    gyf(0) = 2.d0 * gyf(1) - gyf(2)
    gy(0) = 2.d0 * gy(1) - gy(2)

    ocpack(1) = gyf(Nyp - 1)
    ocpack(2) = gyf(Nyp)
    ocpack(3) = gy(Nyp)

    call mpi_send(ocpack, 3 &
                  , mpi_double_precision &
                  , rankY + 1, 1, mpi_comm_y, ierror)

    ! End if RANK=0
  else if (rankY < NprocY - 1) then
    ! Here, we are one of the middle processes and we need to pass data
    ! up and recieve data from below
    ocpack(1) = gyf(Nyp - 1)
    ocpack(2) = gyf(Nyp)
    ocpack(3) = gy(Nyp)

    call mpi_send(ocpack, 3 &
                  , mpi_double_precision &
                  , rankY + 1, 1, mpi_comm_y, ierror)

    call mpi_recv(icpack, 3 &
                  , mpi_double_precision &
                  , rankY - 1, 1, mpi_comm_y, status, ierror)
    ! Now, unpack the data that we have recieved
    gyf(0) = icpack(1)
    gyf(1) = icpack(2)
    gy(1) = icpack(3)

  else
    ! Otherwise, we must be the uppermost process with RANK=Nprocs-1
    ! Set the top ghost cell
    gyf(Nyp + 1) = 2.*gyf(Nyp) - gyf(Nypm1)

    ! Here, we need to recieve data from below, but don't need to send data up
    call mpi_recv(icpack, 3 &
                  , mpi_double_precision &
                  , rankY - 1, 1, mpi_comm_y, status, ierror)
    ! Now, unpack the data that we have recieved
    gyf(0) = icpack(1)
    gyf(1) = icpack(2)
    gy(1) = icpack(3)

  end if
  ! AT this point we have passed data up the chain

  if (rankY == NprocY - 1) then
    ! If we are the higest ranked process, then we don't need to recieve
    ! data at the upper ghost cells, these will be filled with boundary
    ! condition information

    ocpack(1) = gyf(2)
    ocpack(2) = gy(2)

    ! Now, we have packed the data into a compact array, pass the data up
    call mpi_send(ocpack, 3 &
                  , mpi_double_precision &
                  , rankY - 1, 3, mpi_comm_y, ierror)
  else if (rankY > 0) then
    ! Here, we are one of the middle processes and we need to pass data
    ! down and recieve data from above us

    ocpack(1) = gyf(2)
    ocpack(2) = gy(2)
    call mpi_send(ocpack, 3 &
                  , mpi_double_precision &
                  , rankY - 1, 3, mpi_comm_y, ierror)
    call mpi_recv(icpack, 3 &
                  , mpi_double_precision &
                  , rankY + 1, 3, mpi_comm_y, status, ierror)
    ! Now, unpack the data that we have recieved
    gyf(Nyp + 1) = icpack(1)
    gy(Nyp + 1) = icpack(2)

  else
    ! Here, we must be the lowest process (RANK=0) and we need to recieve
    ! data from above
    call mpi_recv(icpack, 3 &
                  , mpi_double_precision &
                  , rankY + 1, 3, mpi_comm_y, status, ierror)
    ! Unpack the data that we have recieved
    gyf(Nyp + 1) = icpack(1)
    gy(Nyp + 1) = icpack(2)

  end if

  return
end


!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine ghost_chan_mpi_j0
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  ! ghost_chan_mpi subroutine to update the j = 0 cell
  ! Needed for les output and epsilon calculation

  include 'header'

  integer i, j, k, n

  ! Define the arrays that will be used for data packing.  This makes the
  ! communication between processes more efficient by only requiring one
  ! send and recieve.
  ! The communication will be done in Fourier space, so these arrays should
  ! be complex arrays to match the velocity
  ! The size of the buffer array is 0:Nkx,0:twoNkz,# of variables
  complex(rkind) ocpack(0:Nxp - 1, 0:twoNkz, 4 + N_th)
  complex(rkind) icpack(0:Nxp - 1, 0:twoNkz, 4 + N_th)

  ! If we are using more than one processor, then we need to pass data

  if (NprocY > 1) then

    if (rankY == 0) then
      ! If we are the lowest ranked process, then we don't need to recieve
      ! data at the lower ghost cells, these will be filled with boundary
      ! condition information
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          ocpack(i, k, 1) = cu1(i, k, Nyp - 1)
          ocpack(i, k, 2) = cu2(i, k, Nyp - 1)
          ocpack(i, k, 3) = cu3(i, k, Nyp - 1)
          ocpack(i, k, 4) = cp(i, k, Nyp - 1)
          do n = 1, N_th
            ocpack(i, k, 4 + n) = cth(i, k, Nyp - 1, n)
          end do
        end do
      end do
      ! Now, we have packed the data into a compact array, pass the data up
      call mpi_send(ocpack, (4 + N_th) * (Nxp) * (twoNkz + 1) &
                    , mpi_double_complex &
                    , rankY + 1, 1, mpi_comm_y, ierror)

      ! End if RANK=0
    else if (rankY < NprocY - 1) then
      ! Here, we are one of the middle processes and we need to pass data
      ! up and recieve data from below
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          ocpack(i, k, 1) = cu1(i, k, Nyp - 1)
          ocpack(i, k, 2) = cu2(i, k, Nyp - 1)
          ocpack(i, k, 3) = cu3(i, k, Nyp - 1)
          ocpack(i, k, 4) = cp(i, k, Nyp - 1)
          do n = 1, N_th
            ocpack(i, k, 4 + n) = cth(i, k, Nyp - 1, n)
          end do
        end do
      end do
      ! Use MPI_SENDRECV since we need to recieve and send data
      call mpi_send(ocpack, (4 + N_th) * (Nxp) * (twoNkz + 1) &
                    , mpi_double_complex &
                    , rankY + 1, 1, mpi_comm_y, ierror)

      call mpi_recv(icpack, (4 + N_th) * (Nxp) * (twoNkz + 1) &
                    , mpi_double_complex &
                    , rankY - 1, 1, mpi_comm_y, status, ierror)
      ! Now, unpack the data that we have recieved
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          cu1(i, k, 0) = icpack(i, k, 1)
          cu2(i, k, 0) = icpack(i, k, 2)
          cu3(i, k, 0) = icpack(i, k, 3)
          cp(i, k, 0) = icpack(i, k, 4)
          do n = 1, N_th
            cth(i, k, 0, n) = icpack(i, k, 4 + n)
          end do
        end do
      end do

    else
      ! Otherwise, we must be the uppermost process with RANK=Nprocs-1
      ! Here, we need to recieve data from below, but don't need to send data up
      call mpi_recv(icpack, (4 + N_th) * (Nxp) * (twoNkz + 1) &
                    , mpi_double_complex &
                    , rankY - 1, 1, mpi_comm_y, status, ierror)
      ! Unpack the data that we have recieved
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          cu1(i, k, 0) = icpack(i, k, 1)
          cu2(i, k, 0) = icpack(i, k, 2)
          cu3(i, k, 0) = icpack(i, k, 3)
          cp(i, k, 0) = icpack(i, k, 4)
          do n = 1, N_th
            cth(i, k, 0, n) = icpack(i, k, 4 + n)
          end do
        end do
      end do

    end if

  end if

  return
end


!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine ghost_CSij2_mpi
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  ! This subroutine is part of the MPI package for the LES subroutine
  ! to compute the strain
  ! Share down the CSij2 Nyp + 1 ghost cell between neighboring processors

  include 'header'
  include 'header_les'

  integer i, j, k, n

  ! Define the arrays that will be used for data packing.  This makes the
  ! communication between processes more efficient by only requiring one
  ! send and recieve.
  ! The communication will be done in Fourier space, so these arrays should
  ! be complex arrays to match the velocity
  ! The size of the buffer array is 0:Nxm1,0:NZP-1
  complex(rkind) ocpack(0:Nxp - 1, 0:twoNkz)
  complex(rkind) icpack(0:Nxp - 1, 0:twoNkz)

  ! If we are using more than one processor, then we need to pass data

  if (NprocY > 1) then

    ! Set the BCs and pass data down

    if (rankY == NprocY - 1) then
      ! If we are the higest ranked process, then we don't need to recieve
      ! data at the upper ghost cells.

      ! Set CSij2 (cu2(i, k, Nyp + 2) - cu2(i, k, Nyp + 1)) / dyf(Nyp + 1)
      ! We don't need CSij2 at the boundaries (dictates CSij4' & 6' only)
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          CSij2(i, k, Nyp + 1) = -10000.d0
        end do
      end do

      ! Now, send data down the chain
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          ocpack(i, k) = CSij2(i, k, 2)
        end do
      end do
      ! Now, we have packed the data into a compact array, pass the data up
      call mpi_send(ocpack, Nxp * (twoNkz + 1) &
                    , mpi_double_complex &
                    , rankY - 1, 3, mpi_comm_y, ierror)
    else if (rankY > 0) then
      ! Here, we are one of the middle processes and we need to pass data
      ! down and recieve data from above us
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          ocpack(i, k) = CSij2(i, k, 2)
        end do
      end do

      call mpi_send(ocpack, Nxp * (twoNkz + 1) &
                    , mpi_double_complex &
                    , rankY - 1, 3, mpi_comm_y, ierror)

      call mpi_recv(icpack, Nxp * (twoNkz + 1) &
                    , mpi_double_complex &
                    , rankY + 1, 3, mpi_comm_y, status, ierror)
      ! Now, unpack the data that we have recieved
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          CSij2(i, k, Nyp + 1) = icpack(i, k)
        end do
      end do
    else
      ! Here, we must be the lowest process (RANK=0) and we need to receive
      ! data from above
      call mpi_recv(icpack, (Nxm1 + 1) * (Nzp) &
                    , mpi_double_complex &
                    , rankY + 1, 3, mpi_comm_y, status, ierror)
      ! Unpack the data that we have recieved
      do k = 0, twoNkz
        do i = 0, Nxp - 1
          CSij2(i, k, Nyp + 1) = icpack(i, k)
        end do
      end do
    end if

  else
    ! Here, NprocY=1, so we just need to set the boundary values
    ! Set CSij2 (cu2(i, k, Nyp + 2) - cu2(i, k, Nyp + 1)) / dyf(Nyp + 1)
    do k = 0, twoNkz
      do i = 0, Nxp - 1
        CSij2(i, k, Nyp + 1) = -10000.d0
      end do
    end do

  end if

  return
end


!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine thomas_forward_real_mpi(a, b, c, g, iNyp, inx)
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  ! This version solves for one full row, then passes the data
  ! This subroutine performs the backward sweep of the Thomas algorithm
  ! Thomas algorithm solves Ax=b for tridiagonal A
  ! The RHS vector and solution are real
  ! Input lower, main, and upper diagonals, ld, md, ud, and rhs x
  ! Returns solution in x
  ! The indexing should be done by ROW, ie.
  ! [ b1  c1   0   0   0 ...
  ! [ a2  b2  c2   0   0 ...
  ! [  0  a3  b3   c3  0 ...

  include 'header'

  integer i, j, inx, iNyp
  real(rkind) a(0:inx - 1, 0:iNyp + 1), b(0:inx - 1, 0:iNyp + 1), c(0:inx - 1, 0:iNyp + 1) &
    , g(0:inx - 1, 0:iNyp + 1)

  real(rkind) ocpack(0:inx - 1, 4), icpack(0:inx - 1, 4)

  if (rankY /= 0) then
    ! If we aren't the lowest process, then wait for data
    call mpi_recv(ocpack, 4 * inx, mpi_double_precision, rankY - 1, 12 &
                  , mpi_comm_y, status, ierror)
    ! Unpack the data
    do i = 0, inx - 1
      a(i, 1) = ocpack(i, 1)
      b(i, 1) = ocpack(i, 2)
      c(i, 1) = ocpack(i, 3)
      g(i, 1) = ocpack(i, 4)
    end do
    ! If we aren't the lowest process, start at J=2
    do j = 2, iNyp
      do i = 0, inx - 1
        a(i, j) = -a(i, j) / b(i, j - 1)
        b(i, j) = b(i, j) + a(i, j) * c(i, j - 1)
        g(i, j) = g(i, j) + a(i, j) * g(i, j - 1)
      end do
    end do
  else
    ! Here, we are the lowest process, start solving at J=1
    do j = 1, iNyp
      do i = 0, inx - 1
        a(i, j) = -a(i, j) / b(i, j - 1)
        b(i, j) = b(i, j) + a(i, j) * c(i, j - 1)
        g(i, j) = g(i, j) + a(i, j) * g(i, j - 1)
      end do
    end do
  end if

  if (rankY /= NprocY - 1) then
    do i = 0, inx - 1
      icpack(i, 1) = a(i, iNyp)
      icpack(i, 2) = b(i, iNyp)
      icpack(i, 3) = c(i, iNyp)
      icpack(i, 4) = g(i, iNyp)
    end do
    call mpi_send(icpack, 4 * inx, mpi_double_precision, rankY + 1, 12 &
                  , mpi_comm_y, ierror)
  else
    ! Here, we are at the upper process, so solve one more row containing
    ! the boundary conditions
    j = iNyp + 1
    do i = 0, inx - 1
      a(i, j) = -a(i, j) / b(i, j - 1)
      b(i, j) = b(i, j) + a(i, j) * c(i, j - 1)
      g(i, j) = g(i, j) + a(i, j) * g(i, j - 1)
    end do
  end if

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|------
subroutine thomas_forward_complex_mpi(a, b, c, g, iNyp, inx)
  !----*|--.---------.---------.---------.---------.---------.---------.-|-----|
  ! This subroutine performs the backward sweep of the Thomas algorithm
  ! Thomas algorithm solves Ax=b for tridiagonal A
  ! The RHS vector and solution are real
  ! Input lower, main, and upper diagonals, ld, md, ud, and rhs x
  ! Returns solution in x
  ! The indexing should be done by ROW, ie.
  ! [ b1  c1   0   0   0 ...
  ! [ a2  b2  c2   0   0 ...
  ! [  0  a3  b3   c3  0 ...

  include 'header'

  integer i, j, inx, iNyp
  real(rkind) a(0:inx, 0:iNyp + 1), b(0:inx, 0:iNyp + 1), c(0:inx, 0:iNyp + 1)
  complex(rkind) g(0:inx, 0:iNyp + 1)

  complex(rkind) ocpack(4), icpack(4)

  do i = 0, inx

    if (rankY /= 0) then
      ! If we aren't the lowest process, then wait for data
      call mpi_recv(ocpack, 4, mpi_double_complex, rankY - 1, 13 &
                    , mpi_comm_y, status, ierror)
      ! Unpack the data
      a(i, 1) = real(ocpack(1))
      b(i, 1) = real(ocpack(2))
      c(i, 1) = real(ocpack(3))
      g(i, 1) = ocpack(4)
    end if

    do j = 2, iNyp
      a(i, j) = -a(i, j) / b(i, j - 1)
      b(i, j) = b(i, j) + a(i, j) * c(i, j - 1)
      g(i, j) = g(i, j) + a(i, j) * g(i, j - 1)
    end do

    if (rankY /= NprocY - 1) then
      icpack(1) = a(i, iNyp)
      icpack(2) = b(i, iNyp)
      icpack(3) = c(i, iNyp)
      icpack(4) = g(i, iNyp)
      call mpi_send(icpack, 4, mpi_double_complex, rankY + 1, 13 &
                    , mpi_comm_y, ierror)
    end if

  end do

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|------
subroutine thomas_backward_real_mpi(a, b, c, g, iNyp, inx)
  !----*|--.---------.---------.---------.---------.---------.---------.-|------
  ! This subroutine performs the backward sweep of the Thomas algorithm
  ! Thomas algorithm solves Ax=b for tridiagonal A
  ! The RHS vector and solution are real
  ! Input lower, main, and upper diagonals, ld, md, ud, and rhs x
  ! Returns solution in x
  ! The indexing should be done by ROW, ie.
  ! ￼￼￼C[b1c1 0 C[a2 b2 c2 C[0a3b3
  ! 0 0... 0 0... c30...

  include 'header'

  integer i, j, inx, iNyp
  real(rkind) a(0:inx - 1, 0:iNyp + 1), b(0:inx - 1, 0:iNyp + 1), c(0:inx - 1, 0:iNyp + 1) &
    , g(0:inx - 1, 0:iNyp + 1)
  real(rkind) icpack(1), ocpack(1)
  do i = 0, inx - 1
    if (rankY /= NprocY - 1) then
      ! If we aren’t the highest process, then wait for data
      call mpi_recv(ocpack, 1, mpi_double_precision, rankY + 1, 10 &
                    , mpi_comm_y, status, ierror)
      g(i, iNyp + 1) = ocpack(1)
    else
      ! Else, if we are the highest process, compute the solution at j=INyp
      g(i, iNyp + 1) = g(i, iNyp + 1) / b(i, iNyp + 1)
    end if
    ! All processes solve from INyp..1
    do j = iNyp, 0, -1
      g(i, j) = (g(i, j) - c(i, j) * g(i, j + 1)) / b(i, j)
    end do
    if (rankY /= 0) then
      icpack(1) = g(i, 2)
      call mpi_send(icpack, 1, mpi_double_precision, rankY - 1, 10 &
                    , mpi_comm_y, ierror)
    end if
  end do
  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|------
subroutine thomas_backward_complex_mpi(a, b, c, g, iNyp, inx)
  !----*|--.---------.---------.---------.---------.---------.---------.-|-----|
  ! This subroutine performs the backward sweep of the Thomas algorithm
  ! Thomas algorithm solves Ax=b for tridiagonal A
  ! The RHS vector and solution are real
  ! Input lower, main, and upper diagonals, ld, md, ud, and rhs x
  ! Returns solution in x
  ! The indexing should be done by ROW, ie.
  ! [ b1  c1   0   0   0 ...
  ! [ a2  b2  c2   0   0 ...
  ! [  0  a3  b3   c3  0 ...

  include 'header'

  integer i, j, inx, iNyp
  real(rkind) a(0:inx, 0:iNyp + 1), b(0:inx, 0:iNyp + 1), c(0:inx, 0:iNyp + 1)
  complex(rkind) g(0:inx, 0:iNyp + 1)

  do i = 0, inx

    if (rankY /= NprocY - 1) then
      ! If we aren't the highest process, then wait for data
      call mpi_recv(g(i, iNyp + 1), 1, mpi_double_complex, rankY + 1, 11 &
                    , mpi_comm_y, status, ierror)
      j = iNyp
      g(i, j) = (g(i, j) - c(i, j) * g(i, j + 1)) / b(i, j)
    else
      ! Else, if we are the highest process, then compute the solution at j=INyp
      g(i, iNyp) = g(i, iNyp) / b(i, iNyp)
    end if

    do j = iNyp - 1, 1, -1
      g(i, j) = (g(i, j) - c(i, j) * g(i, j + 1)) / b(i, j)
    end do

    if (rankY /= 0) then
      call mpi_send(g(i, 2), 1, mpi_double_complex, rankY - 1, 11 &
                    , mpi_comm_y, ierror)
    end if

  end do

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|------
subroutine init_mpi
  !----*|--.---------.---------.---------.---------.---------.---------.-|-----|
  include 'header'

  integer bl(2), disp(2), types(2)

  integer iprocs, type1, comm_cart
  integer dims(2), perdim(2)
  logical mflag(2)

  integer i, j, k, xi, zi

  complex(rkind) tmp(0:Nx / 2, 0:Nzp + 1, 0:Nyp + 1)
  character(len=55) fname

  integer, parameter :: fftw_forward = -1
  integer, parameter :: fftw_backward = 1
  integer, parameter :: fftw_estimate = 0
  integer, parameter :: fftw_measure = 1
  integer, parameter :: fftw_out_of_place = 0
  integer, parameter :: fftw_in_place = 8
  integer, parameter :: fftw_use_wisdom = 16
  integer, parameter :: fftw_threadsafe = 128

  ! This subroutine initializes all mpi variables

  call mpi_init(ierror)
  call mpi_comm_size(mpi_comm_world, iprocs, ierror)
  call mpi_comm_rank(mpi_comm_world, rank, ierror)

  if (Nprocs /= iprocs) then
    if (rank == 0) write (*, '("Error: compiled with ", I10 " cores, &
                                running with ", I10, " cores.")') Nprocs, iprocs
    call mpi_finalize(ierror)
    stop
  end if

  if (mod(Nprocs, NprocY) /= 0) then
    if (rank == 0) write (*, '("Error. Nprocs is not a multiple of NprocZ!")')
    call mpi_finalize(ierror)
    stop
  end if

  dims(2) = NprocY
  dims(1) = NprocZ
  mflag(:) = .false.

  call mpi_cart_create(mpi_comm_world, 2, dims, perdim, .false., &
                       comm_cart, ierror)
  ! In PERDIM I put the information for the remain_dims
  perdim = (/0, 1/)
  call mpi_cart_sub(comm_cart, perdim, mpi_comm_y, ierror)
  perdim = (/1, 0/)
  call mpi_cart_sub(comm_cart, perdim, mpi_comm_z, ierror)

  call mpi_comm_rank(mpi_comm_y, rankY, ierror)
  call mpi_comm_rank(mpi_comm_z, rankZ, ierror)

  if (rank == 0) &
    write (*, '("MPI Initialised with ", I10, " processors")') Nprocs
  call mpi_barrier(mpi_comm_world, ierror)
  if (verbosity > 2 .and. rank == 0) &
    write (*, '("Rank, rankZ, rankY: " 3I10)') rank, rankY, rankZ


  ! ------------------------------
  !  Define datatypes
  ! ------------------------------

  ! Box full x to z(1)
  call mpi_type_vector(Nzp, Nxp, Nx / 2 + 1, &
                       mpi_double_complex, type1, ierror)
  call mpi_type_commit(type1, ierror)

  bl(1:2) = (/1, 1/)
  disp(1:2) = (/0, Nxp * 16/)
  types = (/type1, mpi_ub/)

  call mpi_type_struct(2, bl, disp, types, xy2zy_1, ierror)
  call mpi_type_commit(xy2zy_1, ierror)
  call mpi_type_free(type1, ierror)

  ! Box full x to z(2)
  call mpi_type_vector(Nzp, Nxp, Nxp + 1, &
                       mpi_double_complex, type1, ierror)
  call mpi_type_commit(type1, ierror)

  bl(1:2) = (/1, 1/)
  disp(1:2) = (/0, Nzp * (Nxp + 1) * 16/)
  types = (/type1, mpi_ub/)

  call mpi_type_struct(2, bl, disp, types, xy2zy_2, ierror)
  call mpi_type_commit(xy2zy_2, ierror)
  call mpi_type_free(type1, ierror)


  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|------
subroutine init_chan_mpi
  !----*|--.---------.---------.---------.---------.---------.---------.-|-----|
  ! Initialize any constants here
  include 'header'

  integer j, n

  if (rank == 0) then
    write (*, *)
    write (*, *) '             ****** In Init_Chan_MPI ******'
    write (*, *)
  end if

  pi = 4.d0 * atan(1.d0)

  ! Set the upper and lower bounds for timestepping
  if (rankY == 0) then
    jend = Nyp
    if (u_BC_Ymin == 0) then
      jstart = 2
    else if (u_BC_Ymin == 1) then
      jstart = 1
    else
      jstart = 2
    end if
    ! Now, set the indexing for the scalar equations
    do n = 1, N_th
      jend_th(n) = Nyp
      if (th_BC_Ymin(n) == 0) then
        jstart_th(n) = 2
      else if (th_BC_Ymin(n) == 1) then
        jstart_th(n) = 1
      else
        jstart_th(n) = 2
      end if
    end do
  else if (rankY == NprocY - 1) then
    jstart = 2
    if (u_BC_Ymax == 0) then
      jend = Nyp - 1
    else if (u_BC_Ymax == 1) then
      jend = Nyp
    else
      jend = Nyp - 1
    end if

    ! Set the upper and lower limits of timestepping of the scalar equations
    do n = 1, N_th
      jstart_th(n) = 2
      if (th_BC_Ymax(n) == 0) then
        jend_th(n) = Nyp - 1
      else if (th_BC_Ymax(n) == 1) then
        jend_th(n) = Nyp
      else
        jend_th(n) = Nyp - 1
      end if
    end do

  else
    ! Here, we are on a middle process
    jstart = 2
    jend = Nyp
    do n = 1, N_th
      jstart_th(n) = 2
      jend_th(n) = Nyp
    end do
  end if

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine apply_BC_th_mpi(matl, matd, matu, vec, n)
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  ! This subroutine applies the boundary conditions to the
  ! scalar fields prior to the implicit solve
  include 'header'

  integer n
  ! We first need to check to see which processor we are, if we are
  ! the upper or lowermost process, then apply boundary conditions
  if (rankY == 0) then
    ! If we have the lowest plane, apply the boundary conditions
    call apply_BC_th_lower(matl, matd, matu, vec, n)
  end if
  if (rankY == NprocY - 1) then
    ! If we have the upper plane, apply the boundary conditions
    call apply_BC_th_upper(matl, matd, matu, vec, n)
  end if
  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine apply_BC_th_mpi_c(matl_c, matd_c, matu_c, vec_c, n)
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  ! This subroutine applies the boundary conditions to the
  ! scalar fields prior to the implicit solve
  include 'header'

  integer n
  ! We first need to check to see which processor we are, if we are
  ! the upper or lowermost process, then apply boundary conditions
  if (rank == 0) then
    ! If we have the lowest plane, apply the boundary conditions
    call apply_BC_th_lower_c(matl_c, matd_c, matu_c, vec_c, n)
  end if
  if (rank == Nprocs - 1) then
    ! If we have the upper plane, apply the boundary conditions
    call apply_BC_th_upper_c(matl_c, matd_c, matu_c, vec_c, n)
  end if
  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine apply_BC_u2_mpi(matl, matd, matu, vec)
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  ! This subroutine applies the boundary conditions to the
  ! velocity field prior to the implicit solve
  ! Note, MATL, MATD, etc. are dimensioned in header
  include 'header'

  ! We first need to check to see which processor we are, if we are
  ! the upper or lowermost process, then apply boundary conditions
  if (rankY == 0) then
    ! If we have the lowest plane, apply the boundary conditions
    call apply_BC_2_lower(matl, matd, matu, vec)
  end if
  if (rankY == NprocY - 1) then
    ! If we have the highest plane, apply the boundary conditions
    call apply_BC_2_upper(matl, matd, matu, vec)
  end if
  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine apply_BC_u2_mpi_c(matl_c, matd_c, matu_c, vec_c)
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  ! This subroutine applies the boundary conditions to the
  ! velocity field prior to the implicit solve
  ! Note, MATL, MATD, etc. are dimensioned in header
  include 'header'

  ! We first need to check to see which processor we are, if we are
  ! the upper or lowermost process, then apply boundary conditions
  if (rank == 0) then
    ! If we have the lowest plane, apply the boundary conditions
    call apply_BC_2_lower_c(matl_c, matd_c, matu_c, vec_c)
  end if
  if (rank == Nprocs - 1) then
    ! If we have the highest plane, apply the boundary conditions
    call apply_BC_2_upper_c(matl_c, matd_c, matu_c, vec_c)
  end if
  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine apply_BC_u1_mpi(matl, matd, matu, vec)
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  ! This subroutine applies the boundary conditions to the
  ! velocity field prior to the implicit solve
  ! Note, MATL, MATD, etc. are dimensioned in header
  include 'header'

  ! We first need to check to see which processor we are, if we are
  ! the upper or lowermost process, then apply boundary conditions
  if (rankY == 0) then
    ! If we have the lowest plane, apply the boundary conditions
    call apply_BC_1_lower(matl, matd, matu, vec)
  end if
  if (rankY == NprocY - 1) then
    ! If we have the highest plane, apply the boundary conditions
    call apply_BC_1_upper(matl, matd, matu, vec)
  end if
  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine apply_BC_u1_mpi_c(matl_c, matd_c, matu_c, vec_c)
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  ! This subroutine applies the boundary conditions to the
  ! velocity field prior to the implicit solve
  ! Note, MATL, MATD, etc. are dimensioned in header
  include 'header'

  ! We first need to check to see which processor we are, if we are
  ! the upper or lowermost process, then apply boundary conditions
  if (rank == 0) then
    ! If we have the lowest plane, apply the boundary conditions
    call apply_BC_1_lower_c(matl_c, matd_c, matu_c, vec_c)
  end if
  if (rank == Nprocs - 1) then
    ! If we have the highest plane, apply the boundary conditions
    call apply_BC_1_upper_c(matl_c, matd_c, matu_c, vec_c)
  end if
  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine apply_BC_u3_mpi(matl, matd, matu, vec)
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  ! This subroutine applies the boundary conditions to the
  ! velocity field prior to the implicit solve
  ! Note, MATL, MATD, etc. are dimensioned in header
  include 'header'

  ! We first need to check to see which processor we are, if we are
  ! the upper or lowermost process, then apply boundary conditions
  if (rankY == 0) then
    ! If we have the lowest plane, apply the boundary conditions
    call apply_BC_3_lower(matl, matd, matu, vec)
  end if
  if (rankY == NprocY - 1) then
    ! If we have the highest plane, apply the boundary conditions
    call apply_BC_3_upper(matl, matd, matu, vec)
  end if
  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine apply_BC_u3_mpi_c(matl_c, matd_c, matu_c, vec_c)
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  ! This subroutine applies the boundary conditions to the
  ! velocity field prior to the implicit solve
  ! Note, MATL, MATD, etc. are dimensioned in header
  include 'header'

  ! We first need to check to see which processor we are, if we are
  ! the upper or lowermost process, then apply boundary conditions
  if (rank == 0) then
    ! If we have the lowest plane, apply the boundary conditions
    call apply_BC_3_lower_c(matl_c, matd_c, matu_c, vec_c)
  end if
  if (rank == Nprocs - 1) then
    ! If we have the highest plane, apply the boundary conditions
    call apply_BC_3_upper_c(matl_c, matd_c, matu_c, vec_c)
  end if
  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine apply_BC_rem_div_mpi(matl_c, matd_c, matu_c, vec_c, k)
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  ! This subroutine applies the boundary conditions for the Poisson Eq.
  ! Note, MATL, MATD, etc. are dimensioned in header
  include 'header'

  integer i, k
  ! We first need to check to see which processor we are, if we are
  ! the upper or lowermost process, then apply boundary conditions
  ! Apply the boundary conditions
  if (rankY == 0) then
    do i = 0, Nxp - 1
      ! Use homogeneous dirichlet BCS for kx=kz=0 component at bottom wall
      if ((k == 0) .and. (i == 0) .and. (rankZ == 0)) then
        ! Otherwise the matrix will be singular
        ! Use homogeneous dirichlet BCS for kx=kz=0 component at bottom wall
        matl_c(i, 1) = 0.
        matd_c(i, 1) = 1.
        matu_c(i, 1) = 0.
        vec_c(i, 1) = (0., 0.)
      else
        ! Use Dirichlet boundary conditions, dp/dz=0 at walls
        matl_c(i, 1) = 0.
        matd_c(i, 1) = 1.
        matu_c(i, 1) = -1.
        vec_c(i, 1) = (0., 0.)
      end if
    end do
  end if
  ! Apply the boundary conditions
  if (rankY == NprocY - 1) then
    do i = 0, Nxp - 1
      matl_c(i, Nyp) = 1.
      matd_c(i, Nyp) = -1.
      matu_c(i, Nyp) = 0.
      vec_c(i, Nyp) = (0., 0.)
    end do
  end if

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine apply_BC_poisson_mpi(matl_c, matd_c, matu_c, vec_c, k)
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  ! This subroutine applies the boundary conditions for the Poisson Eq.
  ! Note, MATL, MATD, etc. are dimensioned in header
  include 'header'

  integer i, k
  ! We first need to check to see which processor we are, if we are
  ! the upper or lowermost process, then apply boundary conditions
  ! Use dirichlet boundary condition at the lower wall to
  ! prevent the tridiagonal matrix from becomming singular for i,k=0
  if (rankY == 0) then
    do i = 0, Nxp - 1
      if ((i == 0) .and. (k == 0) .and. (rankZ == 0)) then
        matd_c(i, 1) = 1.
        matu_c(i, 1) = 0.
        vec_c(i, 1) = (0., 0.)
      else
        ! Here, apply Neumann boundary conditions (dp/dz=0) at the walls
        matd_c(i, 1) = 1.
        matu_c(i, 1) = -1.
        vec_c(i, 1) = (0., 0.)
      end if
    end do
  end if
  ! Use dirichlet boundary condition at the lower wall to
  ! prevent the tridiagonal matrix from becomming singular for i,k=0
  if (rankY == NprocY - 1) then
    do i = 0, Nxp - 1
      matd_c(i, Nyp) = -1.
      matl_c(i, Nyp) = 1.
      vec_c(i, Nyp) = (0., 0.)
    end do
  end if

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine apply_BC_vel_mpi
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  ! This subroutine applies the boundary conditions for the Poisson Eq.
  ! Note, MATL, MATD, etc. are dimensioned in header
  include 'header'

  ! Apply Boundary conditions to velocity field
  if (rankY == 0) then
    call apply_BC_vel_lower
  end if
  if (rankY == NprocY - 1) then
    call apply_BC_vel_upper
  end if

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine apply_BC_vel_phys_mpi
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  ! This subroutine applies the boundary conditions for the Poisson Eq.
  ! Note, MATL, MATD, etc. are dimensioned in header
  include 'header'

  ! Apply Boundary conditions to velocity field
  if (rankY == 0) then
    call apply_BC_vel_phys_lower
  end if
  if (rankY == NprocY - 1) then
    call apply_BC_vel_phys_upper
  end if

  return
end


!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine get_minimum_mpi(val)
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|

  include 'header'

  real(rkind) val, vmin

  call mpi_allreduce(val, vmin, 1, mpi_double_precision, &
                     mpi_min, mpi_comm_world, ierror)

  val = vmin

  return

end

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine get_maximum_mpi(val)
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|

  include 'header'

  real(rkind) val, vmax

  call mpi_allreduce(val, vmax, 1, mpi_double_precision, &
                     mpi_max, mpi_comm_world, ierror)

  val = vmax

  return

end

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine fft_xz_to_fourier(v, vv, jmin, jmax)
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|

  include 'header'

  real(rkind) v(0:Nx + 1, 0:Nzp + 1, 0:Nyp + 1)
  complex(rkind) vv(0:Nxp, 0:Nz + 1, 0:Nyp + 1)
  complex(rkind) tmp(0:Nx / 2, 0:Nzp + 1, 0:Nyp + 1)

  integer i, j, k
  integer jmin, jmax


  ! FFT in X
  do j = jmin, jmax
    call rfftwnd_f77_real_to_complex(fftw_x_to_f_plan, Nzp, &
                                     v(0, 0, j), 1, Nx + 2, tmp(0, 0, j), 1, Nx / 2 + 1)
    do k = 0, Nzp - 1
      do i = 0, Nkx
        tmp(i, k, j) = tmp(i, k, j) / Nx
      end do
      do i = Nkx + 1, Nx / 2
        tmp(i, k, j) = cmplx(0.d0, 0.d0)
      end do
    end do
  end do


  do j = jmin, jmax
    call mpi_alltoall(tmp(0, 0, j), 1, xy2zy_1, &
                      vv(0, 0, j), 1, xy2zy_2, mpi_comm_z, ierror)
  end do


  ! FFT in Z
  do j = jmin, jmax
    call fftwnd_f77(fftw_z_to_f_plan, Nxp, &
                    vv(0, 0, j), Nxp + 1, 1, vv(0, 0, j), Nxp + 1, 1)
    do k = 0, NkZ
      do i = 0, Nxp
        vv(i, k, j) = vv(i, k, j) / Nz
      end do
    end do

    do k = 1, NkZ
      do i = 0, Nxp
        vv(i, NkZ + k, j) = vv(i, Nz - 1 + k - NkZ, j) / Nz
      end do
    end do
  end do


end subroutine

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine fft_xz_to_physical(vv, v, jmin, jmax)
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|

  include 'header'

  real(rkind) v(0:Nx + 1, 0:Nzp + 1, 0:Nyp + 1)
  complex(rkind) vv(0:Nxp, 0:Nz + 1, 0:Nyp + 1)
  complex(rkind) tmp(0:Nx / 2, 0:Nzp + 1, 0:Nyp + 1)

  integer i, j, k
  integer jmin, jmax

  ! FFT in Z
  do j = jmin, jmax
    ! UNPACK
    do k = NkZ, 1, -1
      do i = 0, Nxp
        vv(i, Nz - 1 + k - NkZ, j) = vv(i, NkZ + k, j)
      end do
    end do
    do i = 0, Nxp
      do k = NkZ + 1, Nz - NkZ - 1
        vv(i, k, j) = cmplx(0.d0, 0.d0)
      end do

    end do
    call fftwnd_f77(fftw_z_to_p_plan, Nxp, &
                    vv(0, 0, j), Nxp + 1, 1, vv(0, 0, j), Nxp + 1, 1)
  end do

  do j = jmin, jmax
    call mpi_alltoall(vv(0, 0, j), 1, xy2zy_2, &
                      tmp(0, 0, j), 1, xy2zy_1, mpi_comm_z, ierror)
  end do

  ! FFT in X
  do j = jmin, jmax
    call rfftwnd_f77_complex_to_real(fftw_x_to_p_plan, Nzp, &
                                     tmp(0, 0, j), 1, Nx / 2 + 1, v(0, 0, j), 1, Nx + 2)
  end do

end subroutine

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine integrate_y_var(var, res, comm)
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|

  include 'header'

  integer j, comm
  real(rkind) var(0:Nyp + 1), res

  ! Integrate the instantaneous mean profile numerically at GY points
  res = 0.
  if (use_mpi) then
    do j = 2, Nyp
      res = res + 0.5 * (var(j) + var(j - 1)) * dy(j)
    end do
    call mpi_allreduce(mpi_in_place, res, 1, &
                       mpi_double_precision, mpi_sum, comm, ierror)
  else
    do j = 2, Nyp
      res = res + 0.5 * (var(j) + var(j - 1)) * dy(j)
    end do
  end if
  res = res / Ly

end subroutine

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine integrate_z_var(var, res, comm)
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|

  include 'header'

  integer i, k, j, comm
  real(rkind) var(0:Nx + 1, 0:Nzp + 1, 0:Nyp + 1), res(0:Nxm1, 1:Nyp)

  ! Integrate the instantaneous mean profile numerically at GY points

  do i = 0, Nxm1
    do j = 1, Nyp
      res(i, j) = 0.
      do k = 0, Nzp
        res(i, j) = res(i, j) + var(i, k, j) * dz(k)
      end do
    end do
  end do
  if (use_mpi) then
    call mpi_allreduce(mpi_in_place, res, Nx * Nyp, &
                       mpi_double_precision, mpi_sum, comm, ierror)
    call mpi_barrier(mpi_comm_world, ierror)
  end if

  res = res / Lz

end subroutine

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine end_run_mpi(flag)
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|

  include 'header'

  logical flag

  if (rank == 0) then
    call end_run(flag)
  end if
  call mpi_bcast(flag, 1, mpi_logical, 0, mpi_comm_world, ierror)

end
