!******************************************************************************|
! periodic.f90, the fully-periodic-box solvers for diablo.           VERSION 0.9
!
! These solvers were written by Tom Bewley and John Taylor.
!******************************************************************************|

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine init_per
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|

  include 'header'

  ! For forcing of homogeneous turbulence, set the target dissipation rate
  if (N_th > 0) then
    epsilon_target = ((1.d0 / dx(1))**4.d0) * (nu**3.d0) * (Pr(1))**(-2.d0)
    write (*, '("Epsilon_target = " ES26.18)') epsilon_target
  end if

  ! Aspect ratio set to unity (Divides Ro_inv)
  delta = 1.d0

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine rk_per_1
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  ! Main time-stepping algorithm for the fully periodic case
  ! INPUTS  (in Fourier space):  CUi, P, and (if k>1) CFi at (k-1)  (for i=1,2,3)
  ! OUTPUTS (in Fourier space):  CUi, P, and (if k<3) CFi at (k)
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  include 'header'
  real(rkind) temp1, temp2, temp3, temp4, temp5, temp6
  integer i, j, k, n

  ! Compute the RHS in Fourier Space, CRi.

  ! First, define the constants used for time-stepping
  temp1 = nu * h_bar(rk_step) / 2.0
  temp2 = beta_bar(rk_step) * h_bar(rk_step)
  temp3 = zeta_bar(rk_step) * h_bar(rk_step)
  temp4 = h_bar(rk_step)
  temp5 = 0.0

  do j = 0, twoNky
    do k = 0, twoNkz
      do i = 0, Nxp - 1 ! Nkx
        ! Start with the explicit part of the Crank-Nicolson viscous term and
        !  the pressure gradient treated with Explicit Euler:
        temp5 = 1 - temp1 * (kx2(i) + ky2(j) + kz2(k))
        cr1(i, k, j) = temp5 * cu1(i, k, j) - temp4 * (cikx(i) * cp(i, k, j))
        cr2(i, k, j) = temp5 * cu2(i, k, j) - temp4 * (ciky(j) * cp(i, k, j))
        cr3(i, k, j) = temp5 * cu3(i, k, j) - temp4 * (cikz(k) * cp(i, k, j))
        ! For each scalar, start with the explict part of the Crank-Nicolson
        ! diffusive term for each scalar
        do n = 1, N_th
          temp6 = 1 - (temp1 / Pr(n)) * (kx2(i) + ky2(j) + kz2(k))
          crth(i, k, j, n) = temp6 * cth(i, k, j, n)
        end do
      end do
    end do
    if (rk_step > 1) then
      do k = 0, twoNkz
        do i = 0, Nxp - 1 ! Nkx
          ! Add the term: ZETA_BAR(RK_STEP)*R(U(RK_STEP-1))
          cr1(i, k, j) = cr1(i, k, j) + temp3 * cf1(i, k, j)
          cr2(i, k, j) = cr2(i, k, j) + temp3 * cf2(i, k, j)
          cr3(i, k, j) = cr3(i, k, j) + temp3 * cf3(i, k, j)
          ! Do the same for each scalar:
          do n = 1, N_th
            crth(i, k, j, n) = crth(i, k, j, n) + temp3 * cfth(i, k, j, n)
          end do
        end do
      end do
    end if
  end do
  ! If we are considering a linear background scalar gradient then add
  ! the term owing to advection of the background state.
  ! This allows us to consider a mean scalar gradient (ie stratification)
  ! even though the vertical boundary conditions are periodic.
  ! (In this case the passive scalar is a perturbation from a linear
  ! gradient. This gradient and the vertical domain size are used to
  ! make the passive scalar nondimensional, so here the nondimensional
  ! gradient is equal to one
  do n = 1, N_th
    if (background_grad(n)) then
      ! If there is a background scalar gradient add advection term:
      do j = 0, twoNky
        do k = 0, twoNkz
          do i = 0, Nxp - 1 ! Nkx
            cfth(i, k, j, n) = -cu2(i, k, j)
          end do
        end do
      end do
    else
      ! Otherwise don't
      do j = 0, twoNky
        do k = 0, twoNkz
          do i = 0, Nxp - 1 ! Nkx
            cfth(i, k, j, n) = 0.d0
          end do
        end do
      end do
    end if
  end do

  ! Inverse transform to physical space to compute the nonlinear terms
  call fft_xzy_to_physical(cu1, u1)
  call fft_xzy_to_physical(cu2, u2)
  call fft_xzy_to_physical(cu3, u3)

  do n = 1, N_th
    call fft_xzy_to_physical(cth(0, 0, 0, n), th(0, 0, 0, n))
  end do

  ! Compute the nonlinear terms for the passive scalar equation
  ! Do this before the nonlinear momentum terms to use Fi as a working
  !  array before using it for the momentum equation.

  do n = 1, N_th
    do j = 0, Nypm1
      do k = 0, Nzp - 1
        do i = 0, Nxm1
          f1(i, k, j) = u1(i, k, j) * th(i, k, j, n)
          f2(i, k, j) = u2(i, k, j) * th(i, k, j, n)
          f3(i, k, j) = u3(i, k, j) * th(i, k, j, n)
        end do
      end do
    end do
    call fft_xzy_to_fourier(f1, cf1)
    call fft_xzy_to_fourier(f2, cf2)
    call fft_xzy_to_fourier(f3, cf3)
    do j = 0, twoNky
      do k = 0, twoNkz
        do i = 0, Nxp - 1 ! Nkx
          cfth(i, k, j, n) = cfth(i, k, j, n) - cikx(i) * cf1(i, k, j) &
                             - ciky(j) * cf2(i, k, j) &
                             - cikz(k) * cf3(i, k, j)

        end do
      end do
    end do
  end do
  ! The RHS vector for the TH equation is now ready

  ! Compute the nonlinear terms for the momentum equations
  do j = 0, Nypm1
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        f1(i, k, j) = u1(i, k, j) * u1(i, k, j)
        f2(i, k, j) = u1(i, k, j) * u2(i, k, j)
        f3(i, k, j) = u1(i, k, j) * u3(i, k, j)
        s1(i, k, j) = u2(i, k, j) * u2(i, k, j)
      end do
    end do
  end do
  call fft_xzy_to_fourier(f1, cf1)
  call fft_xzy_to_fourier(f2, cf2)
  call fft_xzy_to_fourier(f3, cf3)
  call fft_xzy_to_fourier(s1, cs1)
  ! Here we start constructing the R-K terms in CFi
  ! Note, that the order of the following operations are important
  do j = 0, twoNky
    do k = 0, twoNkz
      do i = 0, Nxp - 1 ! Nkx
        cf1(i, k, j) = -cikx(i) * cf1(i, k, j) &
                       - ciky(j) * cf2(i, k, j) &
                       - cikz(k) * cf3(i, k, j)
        cf2(i, k, j) = -cikx(i) * cf2(i, k, j) &
                       - ciky(j) * cs1(i, k, j)
        cf3(i, k, j) = -cikx(i) * cf3(i, k, j)
      end do
    end do
  end do
  ! At this point, F1,F2,F3 contain some of the nonlinear terms

  ! Compute the remaining nonlinear terms
  ! We cannot use F1,F2,F3 for storage, so use S1 as the working variable
  do j = 0, Nypm1
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        s1(i, k, j) = u2(i, k, j) * u3(i, k, j)
      end do
    end do
  end do
  call fft_xzy_to_fourier(s1, cs1)
  do j = 0, twoNky
    do k = 0, twoNkz
      do i = 0, Nxp - 1 ! Nkx
        cf2(i, k, j) = cf2(i, k, j) - cikz(k) * cs1(i, k, j)
        cf3(i, k, j) = cf3(i, k, j) - ciky(j) * cs1(i, k, j)
      end do
    end do
  end do
  do j = 0, Nypm1
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        s1(i, k, j) = u3(i, k, j) * u3(i, k, j)
      end do
    end do
  end do
  call fft_xzy_to_fourier(s1, cs1)
  do j = 0, twoNky
    do k = 0, twoNkz
      do i = 0, Nxp - 1 ! Nkx
        cf3(i, k, j) = cf3(i, k, j) - cikz(k) * cs1(i, k, j)
      end do
    end do
  end do
  ! Done with the computation of nonlinear terms

  ! Optionally, add forcing terms to the RHS
  ! Here we have U1, U2, U3, and TH in physical space
  call user_rhs_per_physical

  ! If the scalar is active (RI NE 0), add the bouyancy forcing term
  !  as explicit R-K
  do n = 1, N_th
    ! First, convert back to Fourier space
    call fft_xzy_to_fourier(th(0, 0, 0, n), cth(0, 0, 0, n))
    do j = 0, twoNky
      do k = 0, twoNkz
        do i = 0, Nxp - 1 ! Nkx
          cf2(i, k, j) = cf2(i, k, j) + cth(i, k, j, n)
        end do
      end do
    end do
  end do

  ! Now, add the R-K terms to the RHS
  do j = 0, twoNky
    do k = 0, twoNkz
      do i = 0, Nxp - 1 ! Nkx
        cr1(i, k, j) = cr1(i, k, j) + temp2 * cf1(i, k, j)
        cr2(i, k, j) = cr2(i, k, j) + temp2 * cf2(i, k, j)
        cr3(i, k, j) = cr3(i, k, j) + temp2 * cf3(i, k, j)
      end do
    end do
  end do

  ! Add R-K terms for the TH equation to the RHS
  do n = 1, N_th
    do j = 0, twoNky
      do k = 0, twoNkz
        do i = 0, Nxp - 1 ! Nkx
          crth(i, k, j, n) = crth(i, k, j, n) + temp2 * cfth(i, k, j, n)
        end do
      end do
    end do
  end do

  ! Computation of CRi complete.

  ! Now solve the implicit system for the intermediate field.
  ! (In the fully-periodic case, this is easy !)
  do j = 0, twoNky
    do k = 0, twoNkz
      do i = 0, Nxp - 1 ! Nkx
        temp5 = 1 + temp1 * (kx2(i) + ky2(j) + kz2(k))
        cu1(i, k, j) = cr1(i, k, j) / temp5
        cu2(i, k, j) = cr2(i, k, j) / temp5
        cu3(i, k, j) = cr3(i, k, j) / temp5
        do n = 1, N_th
          temp6 = 1 + (temp1 / Pr(n)) * (kx2(i) + ky2(j) + kz2(k))
          cth(i, k, j, n) = crth(i, k, j, n) / temp6
        end do
      end do
    end do
  end do
  ! First step of the Fractional Step algorithm complete.

  ! Begin second step of the Fractional Step algorithm, making u divergence free.

  ! Compute varphi, store in the variable CR1, and project velocity field
  call rem_div_per
  ! Then, update the pressure with phi.
  do j = 0, twoNky
    do k = 0, twoNkz
      do i = 0, Nxp - 1 ! Nkx
        cp(i, k, j) = cp(i, k, j) + cr1(i, k, j) / temp4
      end do
    end do
  end do
  ! Second step of the Fractional Step algorithm complete.

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine rk_per_2
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  ! Alternative time-stepping algorithm for the fully periodic case.
  ! INPUTS  (in Fourier space):  CUi, P, and (if k>1) CFi at (k-1)  (for i=1,2,3)
  ! OUTPUTS (in Fourier space):  CUi, P, and (if k<3) CFi at (k)
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine rem_div_per
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  include 'header'
  integer i, j, k
  real(rkind) temp5

  ! Compute phi, store in the variable CR1.
  ! Note the coefficient H_BAR is absorbed into phi.
  do j = 0, twoNky
    do k = 0, twoNkz
      do i = 0, Nxp - 1 ! Nkx
        temp5 = -(kx2(i) + ky2(j) + kz2(k) + eps)
        cr1(i, k, j) = (cikx(i) * cu1(i, k, j) + ciky(j) * cu2(i, k, j) + &
                        cikz(k) * cu3(i, k, j)) / temp5
      end do
    end do
    ! Then update the CUi to make velocity field divergence-free.
    do k = 0, twoNkz
      do i = 0, Nxp - 1 ! Nkx
        cu1(i, k, j) = cu1(i, k, j) - cikx(i) * cr1(i, k, j)
        cu2(i, k, j) = cu2(i, k, j) - ciky(j) * cr1(i, k, j)
        cu3(i, k, j) = cu3(i, k, j) - cikz(k) * cr1(i, k, j)
      end do
    end do
  end do

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine poisson_p_per
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  include 'header'
  integer i, j, k

  do j = 0, twoNky
    do k = 0, twoNkz
      do i = 0, Nxp - 1 ! Nkx
        cr1(i, k, j) = cikx(i) * cu1(i, k, j)
        cr2(i, k, j) = cikz(k) * cu3(i, k, j)
      end do
    end do
  end do

  call fft_xzy_to_physical(cr1, r1)
  call fft_xzy_to_physical(cr2, r2)

  do j = 0, Nypm1
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        p(i, k, j) = r1(i, k, j) * r1(i, k, j) + r1(i, k, j) * r2(i, k, j) + &
                     r2(i, k, j) * r2(i, k, j)
      end do
    end do
  end do
  do j = 0, twoNky
    do k = 0, twoNkz
      do i = 0, Nxp - 1 ! Nkx
        cf1(i, k, j) = ciky(j) * cu1(i, k, j)
        cf2(i, k, j) = cikx(i) * cu2(i, k, j)
        cf3(i, k, j) = cikz(k) * cu1(i, k, j)
        cr1(i, k, j) = cikx(i) * cu3(i, k, j)
        cr2(i, k, j) = cikz(k) * cu2(i, k, j)
        cr3(i, k, j) = ciky(j) * cu3(i, k, j)
      end do
    end do
  end do
  call fft_xzy_to_physical(cf1, f1)
  call fft_xzy_to_physical(cf2, f2)
  call fft_xzy_to_physical(cf3, f3)
  call fft_xzy_to_physical(cr1, r1)
  call fft_xzy_to_physical(cr2, r2)
  call fft_xzy_to_physical(cr3, r3)
  do j = 0, Nypm1
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        p(i, k, j) = 2 * (p(i, k, j) + f1(i, k, j) * f2(i, k, j) &
                          + f3(i, k, j) * r1(i, k, j) &
                          + r2(i, k, j) * r3(i, k, j))
      end do
    end do
  end do

  call fft_xzy_to_fourier(p, cp)
  do j = 0, twoNky
    do k = 0, twoNkz
      do i = 0, Nxp - 1 ! Nkx
        cp(i, k, j) = cp(i, k, j) / (kx2(i) + ky2(j) + kz2(k) + eps)
      end do
    end do
  end do
  cp(0, 0, 0) = 0.0

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine create_grid_per
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  include 'header'
  integer i, j, k

  write (*, '("Fourier in X")')
  do i = 0, Nx
    gx(i) = (i * Lx) / Nx
    dx(i) = Lx / Nx
    if (verbosity > 3) write (*, *) 'GX(', i, ') = ', gx(i)
  end do
  write (*, '("Fourier in Y")')
  do k = 0, Nz
    gz(k) = (k * Lz) / Nz
    dz(k) = Lz / Nz
    if (verbosity > 3) write (*, *) 'GY(', k, ') = ', gz(k)
  end do
  write (*, '("Fourier in Z")')
  do j = 0, Nyp
    gy(j) = (j * Ly) / Nyp
    dy(j) = Ly / Nyp
    if (verbosity > 3) write (*, *) 'GZ(', j, ') = ', gy(j)
  end do

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine input_per
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  include 'header'

  integer n
  real version, current_version

  ! Read in input parameters specific for channel flow case
  open (11, file='input_per.dat', form='formatted', status='old')
  ! Read input file.

  current_version = 2.0
  read (11, *)
  read (11, *)
  read (11, *)
  read (11, *)
  read (11, *) version
  if (version /= current_version) &
    stop 'Wrong input data format input_chan'
  read (11, *)
  read (11, *) time_ad_meth
  read (11, *)
  read (11, *) les_model_type
  read (11, *)
  read (11, *) IC_type, kick
  read (11, *)
  do n = 1, N_th
    read (11, *)
    read (11, *) background_grad(n)
  end do

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine save_stats_per(final)
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  include 'header'
  character(len=35) fname
  logical final
  real(rkind) mean_u1(0:Nypm1)
  real(rkind) mean_u2(0:Nypm1)
  real(rkind) mean_u3(0:Nypm1)
  real(rkind) mean_th(0:Nypm1, 1:N_th)
  real(rkind) mean_p(0:Nypm1)

  integer i, j, k, n
  real(rkind) uc, ubulk

  ! Note that this routine uses CFi and CFTH for storage, so it should
  ! only be used between full R-K timesteps

  write (*, '("Saving flow statistics.")')

  ! We are in the middle of a run, compile statistics

  ! Store the velocity in Fourier space in CRi
  do j = 0, twoNky
    do k = 0, twoNkz
      do i = 0, Nxp - 1 ! Nkx
        cr1(i, k, j) = cu1(i, k, j)
        cr2(i, k, j) = cu2(i, k, j)
        cr3(i, k, j) = cu3(i, k, j)
      end do
    end do
  end do

  ! Compute the vertical gradients in fourier space, store in CFi
  do j = 0, twoNky
    do k = 0, twoNkz
      do i = 0, Nxp - 1 ! Nkx
        cf1(i, k, j) = ciky(j) * cu1(i, k, j)
        cf2(i, k, j) = ciky(j) * cu2(i, k, j)
        cf3(i, k, j) = ciky(j) * cu3(i, k, j)
      end do
    end do
  end do
  ! Save the  scalar in fourier space in CFTH
  do n = 1, N_th
    do j = 0, twoNky
      do k = 0, twoNkz
        do i = 0, Nxp - 1 ! Nkx
          cfth(i, k, j, n) = cth(i, k, j, n)
        end do
      end do
    end do
  end do

  ! Now, convert the velocity and vertical gradients to physical space
  call fft_xzy_to_physical(cu1, u1)
  call fft_xzy_to_physical(cu2, u2)
  call fft_xzy_to_physical(cu3, u3)
  do n = 1, N_th
    call fft_xzy_to_physical(cth(0, 0, 0, n), th(0, 0, 0, n))
  end do
  call fft_xzy_to_physical(cf1, f1)
  call fft_xzy_to_physical(cf2, f2)
  call fft_xzy_to_physical(cf3, f3)

  ! First get the number of samples taken so far
  Nsamples = Nsamples + 1
  ! Get the mean velocity

  ! First, get the mean in physical space
  do j = 0, Nypm1
    mean_u1(j) = 0.d0
    mean_u2(j) = 0.d0
    mean_u3(j) = 0.d0
    do n = 1, N_th
      mean_th(j, n) = 0.d0
    end do
    mean_p(j) = 0.d0
    do i = 0, Nxm1
      do k = 0, Nzp - 1
        mean_u1(j) = mean_u1(j) + u1(i, k, j)
        mean_u2(j) = mean_u2(j) + u2(i, k, j)
        mean_u3(j) = mean_u3(j) + u3(i, k, j)
        do n = 1, N_th
          mean_th(j, n) = mean_th(j, n) + th(i, k, j, n)
        end do
        mean_p(j) = mean_p(j) + p(i, k, j)
      end do
    end do
    mean_u1(j) = mean_u1(j) / float(Nx * Nz)
    mean_u2(j) = mean_u2(j) / float(Nx * Nz)
    mean_u3(j) = mean_u3(j) / float(Nx * Nz)
    do n = 1, N_th
      mean_th(j, n) = mean_th(j, n) / float(Nx * Nz)
    end do
    mean_p(j) = mean_p(j) / float(Nx * Nz)
  end do

  do j = 0, Nypm1
    ubar(j) = (1./float(Nsamples)) * mean_u1(j) &
              + ((float(Nsamples) - 1.) / float(Nsamples)) * ubar(j)
    vbar(j) = (1./float(Nsamples)) * mean_u2(j) &
              + ((float(Nsamples) - 1.) / float(Nsamples)) * vbar(j)
    wbar(j) = (1./float(Nsamples)) * mean_u3(j) &
              + ((float(Nsamples) - 1.) / float(Nsamples)) * wbar(j)
    do n = 1, N_th
      thbar(j, n) = (1./float(Nsamples)) * mean_th(j, n) &
                    + ((float(Nsamples) - 1.) / float(Nsamples)) * thbar(j, n)
    end do
  end do

  ! Get the turbulent kinetic energy at each level
  do j = 0, Nypm1
    urms(j) = 0.
    vrms(j) = 0.
    wrms(j) = 0.
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        urms(j) = urms(j) + (abs(u1(i, k, j) - mean_u1(j)))**2.
        vrms(j) = vrms(j) + (abs(u2(i, k, j) - mean_u2(j)))**2.
        wrms(j) = wrms(j) + (abs(u3(i, k, j) - mean_u3(j)))**2.
      end do
    end do
    urms(j) = sqrt(urms(j) / (float(Nz) * float(Nx)))
    vrms(j) = sqrt(vrms(j) / (float(Nz) * float(Nx)))
    wrms(j) = sqrt(wrms(j) / (float(Nz) * float(Nx)))
  end do
  ! Get the bulk rms value
  urms_b = 0.
  do j = 1, Nypm1
    urms_b = urms_b + 0.5 * (urms(j) + urms(j - 1)) * (gy(j) - gy(j - 1))
  end do
  urms_b = urms_b / Ly
  ! If we are in 2d:
  if (Nyp == 1) urms_b = urms(0)

  ! Compute the Reynolds stress and mean velocity gradient
  do j = 0, Nypm1
    uv(j) = 0.
    uw(j) = 0.
    wv(j) = 0.
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        uv(j) = uv(j) + (u1(i, k, j) - mean_u1(j)) &
                * (u2(i, k, j) - mean_u2(j))
        wv(j) = wv(j) + (u3(i, k, j) - mean_u3(j)) &
                * (u2(i, k, j) - mean_u2(j))
        uw(j) = uw(j) + (u1(i, k, j) - mean_u1(j)) &
                * (u3(i, k, j) - mean_u3(j))
      end do
    end do
    uv(j) = uv(j) / (float(Nz) * float(Nx))
    uw(j) = uw(j) / (float(Nz) * float(Nx))
    wv(j) = wv(j) / (float(Nz) * float(Nx))
  end do

  ! Get the y-derivative of the mean velocity at GYF points
  do j = 1, Nyp - 2
    dudy(j) = (mean_u1(j + 1) - mean_u1(j - 1)) / (gy(j + 1) - gy(j - 1))
    dwdy(j) = (mean_u3(j + 1) - mean_u3(j - 1)) / (gy(j + 1) - gy(j - 1))
  end do
  j = 0
  dudy(j) = (mean_u1(j + 1) - mean_u1(Nypm1)) / (2.d0 * (gy(j + 1) - gy(j)))
  dwdy(j) = (mean_u3(j + 1) - mean_u3(Nypm1)) / (2.d0 * (gy(j + 1) - gy(j)))
  j = Nypm1
  dudy(j) = (mean_u1(0) - mean_u1(j - 1)) / (2.d0 * (gy(j) - gy(j - 1)))
  dwdy(j) = (mean_u3(0) - mean_u3(j - 1)) / (2.d0 * (gy(j) - gy(j - 1)))

  ! Get the mean square shear
  do j = 0, Nypm1
    shear(j) = 0.d0
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        shear(j) = shear(j) + f1(i, k, j)**2.d0 + f3(i, k, j)**2.d0
      end do
    end do
    shear(j) = shear(j) / float(Nx * Nz)
  end do

  ! Write out the bulk rms velocity
  write (*,  '("<U_rms> = " ES26.18)') urms_b

  ! Convert back to Fourier space
  do n = 1, N_th
    call fft_xzy_to_fourier(th(0, 0, 0, n), cth(0, 0, 0, n))
  end do


  ! Convert velocity back to Fourier space
  call fft_xzy_to_fourier(u1, cu1)
  call fft_xzy_to_fourier(u2, cu2)
  call fft_xzy_to_fourier(u3, cu3)

  call tkebudget_per

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine filter_per
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  ! This subroutine applies a filter to the highest wavenumbers
  ! It should be applied to the scalars in Fourier space
  ! The filter used is a sharpened raised cosine filter

  include 'header'

  integer i, j, k, n

  ! Variables for horizontal filtering
  real(rkind) sigma0

  ! Set the filtering constants for the all directions
  do n = 1, N_th
    do i = 0, Nxp - 1 ! Nkx
      do k = 0, twoNkz
        do j = 0, twoNky
          sigma0 = 0.5d0 * (1.d0 + &
                            cos(sqrt((kx(i) * Lx * 1.d0 / float(Nx))**2.d0 &
                                     + (kz(k) * Lz * 1.d0 / float(Nz))**2.d0 &
                                     + (ky(j) * Ly * 1.d0 / float(Nyp))**2.d0)))
          ! Apply a sharpened raised cosine filter
          cth(i, k, j, n) = cth(i, k, j, n) * &
                            sigma0**4.d0 * (35.d0 - 84.d0 * sigma0 &
                                            + 70.d0 * sigma0**2.d0 - 20.d0 * sigma0**3.d0)
        end do
      end do
    end do
  end do

  return
end

!----*|--.---------.---------.---------.---------.---------.---------.-|-------|
subroutine tkebudget_per
  !----*|--.---------.---------.---------.---------.---------.---------.-|-------|
  ! Note, it is important to only run this routine after complete R-K
  !  time advancement since F1 is overwritten which is needed between R-K steps

  include 'header'

  integer i, j, k
  real(rkind) epsilon_mean, k_eta

  ! Compute the turbulent dissipation rate, epsilon=nu*<du_i/dx_j du_i/dx_j>
  do j = 0, Nypm1
    epsilon(j) = 0.
  end do
  ! Store du/dx in CS1
  do j = 0, twoNky
    do k = 0, twoNkz
      do i = 0, Nxp - 1 ! Nkx
        cs1(i, k, j) = cikx(i) * cu1(i, k, j)
      end do
    end do
  end do
  ! Convert to physical space
  call fft_xzy_to_physical(cs1, s1)
  do j = 0, Nypm1
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        epsilon(j) = epsilon(j) + (s1(i, k, j)**2.0)
      end do
    end do
  end do
  ! Store dv/dx in CS1
  do j = 0, twoNky
    do k = 0, twoNkz
      do i = 0, Nxp - 1 ! Nkx
        cs1(i, k, j) = cikx(i) * cu2(i, k, j)
      end do
    end do
  end do
  ! Convert to physical space
  call fft_xzy_to_physical(cs1, s1)
  do j = 0, Nypm1
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        epsilon(j) = epsilon(j) + 0.5 * (s1(i, k, j)**2.0)
      end do
    end do
  end do
  ! Compute du/dy at GYF gridpoints, note remove mean
  do j = 0, twoNky
    do k = 0, twoNkz
      do i = 0, Nxp - 1 ! Nkx
        cf1(i, k, j) = ciky(j) * cu1(i, k, j)
      end do
    end do
  end do
  ! Convert to physical space
  call fft_xzy_to_physical(cf1, f1)
  do j = 0, Nypm1
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        epsilon(j) = epsilon(j) + 0.5 * (f1(i, k, j)**2.0)
        ! Cross term dvdx*dudy
        epsilon(j) = epsilon(j) + (s1(i, k, j) * f1(i, k, j))
      end do
    end do
  end do
  ! Store dw/dx in CS1
  do j = 0, twoNky
    do k = 0, twoNkz
      do i = 0, Nxp - 1 ! Nkx
        cs1(i, k, j) = cikx(i) * cu3(i, k, j)
      end do
    end do
  end do
  ! Convert to physical space
  call fft_xzy_to_physical(cs1, s1)
  do j = 0, Nypm1
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        epsilon(j) = epsilon(j) + 0.5 * (s1(i, k, j)**2.0)
      end do
    end do
  end do
  ! Compute du/dz at GYF gridpoints, note remove mean
  ! Store du/dz in CS1
  do j = 0, twoNky
    do k = 0, twoNkz
      do i = 0, Nxp - 1 ! Nkx
        cf1(i, k, j) = cikz(k) * cu1(i, k, j)
      end do
    end do
  end do
  ! Convert to physical space
  call fft_xzy_to_physical(cf1, f1)
  do j = 0, Nypm1
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        epsilon(j) = epsilon(j) + 0.5 * (f1(i, k, j)**2.0)
        ! Cross term dudz*dwdx
        epsilon(j) = epsilon(j) + s1(i, k, j) * f1(i, k, j)
      end do
    end do
  end do
  ! Compute dv/dy at GYF gridpoints, note remove mean
  do j = 0, twoNky
    do k = 0, twoNkz
      do i = 0, Nxp - 1 ! Nkx
        cs1(i, k, j) = ciky(j) * cu2(i, k, j)
      end do
    end do
  end do
  ! Convert to physical space
  call fft_xzy_to_physical(cs1, s1)
  do j = 0, Nypm1
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        epsilon(j) = epsilon(j) + (s1(i, k, j)**2.0)
      end do
    end do
  end do
  ! Compute dw/dy at GYF gridpoints, note remove mean
  do j = 0, twoNky
    do k = 0, twoNkz
      do i = 0, Nxp - 1 ! Nkx
        cs1(i, k, j) = ciky(j) * cu3(i, k, j)
      end do
    end do
  end do
  ! Convert to physical space
  call fft_xzy_to_physical(cs1, s1)
  do j = 0, Nypm1
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        epsilon(j) = epsilon(j) + 0.5 * (s1(i, k, j)**2.0)
      end do
    end do
  end do
  ! Store dv/dz in CF1
  do j = 0, twoNky
    do k = 0, twoNkz
      do i = 0, Nxp - 1 ! Nkx
        cf1(i, k, j) = cikz(k) * cu2(i, k, j)
      end do
    end do
  end do
  ! Convert to physical space
  call fft_xzy_to_physical(cf1, f1)
  do j = 0, Nypm1
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        epsilon(j) = epsilon(j) + 0.5 * (f1(i, k, j)**2.0)
        ! Cross term dvdz*dwdy
        epsilon(j) = epsilon(j) + s1(i, k, j) * f1(i, k, j)
      end do
    end do
  end do
  ! Store dw/dz in CS1
  do j = 0, twoNky
    do k = 0, twoNkz
      do i = 0, Nxp - 1 ! Nkx
        cs1(i, k, j) = cikz(k) * cu3(i, k, j)
      end do
    end do
  end do
  ! Convert to physical space
  call fft_xzy_to_physical(cs1, s1)
  do j = 0, Nypm1
    do k = 0, Nzp - 1
      do i = 0, Nxm1
        epsilon(j) = epsilon(j) + (s1(i, k, j)**2.0)
      end do
    end do
  end do
  do j = 0, Nypm1
    epsilon(j) = epsilon(j) / float(Nx * Nz)
  end do

  ! Write out the bulk rms velocity
  write (*,  '("<U_rms> = " ES26.18)') urms_b

  ! Get Kolmogorov wavelength
  epsilon_mean = nu * sum(epsilon(0:Nypm1)) / float(Nyp)

  k_eta = 2.d0 * pi * (nu**3.d0 / epsilon_mean)**(-0.25d0)

  write (*, '("Kolmogorov scale = " ES26.18)') 2.d0 * pi / k_eta

  return
end
